using SimpleJSON;
using System.Collections;
using UnityEngine;
using static EntryPointUtils;

namespace Naturalis
{
    sealed class TittyMagic : MVRScript
    {
        const string NATURALIS_VERSION = "1.1.8";

        bool _initialized;
        string _packagePrefix;

        public override void Init() => StartCoroutine(InitCo());

        IEnumerator InitCo()
        {
            yield return new WaitForEndOfFrame();
            while(SuperController.singleton.isLoading)
            {
                yield return null;
            }

            RegisterBool(new JSONStorableBool(Constant.INTERNAL_IS_MIGRATION_SCRIPT, true));
            string packageId = GetPackageId(manager, GetPluginId(name));
            _packagePrefix = string.IsNullOrEmpty(packageId) ? "" : $"{packageId}:/";
            _initialized = true;
        }

        public override void RestoreFromJSON(
            JSONClass jc,
            bool restorePhysical = true,
            bool restoreAppearance = true,
            JSONArray presetAtoms = null,
            bool setMissingToDefault = true
        ) => StartCoroutine(RestoreFromJSONCo(jc));

        IEnumerator RestoreFromJSONCo(JSONClass jc)
        {
            while(!_initialized)
            {
                yield return null;
            }

            var bootyMagic = this.FirstOwnAtomStorableByRegex(NaturalisBMRegex);
            string versionStr = jc[Constant.NATURALIS_VERSION].Value;

            /* If restoring Naturalis.cslist, let BootyMagic handle adding DLL */
            if(!bootyMagic)
            {
                AddNaturalis(manager, _packagePrefix);
            }

            yield return new WaitForSeconds(1);
            yield return RestorePlugin(jc, versionStr);
            yield return new WaitForSeconds(1.5f);

            /* If restoring Naturalis.cslist, let BootyMagic handle removal */
            if(!bootyMagic)
            {
                manager.RemovePluginWithUID(GetPluginId(name));
            }
        }

        IEnumerator RestorePlugin(JSONClass sourceJSON, string versionStr)
        {
            var targetStorable = this.FirstOwnAtomStorableByRegex(NaturalisTMRegex);
            while(!targetStorable)
            {
                yield return new WaitForSeconds(0.1f);
                targetStorable = this.FirstOwnAtomStorableByRegex(NaturalisTMRegex);
            }

            while(targetStorable.GetBoolParamValue(Constant.INTERNAL_IS_CALIBRATING))
            {
                yield return null;
            }

            yield return new WaitForSeconds(0.05f);
            if(sourceJSON != null)
            {
                var jc = ConvertTittyMagicJSON(sourceJSON, versionStr, NATURALIS_VERSION, pluginLabelJSON);
                targetStorable.RestoreFromJSON(jc);
            }
        }
    }
}
