using NAT.Common;
using NAT.Common.Components;
using System;
using System.Collections;
using UnityEngine;

namespace NAT
{
    sealed class SharedUIModManager : UIModManagerBase
    {
        Bootstrap _bootstrap;
        UIMod _femaleMorphsMod;
        UIMod _maleMorphsMod;
        UIMod _skinMaterials2Mod;
        float _morphsTabVisibleDuration;

        internal bool IsMorphsTabVisibleFor1Sec()
        {
            if(!initialized)
            {
                return false;
            }

            _morphsTabVisibleDuration = _femaleMorphsMod.IsVisible() || _maleMorphsMod.IsVisible()
                ? _morphsTabVisibleDuration + Time.deltaTime
                : 0f;
            return _morphsTabVisibleDuration > 1f;
        }

        internal void Init(Bootstrap bootstrap)
        {
            _bootstrap = bootstrap;
            parentUI = _bootstrap.Person.AtomUIContent;
            _femaleMorphsMod = NewUIMod(
                Constant.FEMALE_MORPHS,
                "Female Morphs",
                CreateFemaleMorphsPauseToggleCo,
                bootstrap.IsPluginEnabled,
                true
            );
            _maleMorphsMod = NewUIMod(
                Constant.MALE_MORPHS,
                "Male Morphs",
                CreateMaleMorphsPauseToggleCo,
                bootstrap.IsPluginEnabled,
                true
            );
            _skinMaterials2Mod = NewUIMod(
                Constant.SKIN_MATERIALS_2_MOVE_BTN_GROUP,
                "Skin Materials 2",
                MoveSkinMaterialsUIButtonGroupCo,
                bootstrap.IsPluginEnabled
            );

            foreach(var uiMod in uiMods)
            {
                uiMod.DeferApply();
            }

            if(!_bootstrap.IsPluginEnabled())
            {
                enabled = false;
            }

            initialized = true;
        }

        IEnumerator CreateFemaleMorphsPauseToggleCo(UIMod uiMod)
        {
            while(_bootstrap.Done != true)
            {
                yield return null;
            }

            try
            {
                _bootstrap.CreatePauseToggle(uiMod, _bootstrap.PauseFemaleMorphUpdatesJsb);
                uiMod.TargetUIEventsListener = uiMod.Target.gameObject.AddComponent<UnityEventsListener>();
            }
            catch(Exception e)
            {
                Loggr.Error($"Error modifying {uiMod.Id} UI: {e}");
            }
        }

        IEnumerator CreateMaleMorphsPauseToggleCo(UIMod uiMod)
        {
            while(_bootstrap.Done != true)
            {
                yield return null;
            }

            try
            {
                _bootstrap.CreatePauseToggle(uiMod, _bootstrap.PauseMaleMorphUpdatesJsb);
                uiMod.TargetUIEventsListener = uiMod.Target.gameObject.AddComponent<UnityEventsListener>();
            }
            catch(Exception e)
            {
                Loggr.Error($"Error modifying {uiMod.Id} UI: {e}");
            }
        }

        IEnumerator MoveSkinMaterialsUIButtonGroupCo(UIMod uiMod)
        {
            while(_bootstrap.Done != true)
            {
                yield return null;
            }

            var rightSide = uiMod.Target.Find("RightSide");
            uiMod.RepositionTransform(rightSide.Find("SavePanel"), new Vector2(-440, 70), new Vector2(-440, 670));
            uiMod.RepositionTransform(rightSide.Find("RestorePanel"), new Vector2(-230, 70), new Vector2(-230, 670));
            uiMod.RepositionTransform(rightSide.Find("Reset"), new Vector2(-70, 70), new Vector2(-70, 670));

            if(_bootstrap.Person.HasDummySkin() || !_bootstrap.IsPluginEnabled())
            {
                uiMod.Disable();
            }
        }

        internal void SetSkinMaterials2ModEnabled(bool value)
        {
            if(value)
            {
                if(!_bootstrap.Person.HasDummySkin())
                {
                    _skinMaterials2Mod.Enable();
                }
            }
            else
            {
                _skinMaterials2Mod.Disable();
            }
        }

        bool _enableSkinMaterials2Mod;

        internal void SetEnabled(bool enableSkinMaterials2Mod = false)
        {
            _enableSkinMaterials2Mod = enableSkinMaterials2Mod;
            enabled = true;
        }

        void OnEnable()
        {
            foreach(var uiMod in uiMods)
            {
                if(_enableSkinMaterials2Mod && uiMod.Id == Constant.SKIN_MATERIALS_2_MOVE_BTN_GROUP)
                {
                    if(!_bootstrap.Person.HasDummySkin())
                    {
                        uiMod.Enable();
                    }
                }
                else
                {
                    uiMod.Enable();
                }
            }
        }
    }
}
