using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.Models;
using NAT.Common.Storables;
using System.Collections.Generic;
using UnityEngine;
using static NAT.Direction;

namespace NAT.BM.Handlers
{
    class GluteForceMorphHandler : ForceHandler
    {
        GluteSoftPhysicsHandler _softPhysicsHandler;

        internal Dictionary<string, List<ConfigurableMorph>> MorphSets { get; private set; }

        internal JSONStorableStringChooser DirectionJssc { get; private set; }

        internal override void Init<T>(ProviderBase provider)
        {
            base.Init<T>(provider);
            _softPhysicsHandler = provider.GetInstance<GluteSoftPhysicsHandler>();
            BaseJsf = NewMultiplierJsf("Force Morphing Base Multiplier");
            UpJsf = NewMultiplierJsf("Force Morphing Up Multiplier");
            DownJsf = NewMultiplierJsf("Force Morphing Down Multiplier");
            DepthOutJsf = NewMultiplierJsf("Force Morphing Depth Out Multiplier");
            DepthInJsf = NewMultiplierJsf("Force Morphing Depth In Multiplier");
            SidewaysInJsf = NewMultiplierJsf("Force Morphing Sideways In Multiplier");
            SidewaysOutJsf = NewMultiplierJsf("Force Morphing Sideways Out Multiplier");
            LoadSettings();
        }

        protected override void OnMultiplierValueChanged(JSONStorableFloat jsf)
        {
            if(_softPhysicsHandler.IsLdpUpdateOn())
            {
                Script.SetCalibrationPending((StorableFloat) jsf);
            }
        }

        protected override void LoadSettings()
        {
            MorphSets = new Dictionary<string, List<ConfigurableMorph>>
            {
                { UP_L, UpForceMorphs("L") },
                { UP_R, UpForceMorphs("R") },
                { DOWN_L, DownForceMorphs("L") },
                { DOWN_R, DownForceMorphs("R") },
                { BACK_L, DepthInForceMorphs("L") },
                { BACK_R, DepthInForceMorphs("R") },
                { FORWARD_L, DepthOutForceMorphs("L") },
                { FORWARD_R, DepthOutForceMorphs("R") },
                { LEFT_L, LeftForceLeftMorphs() },
                { LEFT_R, LeftForceRightMorphs() },
                { RIGHT_L, RightForceLeftMorphs() },
                { RIGHT_R, RightForceRightMorphs() },
            };

            var directions = new List<string>
            {
                UP,
                // UP_C,
                DOWN,
                BACK,
                // BACK_C,
                FORWARD,
                // FORWARD_C,
                LEFT_L,
                RIGHT_L,
                LEFT_R,
                RIGHT_R,
            };
            DirectionJssc = new JSONStorableStringChooser("direction", directions, directions[0], "Direction");
        }

        ConfigurableMorph NewConfigurableMorph(string morphName, bool isNegative, float softMultiplier, float massMultiplier) =>
            new ConfigurableMorph(
                Person.GetGluteMorph(morphName),
                isNegative,
                new JSONStorableFloat("softMultiplier", softMultiplier, -3.00f, 3.00f),
                new JSONStorableFloat("massMultiplier", massMultiplier, -3.00f, 3.00f)
            );

        List<ConfigurableMorph> UpForceMorphs(string side)
        {
            const string prefix = MorphPrefix.UP;
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Glute Rotate Y {side}", true, -0.25f, -0.50f),
                NewConfigurableMorph($"{prefix} Glute UpDown {side}", true, -1.60f, 0.60f),
                NewConfigurableMorph($"{prefix} Glute Move Up {side}", false, 1.80f, -0.50f),
                NewConfigurableMorph($"{prefix} Glute Upper Spread {side}", false, 0.50f, -0.25f),
                NewConfigurableMorph($"{prefix} Glute Width {side}", false, 0.60f, 0.40f),
            };
        }

        List<ConfigurableMorph> DownForceMorphs(string side)
        {
            const string prefix = MorphPrefix.DN;
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Glute Rotate Y {side}", true, -0.50f, -1.00f),
                NewConfigurableMorph($"{prefix} Glute UpDown {side}", false, 0.00f, 0.00f),
                NewConfigurableMorph($"{prefix} Glute Move Down {side}", true, -0.15f, -0.30f),
                NewConfigurableMorph($"{prefix} Glute Height Lower {side}", true, -0.20f, -0.40f),
                NewConfigurableMorph($"{prefix} Glute Line {side}", false, 0.30f, 0.60f),
            };
        }

        List<ConfigurableMorph> DepthInForceMorphs(string side)
        {
            const string prefix = MorphPrefix.BK;
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Glute Scale Z {side}", true, -0.90f, -1.20f),
                NewConfigurableMorph($"{prefix} Glute Height {side}", false, 1.00f, 1.80f),
                NewConfigurableMorph($"{prefix} Glute Width {side}", false, 1.50f, 2.85f),
            };
        }

        List<ConfigurableMorph> DepthOutForceMorphs(string side)
        {
            const string prefix = MorphPrefix.FW;
            return new List<ConfigurableMorph>
            {

                NewConfigurableMorph($"{prefix} Glute Scale Z {side}", false, 0.80f, 1.60f),
                NewConfigurableMorph($"{prefix} Glute Height {side}", true, -0.36f, -0.48f),
                NewConfigurableMorph($"{prefix} Glute Width {side}", true, -0.50f, -0.67f),
            };
        }

        List<ConfigurableMorph> LeftForceLeftMorphs()
        {
            const string prefix = MorphPrefix.LT;
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Glute Rotate Z L", true, -0.25f, -0.50f),
                NewConfigurableMorph($"{prefix} Glute Move Out L", false, 1.00f, -0.20f),
            };
        }

        List<ConfigurableMorph> LeftForceRightMorphs()
        {
            const string prefix = MorphPrefix.LT;
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Glute Rotate Z R", true, -0.45f, -0.40f),
                NewConfigurableMorph($"{prefix} Glute Move In R", false, 1.10f, -0.40f),
            };
        }

        List<ConfigurableMorph> RightForceLeftMorphs()
        {
            const string prefix = MorphPrefix.RT;
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Glute Rotate Z L", true, -0.45f, -0.40f),
                NewConfigurableMorph($"{prefix} Glute Move In L", false, 1.10f, -0.40f),
            };
        }

        List<ConfigurableMorph> RightForceRightMorphs()
        {
            const string prefix = MorphPrefix.RT;
            return new List<ConfigurableMorph>
            {
                NewConfigurableMorph($"{prefix} Glute Rotate Z R", true, -0.25f, -0.50f),
                NewConfigurableMorph($"{prefix} Glute Move Out R", false, 1.00f, -0.20f),
            };
        }

        internal override void PreRegisterParams()
        {
            Script.PreRegisterMorphsParam(BaseJsf);
            Script.PreRegisterMorphsParam(Script.UpDownAngleOffsetJsf);
            Script.PreRegisterMorphsParam(Script.DepthOffsetJsf);
            Script.PreRegisterMorphsParam(Script.LeftRightAngleOffsetJsf);
            Script.PreRegisterMorphsParam(UpJsf);
            Script.PreRegisterMorphsParam(DownJsf);
            Script.PreRegisterMorphsParam(DepthOutJsf);
            Script.PreRegisterMorphsParam(DepthInJsf);
            Script.PreRegisterMorphsParam(SidewaysInJsf);
            Script.PreRegisterMorphsParam(SidewaysOutJsf);
        }

        float _mass;
        float _softness;

        internal float UpLimit { get; private set; }
        internal float DepthOutLimit { get; private set; }
        internal float SidewaysOutLimit { get; private set; }
        internal float SidewaysInLimit { get; private set; }
        internal float UpDownNormalizer { get; private set; }
        internal float DepthInNormalizer { get; private set; }
        internal float DepthOutNormalizer { get; private set; }
        internal float SidewaysNormalizer { get; private set; }

        float _upMultiplier;
        float _downMultiplier;
        float _depthInMultiplier;
        float _depthOutMultiplier;
        float _sidewaysInMultiplier;
        float _sidewaysOutMultiplier;

        static float SoftnessMultiplier(float x) => Calc.InverseSmoothStep(x, 0.80f, 0.03f);

        internal void SetMultipliers(float mass, float softness)
        {
            _mass = mass;
            _softness = SoftnessMultiplier(softness);
            SetUpDownMultipliers();
            SetDepthInMultiplier();
            SetDepthOutMultiplier();
            SetSidewaysMultipliers();
        }

        void SetUpDownMultipliers()
        {
            UpLimit = 30f;
            UpDownNormalizer = 50 * Mathf.Lerp(1, 1.20f, _mass);
            float softnessMultiplier = Mathf.Lerp(0.50f, 1.25f, _softness);
            _upMultiplier = softnessMultiplier * 1.50f * Mathf.Lerp(1, 1.80f, _mass);
            _downMultiplier = softnessMultiplier * 1.80f * Mathf.Lerp(1, 1.80f, _mass);
        }

        void SetDepthInMultiplier()
        {
            DepthInNormalizer = 0.045f * Mathf.Lerp(1, 3.00f, _mass);
            float softnessMultiplier = Mathf.Lerp(0.50f, 1.00f, _softness);
            _depthInMultiplier = softnessMultiplier * 4.40f * Mathf.Lerp(1, 0.50f, _mass);
        }

        void SetDepthOutMultiplier()
        {
            DepthOutLimit = 0.020f;
            DepthOutNormalizer = 0.045f * Mathf.Lerp(1, 3.00f, _mass);
            float softnessMultiplier = Mathf.Lerp(0.65f, 1.125f, _softness);
            _depthOutMultiplier = softnessMultiplier * 2.20f * Mathf.Lerp(1, 1.50f, _mass);
        }

        void SetSidewaysMultipliers()
        {
            SidewaysInLimit = 30f;
            SidewaysOutLimit = Mathf.Lerp(30f, 40f, _mass);
            SidewaysNormalizer = 50 * Mathf.Lerp(1, 1.20f, _mass);
            float softnessMultiplier = Mathf.Lerp(0.50f, 1.25f, _softness);
            _sidewaysInMultiplier = softnessMultiplier * 1.10f * Mathf.Lerp(1, 1.80f, _mass);
            _sidewaysOutMultiplier = softnessMultiplier * 0.80f * Mathf.Lerp(1, 1.80f, _mass);
        }

        internal void Execute()
        {
            AdjustUpMorphs();
            AdjustDownMorphs();
            AdjustDepthOutMorphs();
            AdjustDepthInMorphs();
            AdjustSidewaysMorphs();
        }

        static float UpDownForceEffectCurve(float x) => Calc.Exponential1(x, 1.60f, 2.26f, 1.23f, 0.96f);

        float UpEffect(float angle) =>
            _upMultiplier *
            UserUpMultiplier() *
            UpDownForceEffectCurve(Mathf.Min(Mathf.Abs(angle), UpLimit) / UpDownNormalizer);

        void AdjustUpMorphs()
        {
            if(Person.LeftGlute.AngleY <= 0)
            {
                UpdateMorphs(UP_L, UpEffect(Person.LeftGlute.AngleY));
            }
            else
            {
                ResetMorphs(UP_L);
            }

            if(Person.RightGlute.AngleY <= 0)
            {
                UpdateMorphs(UP_R, UpEffect(Person.RightGlute.AngleY));
            }
            else
            {
                ResetMorphs(UP_R);
            }
        }

        float DownEffect(float angle) => _downMultiplier *
            UserDownMultiplier() *
            UpDownForceEffectCurve(Mathf.Abs(angle) / UpDownNormalizer);

        void AdjustDownMorphs()
        {
            if(Person.LeftGlute.AngleY <= 0)
            {
                ResetMorphs(DOWN_L);
            }
            else
            {
                UpdateMorphs(DOWN_L, DownEffect(Person.LeftGlute.AngleY));
            }

            if(Person.RightGlute.AngleY <= 0)
            {
                ResetMorphs(DOWN_R);
            }
            else
            {
                UpdateMorphs(DOWN_R, DownEffect(Person.RightGlute.AngleY));
            }
        }

        static float DepthForceEffectCurve(float x) => Calc.Exponential1(x, 1.60f, 2.26f, 1.23f, 0.96f);

        float DepthInEffect(float depthDiff) =>
            _depthInMultiplier *
            UserDepthInMultiplier() *
            DepthForceEffectCurve(Mathf.Abs(depthDiff) / DepthInNormalizer);

        void AdjustDepthInMorphs()
        {
            if(Person.LeftGlute.DepthDiff >= 0)
            {
                ResetMorphs(BACK_L);
            }
            else
            {
                UpdateMorphs(BACK_L, DepthInEffect(Person.LeftGlute.DepthDiff));
            }

            if(Person.RightGlute.DepthDiff >= 0)
            {
                ResetMorphs(BACK_R);
            }
            else
            {
                UpdateMorphs(BACK_R, DepthInEffect(Person.RightGlute.DepthDiff));
            }
        }

        float DepthOutEffect(float depthDiff) =>
            _depthOutMultiplier *
            UserDepthOutMultiplier() *
            DepthForceEffectCurve(Mathf.Min(Mathf.Abs(depthDiff), DepthOutLimit) / DepthOutNormalizer);

        void AdjustDepthOutMorphs()
        {
            if(Person.LeftGlute.DepthDiff >= 0)
            {
                UpdateMorphs(FORWARD_L, DepthOutEffect(Person.LeftGlute.DepthDiff));
            }
            else
            {
                ResetMorphs(FORWARD_L);
            }

            if(Person.RightGlute.DepthDiff >= 0)
            {
                UpdateMorphs(FORWARD_R, DepthOutEffect(Person.RightGlute.DepthDiff));
            }
            else
            {
                ResetMorphs(FORWARD_R);
            }
        }

        static float SidewaysForceEffectCurve(float x) => Calc.Exponential1(x, 1.60f, 2.26f, 1.23f, 0.96f);

        float SidewaysInEffect(float angle) =>
            _sidewaysInMultiplier *
            UserSidewaysInMultiplier() *
            SidewaysForceEffectCurve(Mathf.Min(Mathf.Abs(angle), SidewaysInLimit) / SidewaysNormalizer);

        float SidewaysOutEffect(float angle) =>
            _sidewaysOutMultiplier *
            UserSidewaysOutMultiplier() *
            SidewaysForceEffectCurve(Mathf.Min(Mathf.Abs(angle), SidewaysOutLimit) / SidewaysNormalizer);

        void AdjustSidewaysMorphs()
        {
            if(Person.LeftGlute.AngleX <= 0)
            {
                ResetMorphs(LEFT_L);
                UpdateMorphs(RIGHT_L, SidewaysInEffect(Person.LeftGlute.AngleX));
            }
            else
            {
                ResetMorphs(RIGHT_L);
                UpdateMorphs(LEFT_L, SidewaysOutEffect(Person.LeftGlute.AngleX));
            }

            if(Person.RightGlute.AngleX <= 0)
            {
                ResetMorphs(LEFT_R);
                UpdateMorphs(RIGHT_R, SidewaysOutEffect(Person.RightGlute.AngleX));
            }
            else
            {
                ResetMorphs(RIGHT_R);
                UpdateMorphs(LEFT_R, SidewaysInEffect(Person.RightGlute.AngleX));
            }
        }

        void UpdateMorphs(string direction, float effect)
        {
            float halfEffect = 0.5f * effect;
            foreach(var configurableMorph in MorphSets[direction])
            {
                configurableMorph.SetValue(halfEffect, _mass, 1);
            }
        }

        internal override void ResetAll()
        {
            if(MorphSets == null)
            {
                return;
            }

            foreach(var pair in MorphSets)
            {
                foreach(var configurableMorph in pair.Value)
                {
                    configurableMorph.Reset();
                }
            }
        }

        void ResetMorphs(string direction)
        {
            foreach(var configurableMorph in MorphSets[direction])
            {
                configurableMorph.Reset();
            }
        }
    }
}
