using NAT.BM.Handlers;
using NAT.Common;
using NAT.Common.UI;
using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace NAT.BM.UI
{
    sealed class DevMorphWindow : Window
    {
        readonly GluteForceMorphHandler _forceMorphHandler;
        readonly Dictionary<string, UIDynamic> _sectionElements = new Dictionary<string, UIDynamic>();
        readonly JSONStorableString _morphDisplayValueJss = new JSONStorableString("morphDisplayValue", "");

        internal DevMorphWindow(ScriptBase script) : base(script)
        {
            _forceMorphHandler = script.Provider.GetInstance<GluteForceMorphHandler>();
        }

        static string DirectionKey(string value) => value.EndsWith("_C") || value.StartsWith(Direction.LEFT) || value.StartsWith(Direction.RIGHT)
            ? value
            : $"{value}_L";

        protected override void OnBuild()
        {
            BuildLeftSide();
            BuildRightSide();
            RebuildSection(_forceMorphHandler.DirectionJssc.val);
        }

        void BuildLeftSide(bool rightSide = false)
        {
            AddElement(() =>
            {
                var button = script.CreateButton("Print", rightSide);
                button.AddListener(
                    () =>
                    {
                        var configs = _forceMorphHandler.MorphSets[DirectionKey(_forceMorphHandler.DirectionJssc.val)];
                        string[] arr = configs.Select(config => config.ToCodeString("GetGluteMorph")).ToArray();
                        Debug.Log(string.Join("", arr));
                    }
                );
                button.height = 55;
                return button;
            });

            AddElement(() =>
            {
                var button = script.CreateButton("Zero All", rightSide);
                button.AddListener(
                    () =>
                    {
                        var configs = _forceMorphHandler.MorphSets[DirectionKey(_forceMorphHandler.DirectionJssc.val)];
                        foreach(var config in configs)
                        {
                            config.MassMultiplierJsf.val = 0;
                            config.SoftMultiplierJsf.val = 0;
                        }
                    }
                );
                button.height = 55;
                return button;
            });
        }

        void BuildRightSide(bool rightSide = true)
        {
            AddSpacer(55, rightSide);
            AddElement(_forceMorphHandler.DirectionJssc.name, () =>
            {
                var popup = script.CreateFilterablePopup(_forceMorphHandler.DirectionJssc, rightSide);
                script.ConfigureCustomPopup(popup, 360f, false);
                script.AddPrevNextButtons(popup, true);
                popup.popup.onValueChangeHandlers += RebuildSection;
                return popup;
            });
        }

        void RebuildSection(string value)
        {
            ClearSection();
            var leftConfigs = _forceMorphHandler.MorphSets[DirectionKey(value)];
            foreach(var config in leftConfigs)
            {
                AddSectionElement($"{config.MorphName} Label", () =>
                    CreateValueTextField(_morphDisplayValueJss, 28, 50, false)
                );

                AddSectionElement($"{config.MorphName} {config.IsNegativeJsb.name}", () =>
                {
                    var toggle = script.CreateToggle(config.IsNegativeJsb, true);
                    toggle.label = "Is Negative";
                    return toggle;
                });

                AddSectionElement($"{config.MorphName} {config.SoftMultiplierJsf.name}", () =>
                {
                    var slider = script.CreateSlider(config.SoftMultiplierJsf);
                    slider.valueFormat = "F2";
                    slider.label = "Softness Multiplier";
                    return slider;
                });

                AddSectionElement($"{config.MorphName} {config.MassMultiplierJsf.name}", () =>
                {
                    var slider = script.CreateSlider(config.MassMultiplierJsf, true);
                    slider.valueFormat = "F2";
                    slider.label = "Mass Multiplier";
                    return slider;
                });
            }
        }

        void AddSectionElement(string key, Func<UIDynamic> createElement)
        {
            _sectionElements[key] = createElement();
        }

        internal override void Clear()
        {
            base.Clear();
            ClearSection();
            RemoveDirectionPopupChangeHandler();
        }

        void RemoveDirectionPopupChangeHandler()
        {
            var popup = GetUIDynamicAs<UIDynamicPopup>(_forceMorphHandler.DirectionJssc.name);
            if(popup)
            {
                popup.popup.visible = false;
                popup.popup.onValueChangeHandlers -= RebuildSection;
            }
        }

        void ClearSection()
        {
            foreach(var pair in _sectionElements)
            {
                script.RemoveElement(pair.Value);
            }

            _sectionElements.Clear();
        }
    }
}
