using System;
using System.Collections.Generic;

namespace NAT.BM
{
    static class VertexIndices
    {
        internal static IEnumerable<int> AllSoftVerts(string side, string group)
        {
            switch(side)
            {
                case Side.LEFT: return AllLeftSoftVerts(group);
                case Side.RIGHT: return AllRightSoftVerts(group);
                default: throw new ArgumentException($"Invalid side {side}");
            }
        }

        static IEnumerable<int> AllLeftSoftVerts(string group)
        {
            switch(group)
            {
                case SoftVerticesGroupName.OUTER:
                    return new[]
                    {
                        265, 274, 277, 287, 294, 311, 338, 2388, 4036, 7844, 7872, 7881, 8188, 8191, 8193, 8212, 8246, 21567, 21712, 22817, 22856,
                    };
                case SoftVerticesGroupName.MAIN:
                    return new[]
                    {
                        272, 281, 283, 288, 296, 337, 2390, 2391, 2392, 4039, 4040, 4042, 7172, 7173, 7175, 7177, 7821, 7824,
                        7825, 7826, 7833, 7838, 7839, 7845, 7849, 7853, 7864, 7866, 7876, 7877, 8247, 8249, 8250, 8725, 8729,
                    };
                case SoftVerticesGroupName.CENTER:
                    return new[]
                    {
                        7852, 7858, 7862, 7865, 7875, 7878,
                    };
                case SoftVerticesGroupName.LOWER:
                    return new[]
                    {
                        7189, 7820, 7854, 7874, 8187, 8195, 8197, 8252, 21568, 21713, 21714, 21846, 22200, 22201, 22202,
                        22203, 22204, 22859, 22876, 22879, 22886, 22895, 22897, 22901, 22904, 22911, 22912, 22857,
                    };
                default: throw new ArgumentException($"Invalid group {group}");
            }
        }

        static IEnumerable<int> AllRightSoftVerts(string group)
        {
            switch(group)
            {
                case SoftVerticesGroupName.OUTER:
                    return new[]
                    {
                        11193, 11202, 11205, 11215, 11222, 11239, 11266, 13218, 14812, 18524, 18552,
                        18561, 18859, 18862, 18864, 18883, 18917, 21602, 22309, 22819, 22855,
                    };
                case SoftVerticesGroupName.MAIN:
                    return new[]
                    {
                        11200, 11209, 11211, 11216, 11224, 11265, 13220, 13221, 13222, 14815, 14816, 14818,
                        17879, 17880, 17882, 17884, 18501, 18504, 18505, 18506, 18513, 18518, 18519, 18525,
                        18529, 18533, 18544, 18546, 18556, 18557, 18918, 18920, 18921, 19396, 19400,
                    };
                case SoftVerticesGroupName.CENTER:
                    return new[]
                    {
                        18532, 18538, 18542, 18545, 18555, 18558,
                    };
                case SoftVerticesGroupName.LOWER:
                    return new[]
                    {
                        17896, 18500, 18534, 18554, 18858, 18866, 18868, 18923, 21603, 22310, 22311, 22449, 22791, 22792,
                        22793, 22794, 22795, 22860, 22878, 22880, 22890, 22894, 22896, 22900, 22905, 22916, 22917, 22858,
                    };
                default: throw new ArgumentException($"Invalid group {group}");
            }
        }

        internal static IEnumerable<int> PositionalSoftVerts(string side, string group)
        {
            switch(side)
            {
                case Side.LEFT: return PositionalLeftSoftVerts(group);
                case Side.RIGHT: return PositionalRightSoftVerts(group);
                default: throw new ArgumentException($"Invalid side {side}");
            }
        }

        static IEnumerable<int> PositionalLeftSoftVerts(string group)
        {
            switch(group)
            {
                case SoftVerticesGroupName.OUTER:
                    return new[]
                    {
                        265, 287, 294, 338, 2388, 4036, 7844, 7872, 7881, 8188, 8191, 8193, 8212, 8246, 22817, 22856,
                    };
                case SoftVerticesGroupName.MAIN:
                    return new[]
                    {
                        272, 281, 296, 2390, 2392, 4039, 4042, 7172, 7175, 7177, 7824, 7825, 7826, 7833, 7849, 7866, 7877, 8250, 8729,
                    };
                default: throw new ArgumentException($"Invalid group {group}");
            }
        }

        static IEnumerable<int> PositionalRightSoftVerts(string group)
        {
            switch(group)
            {
                case SoftVerticesGroupName.OUTER:
                    return new[]
                    {
                        11193, 11215, 11222, 11266, 13218, 14812, 18524, 18552, 18561, 18859, 18862, 18864, 18883, 18917, 22819, 22855,
                    };
                case SoftVerticesGroupName.MAIN:
                    return new[]
                    {
                        11200, 11209, 11224, 13220, 13222, 14815, 14818, 17879, 17882, 17884,
                        18504, 18505, 18506, 18513, 18529, 18546, 18557, 18921, 19400,
                    };
                default: throw new ArgumentException($"Invalid group {group}");
            }
        }

        // 1 center vert, N main verts, 1 outer vert
        internal static IEnumerable<int[]> CenterToOuterVerts(string side)
        {
            switch(side)
            {
                case Side.LEFT:
                    return new[]
                    {
                        new[] { 7875, 7876, 7881 },
                        new[] { 7875, 288, 287 },
                        new[] { 7875, 4042, 4040, 4039, 4036 },
                        new[] { 7878, 2392, 2391, 2390, 2388 },
                        new[] { 7878, 272, 265 },
                        new[] { 7878, 283, 338 },
                        new[] { 7878, 7877, 22856 },
                        new[] { 7858, 7853, 21712 },
                        new[] { 7865, 7821, 21567 },
                        new[] { 7865, 337, 8188 },
                        new[] { 7865, 7825, 7838, 7175, 8249, 8193 },
                        new[] { 7862, 7826, 7839, 7177, 8247, 8191 },
                        new[] { 7862, 7173, 8246 },
                        new[] { 7862, 7845, 8212 },
                        new[] { 7862, 7824, 7844 },
                        new[] { 7862, 7864, 7872 },
                    };
                case Side.RIGHT:
                    return new[]
                    {
                        new[] { 18555, 18556, 18561 },
                        new[] { 18555, 11216, 11215 },
                        new[] { 18555, 14818, 14816, 14815, 14812 },
                        new[] { 18558, 13222, 13221, 13220, 13218 },
                        new[] { 18558, 11200, 11193 },
                        new[] { 18558, 11211, 11266 },
                        new[] { 18558, 18557, 22855 },
                        new[] { 18538, 18533, 22309 },
                        new[] { 18545, 18501, 21602 },
                        new[] { 18545, 11265, 18859 },
                        new[] { 18545, 18505, 18518, 17882, 18920, 18864 },
                        new[] { 18542, 18506, 18519, 17884, 18918, 18862 },
                        new[] { 18542, 17880, 18917 },
                        new[] { 18542, 18525, 18883 },
                        new[] { 18542, 18504, 18524 },
                        new[] { 18542, 18544, 18552 },
                    };
                default: throw new ArgumentException($"Invalid side {side}");
            }
        }

        // connected verts for calculating averages from
        internal static Dictionary<int, int[]> AdjacentVerts(string side)
        {
            switch(side)
            {
                case Side.LEFT:
                    return new Dictionary<int, int[]>
                    {
                        { 274, new[] { 265, 272 } },
                        { 277, new[] { 265, 338 } },
                        { 281, new[] { 288, 7876 } },
                        { 294, new[] { 287, 7881 } },
                        { 296, new[] { 287, 288 } },
                        { 311, new[] { 7844, 8212 } },
                        { 7172, new[] { 7173, 8246 } },
                        { 7833, new[] { 337, 7821 } },
                        { 7849, new[] { 7864, 7876 } },
                        { 7852, new[] { 7862, 7875 } },
                        { 7866, new[] { 7821, 7853 } },
                        { 8250, new[] { 337, 8188 } },
                        { 8725, new[] { 7173, 8212, 8246 } },
                        { 8729, new[] { 7173, 7844, 8212 } },
                        { 22817, new[] { 21567, 21712 } },
                    };
                case Side.RIGHT:
                    return new Dictionary<int, int[]>
                    {
                        { 11202, new[] { 11193, 11200 } },
                        { 11205, new[] { 11193, 11266 } },
                        { 11209, new[] { 11216, 18556 } },
                        { 11222, new[] { 11215, 18561 } },
                        { 11224, new[] { 11215, 11216 } },
                        { 11239, new[] { 18524, 18883 } },
                        { 17879, new[] { 17880, 18917 } },
                        { 18513, new[] { 11265, 18501 } },
                        { 18529, new[] { 18544, 18556 } },
                        { 18532, new[] { 18542, 18555 } },
                        { 18546, new[] { 18501, 18533 } },
                        { 18921, new[] { 11265, 18859 } },
                        { 19396, new[] { 17880, 18883, 18917 } },
                        { 19400, new[] { 17880, 18524, 18883 } },
                        { 22819, new[] { 21602, 22309 } },
                    };
                default: throw new ArgumentException($"Invalid side {side}");
            }
        }

        internal static Dictionary<int, int[]> NearbyVerts(string side)
        {
            switch(side)
            {
                case Side.LEFT:
                    return new Dictionary<int, int[]>
                    {
                        { 265, new[] { 272, 274, 2388 } },
                        { 272, new[] { 265, 274, 277, 283, 338, 2391, 7877 } },
                        { 274, new[] { 265, 272, 277, 283, 2390 } },
                        { 277, new[] { 272, 274, 283, 338, 7874 } },
                        { 281, new[] { 288, 294, 296, 4042, 7849, 7876 } },
                        { 283, new[] { 272, 274, 277, 338, 2392, 7853, 7874, 7877 } },
                        { 287, new[] { 288, 294, 296, 4036 } },
                        { 288, new[] { 281, 287, 294, 296, 4040, 7876 } },
                        { 294, new[] { 281, 287, 288, 296, 7881 } },
                        { 296, new[] { 281, 287, 288, 294, 4039 } },
                        { 311, new[] { 7173, 7844, 8212, 8725, 8729 } },
                        { 337, new[] { 7175, 7189, 7821, 7833, 8188, 8250, 8252, 21567 } },
                        { 338, new[] { 272, 277, 283, 7854, 7877 } },
                        { 2388, new[] { 265, 2390, 2391, 4036 } },
                        { 2390, new[] { 274, 2388, 2391, 2392, 4039 } },
                        { 2391, new[] { 272, 2388, 2390, 2392, 4040, 7878 } },
                        { 2392, new[] { 283, 2390, 2391, 4042, 7858, 7878 } },
                        { 4036, new[] { 287, 2388, 4039, 4040 } },
                        { 4039, new[] { 296, 2390, 4036, 4040, 4042 } },
                        { 4040, new[] { 288, 2391, 4036, 4039, 4042, 7875 } },
                        { 4042, new[] { 281, 2392, 4039, 4040, 7852, 7875 } },
                        { 7172, new[] { 7173, 7845, 8212, 8246, 8247, 8725, 8729 } },
                        { 7173, new[] { 311, 7172, 7177, 7824, 7845, 8246, 8725, 8729 } },
                        { 7175, new[] { 337, 7177, 7825, 7838, 8193, 8249 } },
                        { 7177, new[] { 7173, 7175, 7826, 7839, 8191, 8247 } },
                        { 7189, new[] { 337, 7820, 7833, 8187, 8197, 8250, 8252, 21567, 21568, 22817 } },
                        { 7872, new[] { 7844, 7864, 7881 } },
                        { 7875, new[] { 4040, 4042, 7852, 7862, 7876, 7878 } },
                        { 7876, new[] { 281, 288, 7849, 7864, 7875, 7881 } },
                        { 7877, new[] { 272, 283, 338, 7853, 7854, 7866, 7874, 7878, 22856 } },
                        { 7878, new[] { 2391, 2392, 7858, 7865, 7875, 7877 } },
                        { 7820, new[] { 7189, 7821, 7833, 7854, 7866, 22817, 22857, 22879 } },
                        { 7821, new[] { 337, 7820, 7825, 7833, 7853, 7866, 22857 } },
                        { 7824, new[] { 7173, 7826, 7845, 7849, 7864 } },
                        { 7825, new[] { 7175, 7821, 7826, 7838, 7858, 7865 } },
                        { 7826, new[] { 7177, 7824, 7825, 7839, 7852, 7862 } },
                        { 7833, new[] { 337, 7189, 7820, 7821, 7838, 7866, 8250 } },
                        { 7838, new[] { 7175, 7825, 7833, 7839, 7865, 8249 } },
                        { 7839, new[] { 7177, 7826, 7838, 7845, 7862, 8247 } },
                        { 7844, new[] { 311, 7845, 7872, 8212, 8729 } },
                        { 7845, new[] { 7172, 7173, 7824, 7839, 7844, 7864, 8729 } },
                        { 7849, new[] { 281, 7824, 7852, 7864, 7876 } },
                        { 7852, new[] { 4042, 7826, 7849, 7858, 7862, 7875 } },
                        { 7853, new[] { 283, 7821, 7854, 7858, 7866, 7874, 7877, 21712 } },
                        { 7854, new[] { 338, 7820, 7853, 7866, 7874, 7877, 21712, 22856, 22897 } },
                        { 7858, new[] { 2392, 7825, 7852, 7853, 7865, 7878 } },
                        { 7862, new[] { 7826, 7839, 7852, 7864, 7865, 7875 } },
                        { 7864, new[] { 7824, 7845, 7849, 7862, 7872, 7876 } },
                        { 7865, new[] { 7825, 7838, 7858, 7862, 7866, 7878 } },
                        { 7866, new[] { 7820, 7821, 7833, 7853, 7854, 7865, 7877 } },
                        { 7874, new[] { 277, 283, 7853, 7854, 7877, 21712, 22856, 22895 } },
                        { 7881, new[] { 294, 7872, 7876 } },
                        { 8187, new[] { 7189, 8188, 8197, 8250, 8252 } },
                        { 8188, new[] { 337, 8187, 8197, 8193, 8250, 8252 } },
                        { 8191, new[] { 7177, 8193, 8246, 8247 } },
                        { 8193, new[] { 7175, 8188, 8191, 8249 } },
                        { 8195, new[] { 8197, 21567, 21568, 21713, 21714, 22904 } },
                        { 8197, new[] { 7189, 8187, 8195, 8252, 21567, 21568 } },
                        { 8212, new[] { 311, 7172, 7844, 8725, 8729 } },
                        { 8246, new[] { 7172, 7173, 8191, 8725 } },
                        { 8247, new[] { 7172, 7177, 7839, 8191, 8249 } },
                        { 8249, new[] { 7175, 7838, 8193, 8247, 8250 } },
                        { 8250, new[] { 337, 7189, 7833, 8187, 8188, 8249, 8252 } },
                        { 8252, new[] { 337, 7189, 8187, 8188, 8197, 8250, 21567 } },
                        { 8725, new[] { 311, 7172, 7173, 8212, 8246, 8729 } },
                        { 8729, new[] { 311, 7172, 7173, 7844, 7845, 8212, 8725 } },
                        { 21567, new[] { 337, 7189, 8195, 8197, 8252, 21568, 21713, 22817, 22857 } },
                        { 21568, new[] { 7189, 8195, 8197, 21567, 21713, 21714, 22817, 22876, 22879 } },
                        { 21712, new[] { 7853, 7854, 7874, 22203, 22856, 22859, 22895, 22897 } },
                        { 21713, new[] { 8195, 21567, 21568, 21714, 22817, 22876, 22879, 22901, 22904 } },
                        { 21714, new[] { 8195, 21568, 21713, 22204, 22876, 22901, 22904 } },
                        { 21846, new[] { 22203, 22856, 22886, 22895, 22897 } },
                        { 22200, new[] { 22201, 22202, 22203, 22204, 22857, 22859, 22876, 22879, 22886, 22901, 22911, 22912 } },
                        { 22201, new[] { 22200, 22202, 22204, 22876, 22886, 22911, 22912 } },
                        { 22202, new[] { 22200, 22201, 22203, 22204, 22859, 22886, 22897, 22911, 22912 } },
                        { 22203, new[] { 21712, 21846, 22200, 22202, 22886, 22859, 22897 } },
                        { 22204, new[] { 21714, 22200, 22201, 22202, 22876, 22901, 22911, 22912 } },
                        { 22817, new[] { 7189, 7820, 21567, 21568, 21713, 22857, 22879 } },
                        { 22856, new[] { 7854, 7874, 7877, 21712, 21846, 22895, 22897 } },
                        { 22857, new[] { 7820, 7821, 21567, 22200, 22817, 22859, 22879 } },
                        { 22859, new[] { 21568, 21712, 22200, 22202, 22203, 22857, 22879, 22886, 22895, 22897 } },
                        { 22876, new[] { 21568, 21713, 21714, 22200, 22201, 22204, 22879, 22901, 22904, 22912 } },
                        { 22879, new[] { 7820, 21568, 21713, 22200, 22817, 22857, 22859, 22876, 22897 } },
                        { 22886, new[] { 21846, 22200, 22202, 22203, 22859, 22886, 22895, 22897 } },
                        { 22895, new[] { 21712, 21846, 22856, 22859, 22886, 22897 } },
                        { 22897, new[] { 7854, 21712, 21846, 22202, 22203, 22856, 22859, 22879, 22886, 22895 } },
                        { 22901, new[] { 21713, 21714, 22200, 22204, 22876, 22904, 22911, 22912 } },
                        { 22904, new[] { 8195, 21713, 21714, 22876, 22901, 22912 } },
                        { 22911, new[] { 22200, 22201, 22202, 22204, 22901, 22912 } },
                        { 22912, new[] { 22200, 22201, 22202, 22204, 22876, 22901, 22904, 22911 } },
                    };
                case Side.RIGHT:
                    return new Dictionary<int, int[]>
                    {
                        { 11193, new[] { 11200, 11202, 13218 } },
                        { 11200, new[] { 11193, 11202, 11205, 11211, 11266, 13221, 18557 } },
                        { 11202, new[] { 11193, 11200, 11205, 11211, 13220 } },
                        { 11205, new[] { 11200, 11202, 11211, 11266, 18554 } },
                        { 11209, new[] { 11216, 11222, 11224, 14818, 18529, 18556 } },
                        { 11211, new[] { 11200, 11202, 11205, 11266, 13222, 18533, 18554, 18557 } },
                        { 11215, new[] { 11216, 11222, 11224, 14812 } },
                        { 11216, new[] { 11209, 11215, 11222, 11224, 14816, 18556 } },
                        { 11222, new[] { 11209, 11215, 11216, 11224, 18561 } },
                        { 11224, new[] { 11209, 11215, 11216, 11222, 14815 } },
                        { 11239, new[] { 17880, 18524, 18883, 19396, 19400 } },
                        { 11265, new[] { 17882, 17896, 18501, 18513, 18859, 18921, 18923, 21602 } },
                        { 11266, new[] { 11200, 11205, 11211, 18534, 18557 } },
                        { 13218, new[] { 11193, 13220, 13221, 14812 } },
                        { 13220, new[] { 11202, 13218, 13221, 13222, 14815 } },
                        { 13221, new[] { 11200, 13218, 13220, 13222, 14816, 18558 } },
                        { 13222, new[] { 11211, 13220, 13221, 14818, 18538, 18558 } },
                        { 14812, new[] { 11215, 13218, 14815, 14816 } },
                        { 14815, new[] { 11224, 13220, 14812, 14816, 14818 } },
                        { 14816, new[] { 11216, 13221, 14812, 14815, 14818, 18555 } },
                        { 14818, new[] { 11209, 13222, 14815, 14816, 18532, 18555 } },
                        { 17879, new[] { 17880, 18525, 18883, 18917, 18918, 19396, 19400 } },
                        { 17880, new[] { 11239, 17879, 17884, 18504, 18525, 18917, 19396, 19400 } },
                        { 17882, new[] { 11265, 17884, 18505, 18518, 18864, 18920 } },
                        { 17884, new[] { 17880, 17882, 18506, 18519, 18862, 18918 } },
                        { 17896, new[] { 11265, 18500, 18513, 18858, 18868, 18921, 18923, 21602, 21603, 22819 } },
                        { 18552, new[] { 18524, 18544, 18561 } },
                        { 18555, new[] { 14816, 14818, 18532, 18542, 18556, 18558 } },
                        { 18556, new[] { 11209, 11216, 18529, 18544, 18555, 18561 } },
                        { 18557, new[] { 11200, 11211, 11266, 18533, 18534, 18546, 18554, 18558, 22855 } },
                        { 18558, new[] { 13221, 13222, 18538, 18545, 18555, 18557 } },
                        { 18500, new[] { 17896, 18501, 18513, 18534, 18546, 22819, 22858, 22880 } },
                        { 18501, new[] { 11265, 18500, 18505, 18513, 18533, 18546, 22858 } },
                        { 18504, new[] { 17880, 18506, 18525, 18529, 18544 } },
                        { 18505, new[] { 17882, 18501, 18506, 18518, 18538, 18545 } },
                        { 18506, new[] { 17884, 18504, 18505, 18519, 18532, 18542 } },
                        { 18513, new[] { 11265, 17896, 18500, 18501, 18518, 18546, 18921 } },
                        { 18518, new[] { 17882, 18505, 18513, 18519, 18545, 18920 } },
                        { 18519, new[] { 17884, 18506, 18518, 18525, 18542, 18918 } },
                        { 18524, new[] { 11239, 18525, 18552, 18883, 19400 } },
                        { 18525, new[] { 17879, 17880, 18504, 18519, 18524, 18544, 19400 } },
                        { 18529, new[] { 11209, 18504, 18532, 18544, 18556 } },
                        { 18532, new[] { 14818, 18506, 18529, 18538, 18542, 18555 } },
                        { 18533, new[] { 11211, 18501, 18534, 18538, 18546, 18554, 18557, 22309 } },
                        { 18534, new[] { 11266, 18500, 18533, 18546, 18554, 18557, 22309, 22855, 22896 } },
                        { 18538, new[] { 13222, 18505, 18532, 18533, 18545, 18558 } },
                        { 18542, new[] { 18506, 18519, 18532, 18544, 18545, 18555 } },
                        { 18544, new[] { 18504, 18525, 18529, 18542, 18552, 18556 } },
                        { 18545, new[] { 18505, 18518, 18538, 18542, 18546, 18558 } },
                        { 18546, new[] { 18500, 18501, 18513, 18533, 18534, 18545, 18557 } },
                        { 18554, new[] { 11205, 11211, 18533, 18534, 18557, 22309, 22855, 22894 } },
                        { 18561, new[] { 11222, 18552, 18556 } },
                        { 18858, new[] { 17896, 18859, 18868, 18921, 18923 } },
                        { 18859, new[] { 11265, 18858, 18868, 18864, 18921, 18923 } },
                        { 18862, new[] { 17884, 18864, 18917, 18918 } },
                        { 18864, new[] { 17882, 18859, 18862, 18920 } },
                        { 18866, new[] { 18868, 21602, 21603, 22310, 22311, 22905 } },
                        { 18868, new[] { 17896, 18858, 18866, 18923, 21602, 21603 } },
                        { 18883, new[] { 11239, 17879, 18524, 19396, 19400 } },
                        { 18917, new[] { 17879, 17880, 18862, 19396 } },
                        { 18918, new[] { 17879, 17884, 18519, 18862, 18920 } },
                        { 18920, new[] { 17882, 18518, 18864, 18918, 18921 } },
                        { 18921, new[] { 11265, 17896, 18513, 18858, 18859, 18920, 18923 } },
                        { 18923, new[] { 11265, 17896, 18858, 18859, 18868, 18921, 21602 } },
                        { 19396, new[] { 11239, 17879, 17880, 18883, 18917, 19400 } },
                        { 19400, new[] { 11239, 17879, 17880, 18524, 18525, 18883, 19396 } },
                        { 21602, new[] { 11265, 17896, 18866, 18868, 18923, 21603, 22310, 22819, 22858 } },
                        { 21603, new[] { 17896, 18866, 18868, 21602, 22310, 22311, 22819, 22878, 22880 } },
                        { 22309, new[] { 18533, 18534, 18554, 22794, 22855, 22860, 22894, 22896 } },
                        { 22310, new[] { 18866, 21602, 21603, 22311, 22819, 22878, 22880, 22900, 22905 } },
                        { 22311, new[] { 18866, 21603, 22310, 22795, 22878, 22900, 22905 } },
                        { 22449, new[] { 22794, 22855, 22890, 22894, 22896 } },
                        { 22791, new[] { 22792, 22793, 22794, 22795, 22858, 22860, 22878, 22880, 22890, 22900, 22916, 22917 } },
                        { 22792, new[] { 22791, 22793, 22795, 22878, 22890, 22916, 22917 } },
                        { 22793, new[] { 22791, 22792, 22794, 22795, 22860, 22890, 22896, 22916, 22917 } },
                        { 22794, new[] { 22309, 22449, 22791, 22793, 22890, 22860, 22896 } },
                        { 22795, new[] { 22311, 22791, 22792, 22793, 22878, 22900, 22916, 22917 } },
                        { 22819, new[] { 17896, 18500, 21602, 21603, 22310, 22858, 22880 } },
                        { 22855, new[] { 18534, 18554, 18557, 22309, 22449, 22894, 22896 } },
                        { 22858, new[] { 18500, 18501, 21602, 22791, 22819, 22860, 22880 } },
                        { 22860, new[] { 21603, 22309, 22791, 22793, 22794, 22858, 22880, 22890, 22894, 22896 } },
                        { 22878, new[] { 21603, 22310, 22311, 22791, 22792, 22795, 22880, 22900, 22905, 22917 } },
                        { 22880, new[] { 18500, 21603, 22310, 22791, 22819, 22858, 22860, 22878, 22896 } },
                        { 22890, new[] { 22449, 22791, 22793, 22794, 22860, 22890, 22894, 22896 } },
                        { 22894, new[] { 22309, 22449, 22855, 22860, 22890, 22896 } },
                        { 22896, new[] { 18534, 22309, 22449, 22793, 22794, 22855, 22860, 22880, 22890, 22894 } },
                        { 22900, new[] { 22310, 22311, 22791, 22795, 22878, 22905, 22916, 22917 } },
                        { 22905, new[] { 18866, 22310, 22311, 22878, 22900, 22917 } },
                        { 22916, new[] { 22791, 22792, 22793, 22795, 22900, 22917 } },
                        { 22917, new[] { 22791, 22792, 22793, 22795, 22878, 22900, 22905, 22916 } },
                    };
                default: throw new ArgumentException($"Invalid side {side}");
            }
        }

        /* In vanilla VAM, these lower outer corner verts have multipliers 3, 4 and 5.
         * In BootyMagic, these verts get a multiplier of 1.5f, 2.25f and 3.5f, respectively
         * (by default their value will already be increased somewhat due to the physics gradient).
         */
        internal static Dictionary<int, float> BaseSpringMultiplierVerts(string side)
        {
            switch(side)
            {
                case Side.LEFT:
                    return new Dictionary<int, float>
                    {
                        { 7844, 2.00f },
                        { 8729, 2.00f },
                        { 7173, 1.33f },
                        { 7845, 1.33f },
                        { 8191, 1.33f },
                        { 8193, 1.33f },
                        { 8246, 1.33f },
                        { 311, 2.67f },
                        { 8725, 2.67f },
                        { 8212, 4.00f },
                    };
                case Side.RIGHT:
                    return new Dictionary<int, float>
                    {
                        { 18524, 2.00f },
                        { 19400, 2.00f },
                        { 17880, 1.33f },
                        { 18525, 1.33f },
                        { 18862, 1.33f },
                        { 18864, 1.33f },
                        { 18917, 1.33f },
                        { 11239, 2.67f },
                        { 19396, 2.67f },
                        { 18883, 4.00f },
                    };
                default: throw new ArgumentException($"Invalid side {side}");
            }

        }

        internal static IEnumerable<int> MainNonGradientVerts(string side)
        {
            switch(side)
            {
                case Side.LEFT: return new[] { 274, 277, 294, 281, 296, 7172, 7833, 7849, 7866, 8250, 7852, 22817, 8729, 311, 8725 };
                case Side.RIGHT:
                    return new[] { 11202, 11205, 11222, 11209, 11224, 17879, 18513, 18529, 18546, 18921, 18532, 22819, 19400, 11239, 19396 };
                default: throw new ArgumentException($"Invalid side {side}");
            }
        }

        internal static IEnumerable<int> LowerReferenceGradientVerts(string side)
        {
            switch(side)
            {
                case Side.LEFT: return new[] { 21567, 21712, 22817, 22856 };
                case Side.RIGHT: return new[] { 21602, 22309, 22819, 22855 };
                default: throw new ArgumentException($"Invalid side {side}");
            }
        }

        internal static int[] CenterPositionalVertsSoftPhysicsOff(string side)
        {
            switch(side)
            {
                case Side.LEFT: return new[] { 7860, 7861, 7875, 7878, 7825, 7826 };
                case Side.RIGHT: return new[] { 18505, 18506, 18540, 18541, 18555, 18558 };
                default: throw new ArgumentException($"Invalid side {side}");
            }
        }
    }
}
