using System.Collections.Generic;
using UnityEngine;

namespace ColliderEditor.Models
{
    sealed class AutoColliderModel : ColliderContainerModelBase<AutoCollider>, IModel
    {
        readonly List<ColliderModel> _ownedColliders = new List<ColliderModel>();
        protected override bool OwnsColliders => true;
        public string Type => "Auto Collider";
        internal AutoCollider AutoCollider => component;

        internal AutoColliderModel(AutoCollider autoCollider, ColliderPreviewConfig config, MaterialHelper materialHelper)
            : base(autoCollider, $"[au] {NameHelper.Simplify(autoCollider.name)}")
        {
            if(component.hardCollider)
            {
                _ownedColliders.Add(ColliderModel.CreateTyped(autoCollider.hardCollider, config, materialHelper));
            }

            if(component.jointCollider)
            {
                _ownedColliders.Add(ColliderModel.CreateTyped(component.jointCollider, config, materialHelper));
            }
        }

        internal override IEnumerable<ColliderModel> GetColliders() => _ownedColliders;

        internal IEnumerable<Rigidbody> GetRigidbodies()
        {
            if(component.jointRB)
            {
                yield return component.jointRB;
            }

            if(component.kinematicRB)
            {
                yield return component.kinematicRB;
            }
        }
    }
}
