using NAT.Common.Storables;
using System.Collections.Generic;
using System.Linq;

namespace NAT.Common
{
    class Bindings
    {
        internal Dictionary<string, string> NamespaceDict { get; private set; }
        internal Dictionary<string, JSONStorableAction> Actions { get; }

        internal Bindings(ScriptBase script)
        {
            NamespaceDict = new Dictionary<string, string>
            {
                { "Namespace", script.DomainName },
            };

            var actionsList = new List<StorableAction>
            {
                new StorableAction(Constant.OPEN_UI, script.DeferOpenUI),
                new StorableAction($"{Constant.OPEN_UI}_{Constant.MAIN}", script.DeferOpenUIMain),
                new StorableAction($"{Constant.OPEN_UI}_{Constant.FRICTION}", script.DeferOpenUIFriction),
                new StorableAction($"{Constant.OPEN_UI}_{Constant.PHYSICS}", script.DeferOpenUIPhysics),
                new StorableAction($"{Constant.OPEN_UI}_{Constant.MORPHS}", script.DeferOpenUIMorphs),
                new StorableAction($"{Constant.OPEN_UI}_{Constant.HARD_COLLIDERS}", script.DeferOpenUIHardColliders),
                new StorableAction($"{Constant.OPEN_UI}_{Constant.CLOTHING}", script.DeferOpenUIClothing),
                new StorableAction(Constant.TOGGLE_ENABLED, () => script.enabledJSON.val = !script.enabledJSON.val),
            };
            foreach(var action in actionsList)
            {
                script.PreRegisterUIActionsParam(action);
            }

            actionsList.Add(script.CalibrationAction);

            if(script.IsDevelopmentEnv)
            {
                actionsList.Add(new StorableAction($"{Constant.OPEN_UI}_Dev", script.DeferOpenUIDev));
                actionsList.Add(new StorableAction($"{Constant.OPEN_UI}_DevMorph", script.DeferOpenUIDevMorph));
            }

            Actions = actionsList.ToDictionary(action => action.name, action => (JSONStorableAction) action);
        }

        internal IEnumerable<object> GetActionsList() => Actions.Values.Select(action => (object) action).ToList();
    }
}
