using NAT.Common.Storables;
using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;

namespace NAT.Common.Components
{
    abstract class UIModManagerBase : MonoBehaviour
    {
        protected Transform parentUI;
        protected readonly List<UIMod> uiMods = new List<UIMod>();
        protected bool initialized;
        protected StorableBool skinMaterials2ModifiedJsb;

        internal JSONStorableAction SwitchToOtherPluginAction { get; set; }
        internal UIDynamicSlider DrySkinFrictionSlider { get; set; }

        protected UIMod NewUIMod(
            string uiModId,
            string targetName,
            Func<UIMod, IEnumerator> changesFunc,
            Func<bool> checkPluginEnabled,
            bool useDisabledStyle = false
        )
        {
            var uiMod = gameObject.AddComponent<UIMod>();
            uiMod.Init(uiModId, parentUI, targetName, changesFunc);
            uiMod.CheckPluginEnabled = checkPluginEnabled;
            uiMod.UseDisabledStyle = useDisabledStyle;
            uiMods.Add(uiMod);
            return uiMod;
        }

        internal virtual void SetFrictionUIVisibility()
        {
        }

        protected void ShowFrictionUI(bool frictionHandlerEnabled)
        {
            if(!enabled)
            {
                return;
            }

            if(frictionHandlerEnabled)
            {
                uiMods.Find(uiMod => uiMod.Id == Constant.SKIN_MATERIALS_2_FRICTION).Enable();
            }

            uiMods.Find(uiMod => uiMod.Id == Constant.SKIN_MATERIALS_2_ADD_BTN).Disable();

            skinMaterials2ModifiedJsb.val = true;
        }

        protected void HideFrictionUI(bool frictionHandlerEnabled)
        {
            if(!enabled)
            {
                return;
            }

            uiMods.Find(uiMod => uiMod.Id == Constant.SKIN_MATERIALS_2_FRICTION).Disable();
            if(frictionHandlerEnabled)
            {
                uiMods.Find(uiMod => uiMod.Id == Constant.SKIN_MATERIALS_2_ADD_BTN).Enable();
            }

            skinMaterials2ModifiedJsb.val = false;
        }

        protected void OnDisable()
        {
            foreach(var uiMod in uiMods)
            {
                uiMod.Disable();
            }
        }

        void OnDestroy()
        {
            foreach(var uiMod in uiMods)
            {
                Destroy(uiMod);
            }
        }
    }
}
