// ReSharper disable MemberCanBePrivate.Global
using System;
using System.Diagnostics.CodeAnalysis;
using UnityEngine;

namespace NAT.Common.Components
{
    sealed class UnityEventsListener : MonoBehaviour
    {
        internal bool IsEnabled { get; private set; }
        internal Action EnabledHandlers { get; set; }
        internal Action DisabledHandlers { get; set; }

        [SuppressMessage("ReSharper", "UnusedAutoPropertyAccessor.Global")]
        internal Action DestroyHandlers { get; set; }

        void OnEnable()
        {
            IsEnabled = true;
            EnabledHandlers?.Invoke();
        }

        void OnDisable()
        {
            IsEnabled = false;
            DisabledHandlers?.Invoke();
        }

        void OnDestroy()
        {
            DestroyHandlers?.Invoke();
        }
    }
}
