using System.Collections.Generic;
using Battlehub.RTCommon;
using UnityEngine;
using UnityEngine.Rendering;

namespace NAT.Common.Diagnostics
{
    sealed class DrawService : MonoBehaviour, IGL
    {
        Material _drawMat;
        Color _defaultColor;
        CameraBehaviour _cameraBehaviour;
        internal List<LineV3> Lines { get; private set; }

        internal void Init(Color defaultColor, bool xray = false)
        {
            var shader = Shader.Find(xray ? "Battlehub/RTHandles/VertexColor" : "Hidden/Internal-Colored");
            _drawMat = new Material(shader)
            {
                hideFlags = HideFlags.HideAndDontSave,
            };
            _drawMat.SetInt("_SrcBlend", (int) BlendMode.SrcAlpha);
            _drawMat.SetInt("_DstBlend", (int) BlendMode.OneMinusSrcAlpha);
            _drawMat.SetInt("_Cull", (int) CullMode.Back);
            _drawMat.SetInt("_ZWrite", 0);
            _defaultColor = defaultColor;

            var vrCamera = GetVRCameraTransform();
            if(vrCamera)
            {
                _cameraBehaviour = vrCamera.gameObject.AddComponent<CameraBehaviour>();
                _cameraBehaviour.PostRenderHandlers += DrawLines;
            }

            if(!GLRenderer.Instance)
            {
                new GameObject().AddComponent<GLRenderer>();
            }

            GLRenderer.Instance.Add(this);

            Lines = new List<LineV3>();
            enabled = false;
        }

        public void Draw(int cullingMask)
        {
            DrawLines();
        }

        void DrawLines()
        {
            if(!enabled)
            {
                return;
            }

            GL.PushMatrix();
            _drawMat.SetPass(0);

            GL.Begin(GL.LINES);
            foreach(var line in Lines)
            {
                GL.Color(line.Color ?? _defaultColor);
                GL.Vertex(line.From);
                GL.Vertex(line.To);
            }

            GL.End();
            GL.PopMatrix();
        }

        void OnDestroy()
        {
            if(GLRenderer.Instance)
            {
                GLRenderer.Instance.Remove(this);
            }

            Destroy(_cameraBehaviour);
        }

        static Transform GetVRCameraTransform()
        {
            if(SuperController.singleton.isOpenVR)
            {
                return SuperController.singleton.ViveCenterCamera.transform;
            }

            return SuperController.singleton.isOVR ? SuperController.singleton.OVRCenterCamera.transform : null;
        }
    }
}
