// ReSharper disable MemberCanBePrivate.Global UnusedMember.Global UnusedMethodReturnValue.Global UnusedType.Global
using NAT.Common;
using System;
using UnityEngine;
using UnityEngine.UI;

namespace NAT
{
    static class MVRScriptExtensions
    {
        internal static string GetPackagePath(this MVRScript script)
        {
            string packageId = script.GetPackageId();
            return packageId == null ? "" : $"{packageId}:/";
        }

        //MacGruber / Discord 20.10.2020
        //Get path prefix of the package that contains this plugin
        internal static string GetPackageId(this MVRScript script)
        {
            string id = script.name.Substring(0, script.name.IndexOf('_'));
            string filename = script.manager.GetJSON()["plugins"][id].Value;
            return FileUtils.ParsePackageIdFromPath(filename);
        }

        internal static JSONStorable GetOwnScript(this MVRScript script, string scriptName)
        {
            string prefix = script.name.Substring(0, script.name.IndexOf('_'));
            return script.containingAtom.GetStorableByID($"{prefix}_{Constant.NATURALIS}.{scriptName}");
        }

        internal static Transform InstantiateTextField(this MVRScript script, Transform parent = null) =>
            Instantiate(script.manager.configurableTextFieldPrefab, parent);

        internal static Transform InstantiateButton(this MVRScript script, Transform parent = null) =>
            Instantiate(script.manager.configurableButtonPrefab, parent);

        internal static Transform InstantiateSlider(this MVRScript script, Transform parent = null) =>
            Instantiate(script.manager.configurableSliderPrefab, parent);

        internal static Transform InstantiateToggle(this MVRScript script, Transform parent = null) =>
            Instantiate(script.manager.configurableTogglePrefab, parent);

        internal static Transform InstantiatePopup(this MVRScript script, Transform parent = null) =>
            Instantiate(script.manager.configurablePopupPrefab, parent);

        internal static Transform InstantiateScrollablePopup(this MVRScript script, Transform parent = null) =>
            Instantiate(script.manager.configurableScrollablePopupPrefab, parent);

        internal static Transform InstantiateFilterablePopup(this MVRScript script, Transform parent = null) =>
            Instantiate(script.manager.configurableFilterablePopupPrefab, parent);

        internal static Transform InstantiateColorPicker(this MVRScript script, Transform parent = null) =>
            Instantiate(script.manager.configurableColorPickerPrefab, parent);

        internal static Transform InstantiateSpacer(this MVRScript script, Transform parent = null) =>
            Instantiate(script.manager.configurableSpacerPrefab, parent);

        static Transform Instantiate(Transform prefab, Transform parent = null)
        {
            var transform = UnityEngine.Object.Instantiate(prefab, parent, false);
            UnityEngine.Object.Destroy(transform.GetComponent<LayoutElement>());
            return transform;
        }

        internal static UIDynamic NewSpacer(
            this MVRScript script,
            float height,
            bool rightSide = false
        )
        {
            if(height <= 0)
            {
                return null;
            }

            var spacer = script.CreateSpacer(rightSide);
            spacer.height = height;
            return spacer;
        }

        internal static void RemoveElement(this MVRScript script, UIDynamic element)
        {
            if(element is UIDynamicTextField)
            {
                script.RemoveTextField((UIDynamicTextField) element);
            }
            else if(element is UIDynamicButton)
            {
                script.RemoveButton((UIDynamicButton) element);
            }
            else if(element is UIDynamicSlider)
            {
                script.RemoveSlider((UIDynamicSlider) element);
            }
            else if(element is UIDynamicToggle)
            {
                script.RemoveToggle((UIDynamicToggle) element);
            }
            else if(element is UIDynamicPopup)
            {
                script.RemovePopup((UIDynamicPopup) element);
            }
            else if(element is UIDynamicColorPicker)
            {
                script.RemoveColorPicker((UIDynamicColorPicker) element);
            }
            else
            {
                script.RemoveSpacer(element);
            }
        }
    }
}
