// ReSharper disable MemberCanBePrivate.Global UnusedMember.Global UnusedMethodReturnValue.Global UnusedType.Global
using NAT.Common;
using UnityEngine;
using UnityEngine.Events;

namespace NAT
{
    static class UIDynamicExtensions
    {
        internal static void AddListener(this UIDynamic uiDynamic, UnityAction<bool> callback)
        {
            if(!uiDynamic)
            {
                return;
            }

            var uiDynamicToggle = uiDynamic as UIDynamicToggle;
            if(!uiDynamicToggle)
            {
                Loggr.Error($"UIDynamic {uiDynamic.name} was null or not an UIDynamicToggle");
                return;
            }

            uiDynamicToggle.toggle.onValueChanged.AddListener(callback);
        }

        internal static void AddListener(this UIDynamic uiDynamic, UnityAction callback)
        {
            if(!uiDynamic)
            {
                return;
            }

            var uiDynamicButton = uiDynamic as UIDynamicButton;
            if(!uiDynamicButton)
            {
                Loggr.Error($"UIDynamic {uiDynamic.name} was null or not an UIDynamicButton");
                return;
            }

            uiDynamicButton.button.onClick.AddListener(callback);
        }

        internal static void AddListener(this UIDynamic uiDynamic, UnityAction<float> callback)
        {
            if(!uiDynamic)
            {
                return;
            }

            var uiDynamicSlider = uiDynamic as UIDynamicSlider;
            if(!uiDynamicSlider)
            {
                Loggr.Error($"UIDynamic {uiDynamic.name} was null or not an UIDynamicSlider");
                return;
            }

            uiDynamicSlider.slider.onValueChanged.AddListener(callback);
        }

        internal static void SetActiveStyle(this UIDynamic uiDynamic, bool isActive, bool setInteractable = false, bool highlightIneffective = false)
        {
            if(!uiDynamic)
            {
                return;
            }

            var color = isActive ? Color.black : Colors.inactive;
            if(uiDynamic is UIDynamicSlider)
            {
                var uiDynamicSlider = (UIDynamicSlider) uiDynamic;
                uiDynamicSlider.slider.interactable = !setInteractable || isActive;
                uiDynamicSlider.quickButtonsEnabled = !setInteractable || isActive;
                uiDynamicSlider.defaultButtonEnabled = !setInteractable || isActive;
                uiDynamicSlider.labelText.color = color;
            }
            else if(uiDynamic is UIDynamicToggle)
            {
                var uiDynamicToggle = (UIDynamicToggle) uiDynamic;
                uiDynamicToggle.toggle.interactable = !setInteractable || isActive;
                if(highlightIneffective && uiDynamicToggle.toggle.isOn && uiDynamicToggle.toggle.interactable)
                {
                    color = isActive ? Color.black : Color.red;
                }

                uiDynamicToggle.labelText.color = color;
            }
            else if(uiDynamic is UIDynamicButton)
            {
                var uiDynamicButton = (UIDynamicButton) uiDynamic;
                uiDynamicButton.button.interactable = !setInteractable || isActive;
                var colors = uiDynamicButton.button.colors;
                colors.disabledColor = colors.normalColor;
                uiDynamicButton.button.colors = colors;
                uiDynamicButton.textColor = color;
            }
            else if(uiDynamic is UIDynamicPopup)
            {
                var uiDynamicPopup = (UIDynamicPopup) uiDynamic;
                uiDynamicPopup.SetInteractable(!setInteractable || isActive);
            }
            else
            {
                Loggr.Error($"SetLabel: UIDynamic {uiDynamic.name} was null, or not an expected type");
            }
        }

        internal static void SetLabel(this UIDynamic uiDynamic, string label)
        {
            if(uiDynamic is UIDynamicSlider)
            {
                var uiDynamicSlider = (UIDynamicSlider) uiDynamic;
                uiDynamicSlider.label = label;
            }
            else if(uiDynamic is UIDynamicToggle)
            {
                var uiDynamicToggle = (UIDynamicToggle) uiDynamic;
                uiDynamicToggle.label = label;
            }
            else if(uiDynamic is UIDynamicButton)
            {
                var uiDynamicButton = (UIDynamicButton) uiDynamic;
                uiDynamicButton.label = label;
            }
            else if(uiDynamic is UIDynamicPopup)
            {
                var uiDynamicPopup = (UIDynamicPopup) uiDynamic;
                uiDynamicPopup.label = label;
            }
            else
            {
                Loggr.Error($"SetLabel: UIDynamic {uiDynamic.name} was null, or not an expected type");
            }
        }
    }
}
