using NAT.Common.Models;
using NAT.Common.Handlers.Configs;
using NAT.Common.Storables;
using NAT.Store;
using SimpleJSON;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace NAT.Common.Handlers
{
    abstract class ClothingHandler : Handler
    {
        ClothingStore _store;
        protected JointPhysicsHandler jointPhysicsHandler;
        protected SoftPhysicsHandler softPhysicsHandler;
        protected GravityPhysicsHandler gravityPhysicsHandler;
        HardColliderHandler _hardColliderHandler;
        internal ClothingPresetHandler PresetHandler { get; set; }

        internal string ClothingProfilesDir { get; set; }
        protected string[] validRegionTags;
        protected string[] invalidRegionTags;
        readonly List<string> _ignorePathStarts = new List<string>
        {
            "HUNTING-SUCCUBUS/Enhanced Eyes",
            "cotyounoyume/Tears",
            "cotyounoyume/Drool",
            "BooMoon/Belly Piercing",
            "WeebU/Fancy heels",
            "VL_13/Acc_B_",
            "JaxZoa/Jax_Effects_",
            "VaM_SleSar/PussyDrip",
            "mopedlampe/moped glasses: ",
            "VL_13/Accessories_G_",
            "paledriver.Eyes",
            "VL_13/Eye_Lines",
        };
        readonly List<string> _ignorePackageUidStarts = new List<string>
        {
            "Skippy.Skippys_CumClothing.",
        };
        readonly HashSet<string> _ignorePaths = new HashSet<string>
        {
            "Oeshii/Iris Reflection",
            "kemenate/Tears",
            "RenVR/Eye Reflections (REN)",
            "Jackaroo/Lips!",
            "Jackaroo/Nails!",
            "Jackaroo/Eyes!",
            "HUNTING-SUCCUBUS/Skin Hider Mask",
            "VaM_SleSar/Drip",
            "VaM_SleSar/MouthDrip",
            "VaM_SleSar/NoseDrip",
        };

        protected Rigidbody leftRb, rightRb;

        internal JSONStorableStringChooser ClothingItemJssc { get; private set; }

        internal JSONStorableBool ItemEnabledJsb { get; private set; }
        internal JSONStorableBool LeftActiveJsb { get; private set; }
        internal JSONStorableBool RightActiveJsb { get; private set; }
        internal JSONStorableBool SymmetricJsb { get; private set; }
        internal JSONStorableFloat ProfileTransitionJsf { get; private set; }
        internal JSONStorableFloat YForceJsf { get; private set; }
        internal JSONStorableFloat XForceJsf { get; private set; }
        internal JSONStorableFloat ZForceJsf { get; private set; }
        internal JSONStorableFloat YXHoldPctJsf { get; private set; }
        internal JSONStorableFloat ZHoldPctJsf { get; private set; }
        internal JSONStorableFloat YXDampeningPctJsf { get; private set; }
        internal JSONStorableFloat ZDampeningPctJsf { get; private set; }
        internal JSONStorableFloat YMorphsOffsetJsf { get; private set; }
        internal JSONStorableFloat XMorphsOffsetJsf { get; private set; }
        internal JSONStorableFloat ZMorphsOffsetJsf { get; private set; }
        internal JSONStorableBool ProfileOffWhenUndressingJsb { get; private set; }
        internal JSONStorableFloat SoftJointSearchRadiusJsf { get; private set; }
        internal JSONStorableStringChooser VisualizeClothingCoverageJssc { get; private set; }
        internal JSONStorableBool AutoSelectVisualizedParamJsb { get; private set; }
        internal SoftPhysicsParamGroup VisualizedParamGroup { get; private set; }

        internal Dictionary<string, JSONStorableFloat> SoftPhysicsJsfs { get; private set; }
        internal Dictionary<string, ClothingMorphConfig> MorphConfigs { get; } = new Dictionary<string, ClothingMorphConfig>();

        internal StorableFloat MorphingTransitionJsf { get; private set; }

        internal Vector3 LeftForce { get; private set; }
        internal Vector3 RightForce { get; private set; }

        internal Vector3 LeftMorphsOffsets { get; private set; }
        internal Vector3 RightMorphsOffsets { get; private set; }
        internal Vector3 LeftMorphsOffsetsCompensation { get; private set; }
        internal Vector3 RightMorphsOffsetsCompensation { get; private set; }

        internal float LeftYXHoldPct { get; private set; }
        internal float RightYXHoldPct { get; private set; }
        internal float LeftZHoldPct { get; private set; }
        internal float RightZHoldPct { get; private set; }
        internal float LeftYXDampeningPct { get; private set; }
        internal float RightYXDampeningPct { get; private set; }
        internal float LeftZDampeningPct { get; private set; }
        internal float RightZDampeningPct { get; private set; }

        protected float upDownMassMultiplier;
        protected float sidewaysMassMultiplier;
        protected float inOutMassMultiplier;

        List<ClothingItem> _itemsToAutoLoadProfile = new List<ClothingItem>();
        internal ClothingItem SelectedItem { get; private set; }
        internal ClothingAdjustmentProfile SelectedProfile { get; private set; }
        readonly List<ClothingAdjustmentProfile> _profilesToPreRegister = new List<ClothingAdjustmentProfile>();
        internal HashSet<ClothingAdjustmentProfile> ActiveProfiles { get; } = new HashSet<ClothingAdjustmentProfile>();
        ClothingAdjustmentParams _selectedParams;

        internal override void Init<T>(ProviderBase provider)
        {
            base.Init<T>(provider);
            _store = Script.Bootstrap.ClothingStore;
            _hardColliderHandler = provider.GetInstance<HardColliderHandler>();

            ClothingItemJssc = new JSONStorableStringChooser(
                "Clothing Item",
                new List<string> { Constant.SELECT },
                new List<string> { Constant.SELECT },
                Constant.SELECT,
                "Clothing Item",
                OnItemSelected
            );
            ItemEnabledJsb = new JSONStorableBool("Enabled", false);
            LeftActiveJsb = new JSONStorableBool("Profile Active", false);
            RightActiveJsb = new JSONStorableBool("Profile Active (Right)", false);
            SymmetricJsb = new JSONStorableBool("Symmetric", true);
            ProfileTransitionJsf = new StorableFloat("Profile Transition %", 0f, 0f, 100f);
            YForceJsf = new JSONStorableFloat("Y Force", 0.00f, -5.00f, 5.00f);
            XForceJsf = new JSONStorableFloat("X Force", 0.00f, -5.00f, 5.00f);
            ZForceJsf = new JSONStorableFloat("Z Force", 0.00f, -5.00f, 5.00f);
            YXHoldPctJsf = new JSONStorableFloat("YX Hold %", 0f, 0f, 100f);
            ZHoldPctJsf = new JSONStorableFloat("Z Hold %", 0f, 0f, 100f);
            YXDampeningPctJsf = new JSONStorableFloat("YX Dampening %", 0f, 0f, 100f);
            ZDampeningPctJsf = new JSONStorableFloat("Z Dampening %", 0f, 0f, 100f);
            YMorphsOffsetJsf = new JSONStorableFloat("Y Morphs Angle Offset", 0.0f, -15.0f, 15.0f);
            XMorphsOffsetJsf = new JSONStorableFloat("X Morphs Angle Offset", 0.0f, Script.IsTittyMagic ? -15.0f : -30.0f, Script.IsTittyMagic ? 15.0f : 30.0f);
            ZMorphsOffsetJsf = new JSONStorableFloat("Z Morphs Offset", 0.0f, -15.0f, 15.0f);
            ProfileOffWhenUndressingJsb = new JSONStorableBool("Set Profile Off When Undressing", false);

            ItemEnabledJsb.setCallbackFunction = value => SelectedItem?.EnabledJsb.SetVal(value);
            LeftActiveJsb.setCallbackFunction = value => SelectedProfile?.LeftActive.SetVal(value);
            RightActiveJsb.setCallbackFunction = value => SelectedProfile?.RightActive.SetVal(value);
            SymmetricJsb.setCallbackFunction = value => SelectedProfile?.Symmetric.SetVal(value);
            ProfileTransitionJsf.setCallbackFunction = value =>
            {
                if(SelectedProfile == null)
                {
                    return;
                }

                if(SelectedProfile.HasTarget)
                {
                    SelectedProfile?.Transition.SetVal(value);
                }
            };
            YForceJsf.setCallbackFunction = value => _selectedParams?.YForce.SetVal(value);
            XForceJsf.setCallbackFunction = value => _selectedParams?.XForce.SetVal(value);
            ZForceJsf.setCallbackFunction = value => _selectedParams?.ZForce.SetVal(value);
            YXHoldPctJsf.setCallbackFunction = value => _selectedParams?.YXHoldPct.SetVal(value);
            ZHoldPctJsf.setCallbackFunction = value => _selectedParams?.ZHoldPct.SetVal(value);
            YXDampeningPctJsf.setCallbackFunction = value => _selectedParams?.YXDampeningPct.SetVal(value);
            ZDampeningPctJsf.setCallbackFunction = value => _selectedParams?.ZDampeningPct.SetVal(value);
            YMorphsOffsetJsf.setCallbackFunction = value => _selectedParams?.YMorphsOffset.SetVal(value);
            XMorphsOffsetJsf.setCallbackFunction = value => _selectedParams?.XMorphsOffset.SetVal(value);
            ZMorphsOffsetJsf.setCallbackFunction = value => _selectedParams?.ZMorphsOffset.SetVal(value);
            ProfileOffWhenUndressingJsb.setCallbackFunction = value => SelectedProfile?.OffWhenUndressing.SetVal(value);

            SoftJointSearchRadiusJsf = new JSONStorableFloat("Soft Joint Search Radius", 0.0100f, 0.0000f, Script.IsBootyMagic ? 0.0500f : 0.0250f);
            SoftPhysicsJsfs = new Dictionary<string, JSONStorableFloat>
            {
                [FAT_SPRING.NAME] = new JSONStorableFloat($"{FAT_SPRING.NAME} %", 0f, 0f, 50f),
                [FAT_DAMPER.NAME] = new JSONStorableFloat($"{FAT_DAMPER.NAME} %", 0f, 0f, 20f),
                [FAT_MASS.NAME] = new JSONStorableFloat($"{FAT_MASS.NAME} %", 0f, -100f, 0f),
                [FAT_COLLIDER_RADIUS.NAME] = new JSONStorableFloat($"{FAT_COLLIDER_RADIUS.NAME} %", 0f, -90f, 0f),
                [FAT_COLLIDER_DEPTH.NAME] = new JSONStorableFloat($"{FAT_COLLIDER_DEPTH.NAME} %", 0f, -100f, 100f),
                [FAT_DISTANCE_LIMIT.NAME] = new JSONStorableFloat($"{FAT_DISTANCE_LIMIT.NAME} %", 0f, -90f, 0f),
            };

            var options = new List<string> { "None", Constant.COVERAGE };
            options.AddRange(SoftPhysicsJsfs.Keys);
            VisualizeClothingCoverageJssc = new JSONStorableStringChooser(
                "Visualize Clothing Coverage",
                options,
                options[0],
                "Visualize"
            )
            {
                setCallbackFunction = value => VisualizedParamGroup = softPhysicsHandler.SoftParamGroups.FirstOrDefault(group => group.Name == value),
            };
            AutoSelectVisualizedParamJsb = new JSONStorableBool("Auto-Select Visualized Parameter", true);

            foreach(var pair in SoftPhysicsJsfs)
            {
                pair.Value.setCallbackFunction = value =>
                {
                    if(_selectedParams == null || !(SelectedProfile?.IsCoverageMapped ?? false))
                    {
                        return;
                    }

                    _selectedParams.SoftPhysicsParams[pair.Key].SetVal(value);
                    if(AutoSelectVisualizedParamJsb.val)
                    {
                        VisualizeClothingCoverageJssc.val = pair.Key;
                    }
                };
            }

            MorphingTransitionJsf = new StorableFloat("Morphing Transition (s)", 0.10f, 0.00f, 1.00f);
        }

        protected override void LoadSettings()
        {
            InitializeProfiles();
            UpdateItemsToAutoLoadProfile();
            RebuildClothingOptions();
            InitActiveProfiles();
        }

        void InitializeProfiles()
        {
            foreach(var item in _store.EnabledItems)
            {
                if(!IsValidItem(item))
                {
                    continue;
                }

                InitializeProfile(item);
            }
        }

        void UpdateItemsToAutoLoadProfile()
        {
            foreach(var item in _store.EnabledItems)
            {
                if(!IsValidItem(item))
                {
                    continue;
                }

                // Debug.Log($"UpdateItemsToAutoLoadProfile: {item.DazItem.uid}");
                _itemsToAutoLoadProfile.Add(item);
            }
        }

        protected void AddMorphConfigCallbacks()
        {
            foreach(var pair in MorphConfigs)
            {
                pair.Value.StorableFloat.setCallbackFunction = value => _selectedParams?.MorphParams[pair.Key].SetVal(value);
            }
        }

        void SyncActiveProfiles(ClothingItem item)
        {
            if(item == null)
            {
                return;
            }

            var profile = GetProfile(item);
            if(IsValidItem(item) && item.EnabledJsb.val && profile.IsActive)
            {
                ActiveProfiles.Add(profile);
            }
            else
            {
                ActiveProfiles.Remove(profile);
            }
        }

        internal void SetAllAdjustmentValues()
        {
            foreach(var profile in ActiveProfiles)
            {
                profile.SetAllAdjustmentValues(upDownMassMultiplier, sidewaysMassMultiplier, inOutMassMultiplier, Script.IsBootyMagic);
            }
        }

        protected virtual void SyncJointPhysicsAdjustments()
        {
            LeftYXHoldPct = 0f;
            RightYXHoldPct = 0f;
            LeftZHoldPct = 0f;
            RightZHoldPct = 0f;
            LeftYXDampeningPct = 0f;
            RightYXDampeningPct = 0f;
            LeftZDampeningPct = 0f;
            RightZDampeningPct = 0f;

            foreach(var profile in ActiveProfiles)
            {
                if(profile.LeftActive.Value)
                {
                    LeftYXHoldPct = Mathf.Max(LeftYXHoldPct, profile.GetLeftYXHoldPct());
                    LeftZHoldPct = Mathf.Max(LeftZHoldPct, profile.GetLeftZHoldPct());
                    LeftYXDampeningPct = Mathf.Max(LeftYXDampeningPct, profile.GetLeftYXDampeningPct());
                    LeftZDampeningPct = Mathf.Max(LeftZDampeningPct, profile.GetLeftZDampeningPct());
                }

                if(profile.RightActive.Value)
                {
                    RightYXHoldPct = Mathf.Max(RightYXHoldPct, profile.GetRightYXHoldPct());
                    RightZHoldPct = Mathf.Max(RightZHoldPct, profile.GetRightZHoldPct());
                    RightYXDampeningPct = Mathf.Max(RightYXDampeningPct, profile.GetRightYXDampeningPct());
                    RightZDampeningPct = Mathf.Max(RightZDampeningPct, profile.GetRightZDampeningPct());
                }
            }

            float leftYXHoldAmount = CalculateHoldAmount(LeftYXHoldPct);
            float rightYXHoldAmount = CalculateHoldAmount(RightYXHoldPct);
            float leftZHoldAmount = CalculateHoldAmount(LeftZHoldPct);
            float rightZHoldAmount = CalculateHoldAmount(RightZHoldPct);

            jointPhysicsHandler.RotationSpring.Left.UpdateClothingAdjustmentAmount(leftYXHoldAmount);
            jointPhysicsHandler.RotationSpring.Right.UpdateClothingAdjustmentAmount(rightYXHoldAmount);
            jointPhysicsHandler.RotationDamper.Left.UpdateClothingAdjustmentAmount(0.01f * LeftYXDampeningPct);
            jointPhysicsHandler.RotationDamper.Right.UpdateClothingAdjustmentAmount(0.01f * RightYXDampeningPct);
            jointPhysicsHandler.InOutSpring.Left.UpdateClothingAdjustmentAmount(CalculateHoldAmount(LeftZHoldPct));
            jointPhysicsHandler.InOutSpring.Right.UpdateClothingAdjustmentAmount(CalculateHoldAmount(RightZHoldPct));
            jointPhysicsHandler.InOutDamper.Left.UpdateClothingAdjustmentAmount(0.01f * LeftZDampeningPct);
            jointPhysicsHandler.InOutDamper.Right.UpdateClothingAdjustmentAmount(0.01f * RightZDampeningPct);

            float collidersMassMultiplierLeft = 1 - (0.6f * leftYXHoldAmount + 0.28f * leftZHoldAmount);
            float collidersMassMultiplierRight = 1 - (0.6f * rightYXHoldAmount + 0.28f * rightZHoldAmount);
            _hardColliderHandler.UpdateClothingAdjustmentAmounts(collidersMassMultiplierLeft, collidersMassMultiplierRight);
            _hardColliderHandler.SyncCollidersMass();
        }

        protected static float CalculateHoldAmount(float percentage) =>
            Calc.Exponential1(0.01f * percentage, 0.50f, 3.40f, 1.20f);

        void SyncJointAngleOffsets()
        {
            var leftMin = Vector3.zero;
            var leftMax = Vector3.zero;
            var leftSum = Vector3.zero;
            var rightMin = Vector3.zero;
            var rightMax = Vector3.zero;
            var rightSum = Vector3.zero;

            foreach(var profile in ActiveProfiles)
            {
                if(profile.LeftActive.Value)
                {
                    var left = new Vector3(profile.GetLeftXMorphsOffset(), profile.GetLeftYMorphsOffset(), profile.GetLeftZMorphsOffset());
                    leftMin = Vector3.Min(left, leftMin);
                    leftMax = Vector3.Max(left, leftMax);
                    leftSum += left;
                }

                if(profile.RightActive.Value)
                {
                    var right = new Vector3(-profile.GetRightXMorphsOffset(), profile.GetRightYMorphsOffset(), profile.GetRightZMorphsOffset());
                    rightMin = Vector3.Min(right, rightMin);
                    rightMax = Vector3.Max(right, rightMax);
                    rightSum += right;
                }
            }

            LeftMorphsOffsets = ClampVector(leftSum, leftMin, leftMax);
            RightMorphsOffsets = ClampVector(rightSum, rightMin, rightMax);
        }

        internal void SyncMorphsOffsetsCompensation()
        {
            // must be synced before syncing offsets compensation since the compensation value depends on morph values being up to date
            foreach(var pair in MorphConfigs)
            {
                SyncMorph(pair.Key, pair.Value);
            }

            Vector3 left = Vector3.zero, right = Vector3.zero;

            foreach(var pair in MorphConfigs)
            {
                var config = pair.Value;
                left += config.GetLeftMorphsOffsetsCompensation();
                right += config.GetRightMorphsOffsetsCompensation();
            }

            LeftMorphsOffsetsCompensation = left;
            RightMorphsOffsetsCompensation = right;
        }

        void SyncSoftPhysicsAdjustments()
        {
            if(!softPhysicsHandler.SoftPhysicsOn)
            {
                return;
            }

            foreach(var pair in SoftPhysicsJsfs)
            {
                SyncSoftPhysicsParam(pair.Key);
            }

            SetUseDefaultSoftColliderFriction(Side.LEFT);
            SetUseDefaultSoftColliderFriction(Side.RIGHT);
        }

        void SyncSoftPhysicsParam(string paramName)
        {
            SetParamClothingAdjustment(Side.LEFT, paramName);
            SetParamClothingAdjustment(Side.RIGHT, paramName);
            CalculateAverageClothingAdjustments(paramName);
        }

        void SetParamClothingAdjustment(string side, string paramName)
        {
            foreach(var softJoint in softPhysicsHandler.AllSoftJoints[side])
            {
                float min = 0f;
                float max = 0f;
                float sum = 0f;
                foreach(var profile in ActiveProfiles)
                {
                    if(profile.ActiveStorableBools[side].Value && profile.CoveredSoftJoints[side].Contains(softJoint))
                    {
                        float current = profile.GetSoftPhysicsParamValue(side, paramName);
                        min = Mathf.Min(min, current);
                        max = Mathf.Max(max, current);
                        sum += current;
                    }
                }

                float adjustmentPct = sum > 0 ? Mathf.Min(sum, max) : Mathf.Max(sum, min);
                softJoint.SetClothingAdjustment(paramName, 0.01f * adjustmentPct);
            }
        }

        void CalculateAverageClothingAdjustments(string paramName)
        {
            var group = softPhysicsHandler.SoftParamGroups.First(x => x.Name == paramName);
            group.LeftClothingAdjustmentAverage = CalculateAverageClothingAdjustment(Side.LEFT, paramName);
            group.RightClothingAdjustmentAverage = CalculateAverageClothingAdjustment(Side.RIGHT, paramName);
        }

        float CalculateAverageClothingAdjustment(string side, string paramName)
        {
            var set = new HashSet<SoftJoint>(ActiveProfiles.SelectMany(profile => profile.CoveredSoftJoints[side]));
            return set.Select(softJoint => softJoint.GetClothingAdjustmentValue(paramName))
                .DefaultIfEmpty(0)
                .Average();
        }

        void SetUseDefaultSoftColliderFriction(string side)
        {
            var coveredJoints = ActiveProfiles
                .Where(profile => profile.ActiveStorableBools[side].Value)
                .SelectMany(profile => profile.CoveredSoftJoints[side])
                .ToList();

            foreach(var softJoint in coveredJoints)
            {
                softJoint.UseDefaultFriction = true;
            }

            var exposedJoints = softPhysicsHandler.AllSoftJoints[side].Except(coveredJoints);
            foreach(var softJoint in exposedJoints)
            {
                softJoint.UseDefaultFriction = false;
            }
        }

        internal void AutoLoadDefaultProfiles()
        {
            if(_itemsToAutoLoadProfile == null)
            {
                return;
            }

            foreach(var item in _itemsToAutoLoadProfile)
            {
                // Debug.Log($"AutoLoadDefaultProfiles: Loading default profile for {item.Label}");
                PresetHandler.UpdatePresetDir(item.PresetSubDir);
                PresetHandler.LoadDefaultPreset(item, true);
            }

            _itemsToAutoLoadProfile = null;
        }

        internal void AutoLoadProfiles(IEnumerable<JSONNode> activeProfiles, string packageId)
        {
            if(Script.FirstCalibrationDone)
            {
                UpdateItemsToAutoLoadProfile();
            }

            foreach(var jsonNode in activeProfiles)
            {
                string path = ParsePresetPath(jsonNode, packageId);
                if(!string.IsNullOrEmpty(path) && FileUtils.FileExists(path))
                {
                    PresetHandler.LoadEnabledItemProfileFromPath(path);
                }
            }

            AutoLoadDefaultProfiles();
        }

        internal void ResetAll()
        {
            ResetMorphs();
            foreach(var param in jointPhysicsHandler.ParamGroups)
            {
                param.ResetClothingAdjustmentAmount();
            }
        }

        internal void OnItemToggled(ClothingItem item)
        {
            /* Prevent null exception when exiting scene */
            if(Script)
            {
                Script.StartCoroutine(OnItemToggledCo(item));
            }
        }

        internal void OnItemAllowDetachToggled(ClothingItem item, bool value)
        {
            if(!item.SimControl || !item.SimControl.simEnabledJSON.val)
            {
                return;
            }

            var profile = GetProfile(item);
            if(profile == null)
            {
                return;
            }

            if(profile.OffWhenUndressing.Value && value)
            {
                profile.ActiveStorableBools[Side.LEFT].SetVal(false);
                if(!profile.Symmetric.Value)
                {
                    profile.ActiveStorableBools[Side.RIGHT].SetVal(false);
                }
            }
        }

        IEnumerator OnItemToggledCo(ClothingItem item)
        {
            if(item.EnabledJsb.val)
            {
                if(IsValidItem(item))
                {
                    RebuildClothingOptions();
                    yield return WaitForClothingItemToLoad(item);
                    var profile = GetProfile(item);
                    if(profile == null)
                    {
                        InitializeProfile(item);
                    }
                    else
                    {
                        SetProfile(item, profile);
                    }

                    ClothingItemJssc.val = item.DazItem.uid;
                }
            }
            else
            {
                bool isClothingTabActive = Script.Tabs?.IsClothingTabActive() ?? false;
                if(!isClothingTabActive && ClothingItemJssc.val == item.DazItem.uid)
                {
                    SelectFirstActiveItem();
                }
            }

            SyncActiveProfiles(item);
            SyncAdjustments();
            Script.SyncMorphsAngleOffsets(MorphingTransitionJsf.val);

            if(item == SelectedItem)
            {
                ItemEnabledJsb.valNoCallback = item.EnabledJsb.val;
            }
        }

        // TODO simplify
        IEnumerator WaitForClothingItemToLoad(ClothingItem item)
        {
            yield return null;
            float timePassed = 0f;
            var thresholds = new Dictionary<float, bool>
            {
                { 30, false },
                { 60, false },
                { 90, false },
                { 120, false },
                { 150, false },
            };

            var wraps = item.DazItem.GetComponentsInChildren<DAZSkinWrap>();
            while(wraps.Length == 0)
            {
                yield return new WaitForSeconds(0.1f);
                timePassed += 0.1f;
                wraps = item.DazItem.GetComponentsInChildren<DAZSkinWrap>();

                foreach(var pair in thresholds)
                {
                    if(timePassed > pair.Key && !pair.Value)
                    {
                        Script.Loggr.LogMessage($"Item {item.Label} DAZSkinWrap(s) not found in {pair.Key:0} sec; still waiting before initializing profile");
                        thresholds[pair.Key] = true;
                    }
                }

                if(timePassed > 180)
                {
                    Script.Loggr.LogError($"Item {item.Label} DAZSkinWrap(s) not found in 180 sec; initializing profile anyway");
                    yield break;
                }
            }
        }

        internal void AutoSelectItem()
        {
            if(ClothingItemJssc.val == Constant.SELECT)
            {
                SelectFirstActiveItem();
            }
        }

        void SelectFirstActiveItem()
        {
            var firstActiveItem = _store.EnabledItems.FirstOrDefault(item => IsValidItem(item) && GetProfile(item).IsActive);
            ClothingItemJssc.val = firstActiveItem?.DazItem.uid ?? Constant.SELECT;
        }

        bool IsValidItem(ClothingItem item) =>
            CheckTags(item.DazItem.tagsArray) &&
            !_ignorePaths.Contains(item.PresetSubDir) &&
            CheckPresetSubDir(item) &&
            CheckPackageUid(item);

        bool CheckTags(string[] tags)
        {
            if(tags == null)
            {
                return true;
            }

            return tags.Any(tag => validRegionTags.Contains(tag)) ||
                !tags.Any(tag => invalidRegionTags.Contains(tag));
        }

        bool CheckPresetSubDir(ClothingItem item)
        {
            foreach(string x in _ignorePathStarts)
            {
                if(item.PresetSubDir.StartsWith(x))
                {
                    return false;
                }
            }

            return true;
        }

        bool CheckPackageUid(ClothingItem item)
        {
            foreach(string x in _ignorePackageUidStarts)
            {
                if(item.DazItem.packageUid?.StartsWith(x) ?? false)
                {
                    return false;
                }
            }

            return true;
        }

        internal void SyncAdjustments()
        {
            SyncJointPhysicsAdjustments();
            SyncJointAngleOffsets();
            SyncMorphsOffsetsCompensation();
            SyncSoftPhysicsAdjustments();
        }

        void InitializeProfile(ClothingItem item)
        {
            // Debug.Log($"{Script.DomainName}: InitializeProfile for {item.Label}");
            if(GetProfile(item) != null)
            {
                return;
            }

            var profile = CreateNewProfile(item);
            SetProfile(item, profile);
            PresetHandler.UpdatePresetDir(item.PresetSubDir);
            if(initialized) // during initialization, default presets are loaded separately with AutoLoadDefaultProfiles
            {
                PresetHandler.LoadDefaultPreset(item, true);
            }

            profile.SaveValues();
        }

        ClothingAdjustmentProfile CreateNewProfile(ClothingItem item)
        {
            // Debug.Log($"{Script.DomainName}: CreateNewProfile for {item.Label}");
            string prefix = $"{item.Label}:";
            var profile = new ClothingAdjustmentProfile(
                item,
                prefix,
                new StorableBool($"{prefix}{SymmetricJsb.name}", true)
                {
                    Temporary = true,
                },
                new StorableBool($"{prefix}{LeftActiveJsb.name}", true)
                {
                    Temporary = true,
                },
                new StorableBool($"{prefix}{RightActiveJsb.name}", true)
                {
                    Temporary = true,
                },
                StorableFloat.Copy(ProfileTransitionJsf, prefix, "", true),
                new StorableBool($"{prefix}{ProfileOffWhenUndressingJsb.name}", false)
                {
                    Temporary = true,
                }
            );

            if(!Person.IsFemale)
            {
                profile.IsCoverageMapped = true;
                profile.ExposedSoftJoints[Side.LEFT] = softPhysicsHandler.AllSoftJoints[Side.LEFT].ToList();
                profile.ExposedSoftJoints[Side.RIGHT] = softPhysicsHandler.AllSoftJoints[Side.RIGHT].ToList();
            }

            profile.Symmetric.StorableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                if(value)
                {
                    RightActiveJsb.valNoCallback = LeftActiveJsb.val;
                }

                profile.SyncSymmetric(value);
                if(profile == SelectedProfile)
                {
                    SelectLeft();
                }

                SyncActiveProfiles(item);
                SyncAdjustments();
                Script.SyncMorphsAngleOffsets(MorphingTransitionJsf.val);
            };
            profile.LeftActive.StorableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                if(profile.Symmetric.Value)
                {
                    profile.RightActive.SetValNoCallback(value);
                }

                if(profile == SelectedProfile)
                {
                    LeftActiveJsb.valNoCallback = value;
                    if(profile.Symmetric.Value)
                    {
                        RightActiveJsb.valNoCallback = value;
                    }
                }

                SyncActiveProfiles(item);
                SyncAdjustments();
                Script.SyncMorphsAngleOffsets(MorphingTransitionJsf.val);
            };
            profile.RightActive.StorableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                if(profile.Symmetric.Value)
                {
                    profile.LeftActive.SetValNoCallback(value);
                }

                if(profile == SelectedProfile)
                {
                    RightActiveJsb.valNoCallback = value;
                    if(profile.Symmetric.Value)
                    {
                        LeftActiveJsb.valNoCallback = value;
                    }
                }

                SyncActiveProfiles(item);
                SyncAdjustments();
                Script.SyncMorphsAngleOffsets(MorphingTransitionJsf.val);
            };
            profile.Transition.StorableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                profile.SetTransitionValue();
                if(profile == SelectedProfile)
                {
                    ProfileTransitionJsf.valNoCallback = value;
                }

                SyncAdjustments();
                Script.SyncMorphsAngleOffsets();
            };
            profile.OffWhenUndressing.StorableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                if(profile == SelectedProfile)
                {
                    ProfileOffWhenUndressingJsb.valNoCallback = value;
                }
            };

            profile.Params[Side.LEFT] = NewClothingAdjustmentParams(prefix, false);
            profile.RightCachedParams = NewClothingAdjustmentParams(prefix, true);
            profile.Params[Side.RIGHT] = profile.Params[Side.LEFT];

            if(Script.Initialized != true)
            {
                _profilesToPreRegister.Add(profile);
            }
            else
            {
                RegisterProfileParams(profile);
            }

            return profile;
        }

        ClothingAdjustmentParams NewClothingAdjustmentParams(string prefix, bool right, bool setCallbacks = true)
        {
            string suffix = right ? " (Right)" : "";
            var @params = new ClothingAdjustmentParams
            (
                YForceJsf,
                XForceJsf,
                ZForceJsf,
                YXHoldPctJsf,
                ZHoldPctJsf,
                YXDampeningPctJsf,
                ZDampeningPctJsf,
                YMorphsOffsetJsf,
                XMorphsOffsetJsf,
                ZMorphsOffsetJsf,
                SoftPhysicsJsfs,
                MorphConfigs,
                prefix,
                suffix
            );
            if(setCallbacks)
            {
                SetParamsCallbackFunctions(@params);
            }

            return @params;
        }

        void SetParamsCallbackFunctions(ClothingAdjustmentParams @params)
        {
            @params.YForce.StorableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                @params.SetYForce(value, upDownMassMultiplier);
                if(@params == _selectedParams)
                {
                    YForceJsf.valNoCallback = value;
                }
            };
            @params.XForce.StorableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                @params.SetXForce(value, sidewaysMassMultiplier);
                if(@params == _selectedParams)
                {
                    XForceJsf.valNoCallback = value;
                }
            };
            @params.ZForce.StorableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                @params.SetZForce(value, inOutMassMultiplier, !Script.IsBootyMagic);
                if(@params == _selectedParams)
                {
                    ZForceJsf.valNoCallback = value;
                }
            };
            @params.YXHoldPct.StorableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                ClothingAdjustmentParams.SetPercentageValue(@params.YXHoldPct);
                if(@params == _selectedParams)
                {
                    YXHoldPctJsf.valNoCallback = value;
                }

                SyncJointPhysicsAdjustments();
            };
            @params.ZHoldPct.StorableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                ClothingAdjustmentParams.SetPercentageValue(@params.ZHoldPct);
                if(@params == _selectedParams)
                {
                    ZHoldPctJsf.valNoCallback = value;
                }

                SyncJointPhysicsAdjustments();
            };
            @params.YXDampeningPct.StorableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                ClothingAdjustmentParams.SetPercentageValue(@params.YXDampeningPct);
                if(@params == _selectedParams)
                {
                    YXDampeningPctJsf.valNoCallback = value;
                }

                SyncJointPhysicsAdjustments();
            };
            @params.ZDampeningPct.StorableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                ClothingAdjustmentParams.SetPercentageValue(@params.ZDampeningPct);
                if(@params == _selectedParams)
                {
                    ZDampeningPctJsf.valNoCallback = value;
                }

                SyncJointPhysicsAdjustments();
            };
            @params.YMorphsOffset.StorableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                @params.SetYMorphsOffset(value, Script.IsBootyMagic);
                if(@params == _selectedParams)
                {
                    YMorphsOffsetJsf.valNoCallback = value;
                }

                SyncJointAngleOffsets();
                Script.SyncMorphsAngleOffsets();
            };
            @params.XMorphsOffset.StorableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                @params.SetXMorphsOffset(value, Script.IsBootyMagic);
                if(@params == _selectedParams)
                {
                    XMorphsOffsetJsf.valNoCallback = value;
                }

                SyncJointAngleOffsets();
                Script.SyncMorphsAngleOffsets();
            };
            @params.ZMorphsOffset.StorableParam.setCallbackFunction = value =>
            {
                if(Script.IsInsideRestore)
                {
                    return;
                }

                @params.SetZMorphsOffset(value, Script.IsBootyMagic);
                if(@params == _selectedParams)
                {
                    ZMorphsOffsetJsf.valNoCallback = value;
                }

                SyncJointAngleOffsets();
                Script.SyncMorphsAngleOffsets();
            };

            foreach(var pair in @params.SoftPhysicsParams)
            {
                pair.Value.StorableParam.setCallbackFunction = value =>
                {
                    if(Script.IsInsideRestore)
                    {
                        return;
                    }

                    ClothingAdjustmentParams.SetPercentageValue(pair.Value);
                    if(@params == _selectedParams)
                    {
                        SoftPhysicsJsfs[pair.Key].valNoCallback = value;
                    }

                    SyncSoftPhysicsParam(pair.Key);
                };
            }

            foreach(var pair in @params.MorphParams)
            {
                pair.Value.StorableParam.setCallbackFunction += value =>
                {
                    if(Script.IsInsideRestore)
                    {
                        return;
                    }

                    if(@params == _selectedParams)
                    {
                        MorphConfigs[pair.Key].StorableFloat.valNoCallback = value;
                    }

                    SyncMorphsOffsetsCompensation(); // sync all morphs at once to ensure correct morphs offsets compensation
                    Script.SyncMorphsAngleOffsets();
                };
            }
        }

        internal void OnItemsReloaded()
        {
            InitializeProfiles();
            RebuildClothingOptions(true);
            InitActiveProfiles();
            ClothingItemJssc.val = Constant.SELECT;
        }

        void RebuildClothingOptions(bool removeDisabled = false)
        {
            ClothingItemJssc.choices.Remove(Constant.SELECT);
            ClothingItemJssc.displayChoices.Remove(Constant.SELECT);

            var choices = removeDisabled ? new List<string>() : new List<string>(ClothingItemJssc.choices);
            var displayChoices = removeDisabled ? new List<string>() : new List<string>(ClothingItemJssc.displayChoices);

            foreach(var item in _store.EnabledItems)
            {
                if(!IsValidItem(item))
                {
                    continue;
                }

                string uid = item.DazItem.uid;
                string label = item.Label;
                if(!choices.Contains(uid))
                {
                    choices.Add(uid);
                    displayChoices.Add(label);
                }
            }

            ClothingItemJssc.choices = choices;
            ClothingItemJssc.displayChoices = displayChoices;
        }

        void InitActiveProfiles()
        {
            ActiveProfiles.Clear();
            foreach(var item in _store.EnabledItems)
            {
                if(!IsValidItem(item))
                {
                    continue;
                }

                var profile = GetProfile(item);
                if(profile.IsActive)
                {
                    ActiveProfiles.Add(profile);
                }
            }
        }

        internal void SelectLeft()
        {
            if(SelectedProfile == null)
            {
                return;
            }

            SyncSelectedProfile(Side.LEFT);
        }

        internal void SelectRight()
        {
            if(SelectedProfile == null)
            {
                return;
            }

            SyncSelectedProfile(Side.RIGHT);
        }

        void OnItemSelected(string option)
        {
            if(option == Constant.SELECT)
            {
                ItemEnabledJsb.SetDefaultValNoCallback();
                LeftActiveJsb.SetDefaultValNoCallback();
                RightActiveJsb.SetDefaultValNoCallback();
                SymmetricJsb.SetDefaultValNoCallback();
                ProfileTransitionJsf.SetDefaultValNoCallback();
                YForceJsf.SetDefaultValNoCallback();
                XForceJsf.SetDefaultValNoCallback();
                ZForceJsf.SetDefaultValNoCallback();
                YXHoldPctJsf.SetDefaultValNoCallback();
                ZHoldPctJsf.SetDefaultValNoCallback();
                YXDampeningPctJsf.SetDefaultValNoCallback();
                ZDampeningPctJsf.SetDefaultValNoCallback();
                YMorphsOffsetJsf.SetDefaultValNoCallback();
                XMorphsOffsetJsf.SetDefaultValNoCallback();
                ZMorphsOffsetJsf.SetDefaultValNoCallback();

                foreach(var pair in SoftPhysicsJsfs)
                {
                    pair.Value.SetDefaultValNoCallback();
                }

                foreach(var pair in MorphConfigs)
                {
                    pair.Value.StorableFloat.SetDefaultValNoCallback();
                }

                VisualizeClothingCoverageJssc.val = "None";
                SelectedItem = null;
                SelectedProfile = null;
                _selectedParams = null;
                Script.Tabs.ClothingWindow.Refresh();
                return;
            }

            // Sanity checks
            if(!_store.Items.ContainsKey(option))
            {
                Script.Loggr.LogError($"Clothing item with key {option} not found");
                return;
            }

            SelectedItem = _store.Items[option];
            if(SelectedItem == null)
            {
                Script.Loggr.LogError("Selected item is null");
                return;
            }

            if(SelectedItem.EnabledJsb == null)
            {
                Script.Loggr.LogError($"Selected item {SelectedItem.Label} enabled toggle is null");
                return;
            }

            ItemEnabledJsb.valNoCallback = SelectedItem.EnabledJsb.val;
            SelectedProfile = GetProfile(SelectedItem);

            if(SelectedProfile == null)
            {
                Script.Loggr.LogError($"Selected item {SelectedItem.Label} profile is null");
                return;
            }

            SyncSelectedProfile(SelectedProfile.IsRightSelected ? Side.RIGHT : Side.LEFT);
        }

        internal void FindCoveredSoftJoints()
        {
            if(SelectedItem == null)
            {
                Script.Loggr.LogMessage("Select a clothing item first.");
                return;
            }

            if(!SelectedItem.EnabledJsb.val)
            {
                Script.Loggr.LogMessage("Item must be enabled in order to map its skin coverage.");
                return;
            }

            softPhysicsHandler.ResetSoftJointsClothingAdjustments();

            var leftJoints = new List<SoftJoint>();
            var rightJoints = new List<SoftJoint>();

            var wrap = SelectedItem.DazItem.GetComponentInChildren<DAZSkinWrap>();
            wrap.UpdateVerts();
            var vertices = wrap.Mesh.vertices;

            foreach(var vertex in vertices)
            {
                leftJoints.AddRange(NearbyJoints(vertex, Side.LEFT));
                rightJoints.AddRange(NearbyJoints(vertex, Side.RIGHT));
            }

            AssignJoints(leftJoints.Distinct().ToList(), SelectedProfile, Side.LEFT);
            AssignJoints(rightJoints.Distinct().ToList(), SelectedProfile, Side.RIGHT);
            SelectedProfile.IsCoverageMapped = true;

            VisualizeClothingCoverageJssc.val = Constant.COVERAGE;

            if(!SelectedProfile.IsCoverageMapped)
            {
                foreach(var pair in SoftPhysicsJsfs)
                {
                    pair.Value.valNoCallback = 0f;
                }
            }

            SyncSoftPhysicsAdjustments();
            Script.Tabs.ClothingWindow.Refresh();
        }

        IEnumerable<SoftJoint> NearbyJoints(Vector3 vertex, string side)
        {
            foreach(var softJoint in softPhysicsHandler.AllSoftJoints[side])
            {
                var position = softJoint.JointRb.position;
                float radius = SoftJointSearchRadiusJsf.val;
                float distance = Vector3.Distance(position, vertex);
                if(distance < radius)
                {
                    yield return softJoint;
                }
            }
        }

        void AssignJoints(
            List<SoftJoint> coveredJoints,
            ClothingAdjustmentProfile profile,
            string side
        )
        {
            profile.CoveredSoftJoints[side] = coveredJoints;
            profile.ExposedSoftJoints[side] = softPhysicsHandler.AllSoftJoints[side].Except(coveredJoints).ToList();
        }

        internal void SetAllJointsCovered(string side)
        {
            SelectedProfile.CoveredSoftJoints[side] = softPhysicsHandler.AllSoftJoints[side].ToList();
            SelectedProfile.ExposedSoftJoints[side].Clear();
            VisualizeClothingCoverageJssc.val = Constant.COVERAGE;
        }

        internal void SetAllJointsExposed(string side)
        {
            SelectedProfile.CoveredSoftJoints[side].Clear();
            SelectedProfile.ExposedSoftJoints[side] = softPhysicsHandler.AllSoftJoints[side].ToList();
            VisualizeClothingCoverageJssc.val = Constant.COVERAGE;
        }

        internal virtual void MirrorCoverageJoints(string fromSide, string toSide)
        {
        }

        const float HOVER_RADIUS = 80f;
        bool _jointToggled;
        float _prevDist;
        string _hoveredOverSide;

        internal void HandleToggleCoverageJoint()
        {
            /* Prevent toggling again while key is down */
            if(_jointToggled)
            {
                if(!Input.GetKey(KeyCode.LeftControl))
                {
                    _jointToggled = false;
                }

                return;
            }

            SelectedProfile.HoveredOverJoint = null;
            _hoveredOverSide = null;
            _prevDist = HOVER_RADIUS;
            HighlightHoveredOverJoint(Side.LEFT, Script.Bootstrap.Camera);
            HighlightHoveredOverJoint(Side.RIGHT, Script.Bootstrap.Camera);

            if(SelectedProfile.HoveredOverJoint != null && Input.GetKey(KeyCode.LeftControl))
            {
                _jointToggled = true;
                ToggleSoftJointSelection(_hoveredOverSide, SelectedProfile.HoveredOverJoint);
                SyncSoftPhysicsAdjustments();
            }
        }

        void HighlightHoveredOverJoint(string side, Camera camera)
        {
            var softJoints = softPhysicsHandler.AllSoftJoints[side];
            foreach(var softJoint in softJoints)
            {
                var targetPos = camera.WorldToScreenPoint(softJoint.JointRb.position);
                float dist = Vector2.Distance(Input.mousePosition, targetPos);
                if(dist < HOVER_RADIUS && dist < _prevDist)
                {
                    SelectedProfile.HoveredOverJoint = softJoint;
                    _hoveredOverSide = side;
                    _prevDist = dist;
                }
            }
        }

        void ToggleSoftJointSelection(string side, SoftJoint softJoint)
        {
            if(SelectedProfile.CoveredSoftJoints[side].Contains(softJoint))
            {
                SelectedProfile.CoveredSoftJoints[side].Remove(softJoint);
                if(!SelectedProfile.ExposedSoftJoints[side].Contains(softJoint))
                {
                    SelectedProfile.ExposedSoftJoints[side].Add(softJoint);
                }
            }
            else if(SelectedProfile.ExposedSoftJoints[side].Contains(softJoint))
            {
                SelectedProfile.ExposedSoftJoints[side].Remove(softJoint);
                if(!SelectedProfile.CoveredSoftJoints[side].Contains(softJoint))
                {
                    SelectedProfile.CoveredSoftJoints[side].Add(softJoint);
                }
            }
        }

        internal void SaveProfile()
        {
            if(SelectedItem == null)
            {
                return;
            }

            PresetHandler.UpdatePresetDir(SelectedItem.PresetSubDir);
            PresetHandler.SavePresetJsu.FileBrowse();
        }

        internal void LoadProfile(bool isSettingTarget = false)
        {
            if(SelectedItem == null)
            {
                return;
            }

            PresetHandler.UpdatePresetDir(SelectedItem.PresetSubDir);
            PresetHandler.IsSettingTarget = isSettingTarget;
            PresetHandler.LoadPresetJsu.FileBrowse();
        }

        internal void SaveDefault()
        {
            if(SelectedItem == null)
            {
                return;
            }

            PresetHandler.UpdatePresetDir(SelectedItem.PresetSubDir);
            PresetHandler.SaveDefaultPreset();
        }

        internal void LoadDefault()
        {
            if(SelectedItem == null)
            {
                return;
            }

            PresetHandler.UpdatePresetDir(SelectedItem.PresetSubDir);
            PresetHandler.LoadDefaultPreset(SelectedItem);
        }

        internal void ResetProfile()
        {
            if(SelectedItem == null)
            {
                return;
            }

            var profile = GetProfile(SelectedItem);
            profile.Reset(Person.IsFemale);
            SyncActiveProfiles(SelectedItem);
            profile.SetAllAdjustmentValues(upDownMassMultiplier, sidewaysMassMultiplier, inOutMassMultiplier, Script.IsBootyMagic);
            SyncAdjustments();
            Script.SyncMorphsAngleOffsets();
            SyncSelectedProfile(Side.LEFT);
        }

        internal void ResetTargetProfile()
        {
            if(SelectedItem == null || SelectedProfile == null)
            {
                ProfileTransitionJsf.val = 0f;
                return;
            }

            var profile = GetProfile(SelectedItem);
            profile.Transition.StorableParam.val = 0f;
            profile.ResetTarget();
        }

        internal void SetProfileOffAsTarget(ClothingAdjustmentProfile profile)
        {
            if(profile == null)
            {
                return;
            }

            var offParams = NewClothingAdjustmentParams(profile.ParamNamePrefix, false, false);
            profile.SetProfileOffAsTarget(offParams);
        }

        internal void SetCurrentTransitionStateAsCurrentProfile()
        {
            if(SelectedItem == null || SelectedProfile == null)
            {
                return;
            }

            var profile = GetProfile(SelectedItem);
            if(!profile.HasTarget || profile.Transition.StorableParam.val == 0f)
            {
                return;
            }

            profile.SetValuesFromCurrentTransitionState();
            profile.SetAllAdjustmentValues(upDownMassMultiplier, sidewaysMassMultiplier, inOutMassMultiplier, Script.IsBootyMagic);
            SyncAdjustments();
            Script.SyncMorphsAngleOffsets();
            SyncSelectedProfile(Side.LEFT);
        }

        protected abstract ClothingAdjustmentProfile GetProfile(ClothingItem item);
        protected abstract void SetProfile(ClothingItem item, ClothingAdjustmentProfile profile);

        void SyncSelectedProfile(string side)
        {
            bool isRight = side == Side.RIGHT;
            if(isRight && SelectedProfile.Symmetric.Value)
            {
                Script.Loggr.LogError("SyncSelectedProfile: Cannot sync right params to storables if profile is symmetric");
                return;
            }

            SymmetricJsb.valNoCallback = SelectedProfile.Symmetric.Value;
            LeftActiveJsb.valNoCallback = SelectedProfile.LeftActive.Value;
            RightActiveJsb.valNoCallback = SelectedProfile.RightActive.Value;
            ProfileTransitionJsf.valNoCallback = SelectedProfile.Transition.StorableParam.val;
            _selectedParams = SelectedProfile.Params[side];
            YForceJsf.valNoCallback = _selectedParams.YForce.StorableParam.val;
            XForceJsf.valNoCallback = _selectedParams.XForce.StorableParam.val;
            ZForceJsf.valNoCallback = _selectedParams.ZForce.StorableParam.val;
            YXHoldPctJsf.valNoCallback = _selectedParams.YXHoldPct.StorableParam.val;
            ZHoldPctJsf.valNoCallback = _selectedParams.ZHoldPct.StorableParam.val;
            YXDampeningPctJsf.valNoCallback = _selectedParams.YXDampeningPct.StorableParam.val;
            ZDampeningPctJsf.valNoCallback = _selectedParams.ZDampeningPct.StorableParam.val;
            YMorphsOffsetJsf.valNoCallback = _selectedParams.YMorphsOffset.StorableParam.val;
            XMorphsOffsetJsf.valNoCallback = _selectedParams.XMorphsOffset.StorableParam.val;
            ZMorphsOffsetJsf.valNoCallback = _selectedParams.ZMorphsOffset.StorableParam.val;

            foreach(var pair in SoftPhysicsJsfs)
            {
                pair.Value.valNoCallback = _selectedParams.SoftPhysicsParams[pair.Key].StorableParam.val;
            }

            foreach(var pair in MorphConfigs)
            {
                pair.Value.StorableFloat.valNoCallback = _selectedParams.MorphParams[pair.Key].Value;
            }

            SelectedProfile.IsRightSelected = isRight;

            if(!SelectedProfile.HasSoftJoints())
            {
                VisualizeClothingCoverageJssc.val = Person.IsFemale ? "None" : Constant.COVERAGE;
            }

            Script.Tabs.ClothingWindow.Refresh();
        }

        float _relativeXGravityLeft, _relativeXGravityRight;
        float _relativeYGravityLeft, _relativeYGravityRight;
        float _relativeZGravityLeft, _relativeZGravityRight;

        internal override void PreRegisterParams()
        {
            Script.PreRegisterClothingParam(MorphingTransitionJsf);
            _profilesToPreRegister.ForEach(PreRegisterProfileParams);
        }

        void PreRegisterProfileParams(ClothingAdjustmentProfile profile)
        {
            Script.PreRegisterClothingParam(profile.Symmetric.StorableParam);
            Script.PreRegisterClothingParam(profile.LeftActive.StorableParam);
            Script.PreRegisterClothingParam(profile.RightActive.StorableParam);
            Script.PreRegisterClothingParam(profile.Transition.StorableParam);
            PreRegisterAdjustmentParams(profile.Params[Side.LEFT]);
            PreRegisterAdjustmentParams(profile.RightCachedParams);
        }

        void PreRegisterAdjustmentParams(ClothingAdjustmentParams @params)
        {
            foreach(var pair in @params.AllParams)
            {
                Script.PreRegisterClothingParam(pair.Value.StorableParam);
            }

            foreach(var pair in @params.MorphParams)
            {
                Script.PreRegisterClothingParam(pair.Value.StorableParam);
            }
        }

        void RegisterProfileParams(ClothingAdjustmentProfile profile)
        {
            Script.RegisterBool(profile.Symmetric.StorableParam);
            Script.RegisterBool(profile.LeftActive.StorableParam);
            Script.RegisterBool(profile.RightActive.StorableParam);
            Script.RegisterFloat(profile.Transition.StorableParam);
            RegisterAdjustmentParams(profile.Params[Side.LEFT]);
            RegisterAdjustmentParams(profile.RightCachedParams);
        }

        void RegisterAdjustmentParams(ClothingAdjustmentParams @params)
        {
            foreach(var pair in @params.AllParams)
            {
                Script.RegisterFloat(pair.Value.StorableParam);
            }

            foreach(var pair in @params.MorphParams)
            {
                Script.RegisterFloat(pair.Value.StorableParam);
            }
        }

        internal void Execute()
        {
            UpdateForces();
            foreach(var pair in MorphConfigs)
            {
                SyncMorph(pair.Key, pair.Value);
            }
        }

        void UpdateForces()
        {
            int count = ActiveProfiles.Count;
            if(count == 0)
            {
                return;
            }

            _relativeXGravityLeft = gravityPhysicsHandler.CustomGravityLeft.RelativeTo(leftRb.transform, Vector3.right);
            _relativeXGravityRight = gravityPhysicsHandler.CustomGravityRight.RelativeTo(rightRb.transform, Vector3.right);
            _relativeYGravityLeft = gravityPhysicsHandler.CustomGravityLeft.RelativeTo(leftRb.transform, Vector3.up);
            _relativeYGravityRight = gravityPhysicsHandler.CustomGravityRight.RelativeTo(rightRb.transform, Vector3.up);
            _relativeZGravityLeft = gravityPhysicsHandler.CustomGravityLeft.RelativeTo(leftRb.transform, Vector3.forward);
            _relativeZGravityRight = gravityPhysicsHandler.CustomGravityRight.RelativeTo(rightRb.transform, Vector3.forward);

            var leftForce = Vector3.zero;
            var rightForce = Vector3.zero;

            if(count == 1)
            {
                var profile = ActiveProfiles.First();
                if(profile.LeftActive.Value)
                {
                    leftForce = new Vector3(
                        CalculateXForceLeft(profile),
                        CalculateYForceLeft(profile),
                        CalculateZForceLeft(profile)
                    );
                }

                if(profile.RightActive.Value)
                {
                    rightForce = new Vector3(
                        CalculateXForceRight(profile),
                        CalculateYForceRight(profile),
                        CalculateZForceRight(profile)
                    );
                }
            }
            else
            {
                Vector3 leftMin = Vector3.zero, rightMin = Vector3.zero;
                Vector3 leftMax = Vector3.zero, rightMax = Vector3.zero;
                Vector3 leftSum = Vector3.zero, rightSum = Vector3.zero;

                foreach(var profile in ActiveProfiles)
                {
                    if(profile.LeftActive.Value)
                    {
                        var left = new Vector3(CalculateXForceLeft(profile), CalculateYForceLeft(profile), CalculateZForceLeft(profile));
                        leftMin = Vector3.Min(left, leftMin);
                        leftMax = Vector3.Max(left, leftMax);
                        leftSum += left;
                    }

                    if(profile.RightActive.Value)
                    {
                        var right = new Vector3(CalculateXForceRight(profile), CalculateYForceRight(profile), CalculateZForceRight(profile));
                        rightMin = Vector3.Min(right, rightMin);
                        rightMax = Vector3.Max(right, rightMax);
                        rightSum += right;
                    }
                }

                leftForce = ClampVector(leftSum, leftMin, leftMax);
                rightForce = ClampVector(rightSum, rightMin, rightMax);
            }

            LeftForce = leftForce;
            RightForce = rightForce;

            leftRb.AddForce(leftRb.transform.TransformDirection(leftForce), ForceMode.Acceleration);
            rightRb.AddForce(rightRb.transform.TransformDirection(rightForce), ForceMode.Acceleration);
        }

        static Vector3 ClampVector(Vector3 vector, Vector3 min, Vector3 max) => new Vector3(
            vector.x > 0 ? Mathf.Min(vector.x, max.x) : Mathf.Max(vector.x, min.x),
            vector.y > 0 ? Mathf.Min(vector.y, max.y) : Mathf.Max(vector.y, min.y),
            vector.z > 0 ? Mathf.Min(vector.z, max.z) : Mathf.Max(vector.z, min.z)
        );

        float CalculateXForceLeft(ClothingAdjustmentProfile profile)
        {
            float force = profile.GetLeftXForce();
            float target = force - _relativeXGravityLeft;

            /* Squeeze */
            if(force > 0)
            {
                return _relativeXGravityLeft > 0
                    ? Mathf.Max(0, target) /* Rotate right */
                    : force; /* Rotate left */
            }

            /* Separate */
            return _relativeXGravityLeft > 0
                ? force /* Rotate right */
                : Mathf.Min(0, target); /* Rotate left */
        }

        float CalculateXForceRight(ClothingAdjustmentProfile profile)
        {
            float force = profile.GetRightXForce();
            float target = force + _relativeXGravityRight;

            /* Squeeze */
            if(force > 0)
            {
                return _relativeXGravityRight > 0
                    ? -force /* Rotate right */
                    : -Mathf.Max(0, target); /* Rotate left */
            }

            /* Separate */
            return _relativeXGravityRight > 0
                ? -Mathf.Min(0, target) /* Rotate right */
                : -force; /* Rotate left */
        }

        float CalculateYForceLeft(ClothingAdjustmentProfile profile) => CalculateYForce(
            _relativeYGravityLeft,
            gravityPhysicsHandler.RefDownGravityLeft.y,
            profile.GetLeftYForce()
        );

        float CalculateYForceRight(ClothingAdjustmentProfile profile) => CalculateYForce(
            _relativeYGravityRight,
            gravityPhysicsHandler.RefDownGravityRight.y,
            profile.GetRightYForce()
        );

        static float CalculateYForce(float relativeGravity, float refDown, float gForce)
        {
            float target = gForce - (relativeGravity - refDown);
            return gForce > 0
                ? Mathf.Max(0, target) /* Lift up */
                : gForce; /* Press down */
        }

        float CalculateZForceLeft(ClothingAdjustmentProfile profile) => CalculateZForce(_relativeZGravityLeft, profile.GetLeftZForce());
        float CalculateZForceRight(ClothingAdjustmentProfile profile) => CalculateZForce(_relativeZGravityRight, profile.GetRightZForce());

        static float CalculateZForce(float relativeGravity, float gForce)
        {
            float target = gForce - relativeGravity;

            /* Pull */
            if(gForce > 0)
            {
                return relativeGravity > 0
                    ? Mathf.Max(0, target) /* Rotate forward */
                    : gForce; /* Rotate back */
            }

            /* Push */
            return relativeGravity > 0
                ? gForce /* Rotate back */
                : Mathf.Min(0, target); /* Rotate forward */
        }

        void SyncMorph(string id, ClothingMorphConfig config)
        {
            float leftValue = 0f, rightValue = 0f;

            foreach(var profile in ActiveProfiles)
            {
                if(profile.LeftActive.Value)
                {
                    leftValue = Mathf.Max(leftValue, profile.GetLeftMorphParamValue(id));
                }

                if(profile.RightActive.Value)
                {
                    rightValue = Mathf.Max(rightValue, profile.GetRightMorphParamValue(id));
                }
            }

            config.Left.morphValue = leftValue;
            config.Right.morphValue = rightValue;
        }

        internal void ResetMorphs()
        {
            foreach(var pair in MorphConfigs)
            {
                var config = pair.Value;
                config.Left.morphValue = 0f;
                config.Right.morphValue = 0f;
            }
        }

        internal void LoadSelectedItemProfile(string fullPath, JSONClass jc, bool isSettingTarget = false)
        {
            // Debug.Log($"{Script.DomainName}: LoadSelectedItemProfile fullPath {fullPath}, isSettingTarget {isSettingTarget}");
            var parsedPath = ParseProfilePath(fullPath);
            if(string.IsNullOrEmpty(parsedPath.itemSubDir))
            {
                Script.Loggr.LogError($"LoadSelectedItemProfile: Invalid profile path {fullPath}", false);
                return;
            }

            FinishLoadProfile(SelectedItem, parsedPath, jc, isSettingTarget);
        }

        internal void LoadItemByUidProfile(string uid, string fullPath, JSONClass jc)
        {
            // Debug.Log($"{Script.DomainName}: LoadItemByUidProfile uid {uid}, fullPath {fullPath}");
            var parsedPath = ParseProfilePath(fullPath);
            if(string.IsNullOrEmpty(parsedPath.itemSubDir))
            {
                Script.Loggr.LogError($"LoadItemByUidProfile: Invalid profile path {fullPath}", false);
                return;
            }

            var item = _store.Items[uid];
            FinishLoadProfile(item, parsedPath, jc);
        }

        internal void LoadEnabledItemProfileByPath(string fullPath, JSONClass jc, bool isSettingTarget = false)
        {
            // Debug.Log($"{Script.DomainName}: LoadEnabledItemProfileByPath fullPath {fullPath}, isSettingTarget {isSettingTarget}");
            var parsedPath = ParseProfilePath(fullPath);
            if(string.IsNullOrEmpty(parsedPath.itemSubDir))
            {
                if(Script.FirstCalibrationDone)
                {
                    Script.Loggr.LogError($"LoadEnabledItemByPathProfile: Invalid profile path {fullPath}", false);
                }

                return;
            }

            var firstItem = _store.EnabledItems.FirstOrDefault(item => IsValidItem(item) && item.PresetSubDir == parsedPath.itemSubDir);
            if(firstItem == null)
            {
                if(Script.FirstCalibrationDone)
                {
                    Script.Loggr.LogError($"No enabled clothing item found for profile '{parsedPath.fileName}' (Full path: {fullPath})", false);
                }

                return;
            }

            FinishLoadProfile(firstItem, parsedPath, jc, isSettingTarget);
        }

        void FinishLoadProfile(ClothingItem item, ParsedPath parsedPath, JSONClass jc, bool isSettingTarget = false)
        {
            // Debug.Log($"{Script.DomainName}: FinishLoadProfile {item.Label}, path {parsedPath.fullPath}, isSettingTarget {isSettingTarget}");
            var profile = GetProfile(item);
            bool isSameItem = DetermineIsSameItem(item, parsedPath);

            if(isSettingTarget)
            {
                RestoreTargetProfileFromJSON(item, profile, jc);
                profile.TargetName = BuildProfileName(parsedPath, isSameItem);
                profile.TargetFullPath.SetVal(parsedPath.fullPath);
            }
            else
            {
                profile.LoadedName = BuildProfileName(parsedPath, isSameItem);
                profile.LoadedFullPath = parsedPath.fullPath;
                RestoreProfileFromJSON(item, profile, jc, parsedPath.packageId, isSameItem);

                if(Script.FirstCalibrationDone)
                {
                    SyncActiveProfiles(item);
                    profile.SetAllAdjustmentValues(upDownMassMultiplier, sidewaysMassMultiplier, inOutMassMultiplier, Script.IsBootyMagic);
                    SyncAdjustments();
                    Script.SyncMorphsAngleOffsets(MorphingTransitionJsf.val);
                }

                if(profile == SelectedProfile)
                {
                    SyncSelectedProfile(profile.IsRightSelected ? Side.RIGHT : Side.LEFT);
                }
            }
        }

        static bool DetermineIsSameItem(ClothingItem item, ParsedPath parsedPath)
        {
            bool isSameItem = item.PresetSubDir == parsedPath.itemSubDir;
            if(item.IsBuiltIn && !isSameItem)
            {
                bool isSameDefaultBuiltInItem =
                    item.DazItem.creatorName == parsedPath.itemSubDir &&
                    item.DazItem.displayName == parsedPath.fileName.Replace(".json", "");
                isSameItem = isSameDefaultBuiltInItem;
            }

            return isSameItem;
        }

        ParsedPath ParseProfilePath(string fullPath) => new ParsedPath(fullPath, $"{FileUtils.savesDir}/{ClothingProfilesDir}");

        internal void SetSelectedProfileLoadedInfo(string fullPath)
        {
            if(SelectedItem == null || SelectedProfile == null)
            {
                Script.Loggr.LogError("SetSelectedProfileLoadedInfo: Selected item or profile was null");
                return;
            }

            var parsedPath = ParseProfilePath(fullPath);
            if(string.IsNullOrEmpty(parsedPath.itemSubDir))
            {
                Script.Loggr.LogError($"SetSelectedProfileLoadedInfo: Invalid profile path {fullPath}", false);
                return;
            }

            SelectedProfile.LoadedName = BuildProfileName(parsedPath, true);
            SelectedProfile.LoadedFullPath = parsedPath.fullPath;
            SelectedProfile.SaveValues();
        }

        static string BuildProfileName(ParsedPath parsedPath, bool isSameItem)
        {
            string name = parsedPath.fileName.Replace(".json", "");
            if(!string.IsNullOrEmpty(parsedPath.packageId))
            {
                name += $" ({parsedPath.packageId}.var)";
            }
            else if(!isSameItem)
            {
                name += $" ({parsedPath.itemSubDir})";
            }

            return name;
        }

        void RestoreProfileFromJSON(ClothingItem item, ClothingAdjustmentProfile profile, JSONClass jc, string packageId, bool isSameItem)
        {
            if(item == null)
            {
                return;
            }

            string prefix = $"{item.Label}:";
            foreach(string key in jc.Keys.ToList())
            {
                JSONUtils.RenameKey(jc, key, $"{prefix}{key}");
            }

            profile.RestoreFromJSON(jc);
            AssignCoveredJointsFromJSON(jc, $"{prefix}Covered Joints", profile, isSameItem);
            SetTargetProfileFromJSON(jc, $"{prefix}Target Profile", profile, packageId);
            profile.SaveValues();
        }

        void AssignCoveredJointsFromJSON(JSONClass jc, string key, ClothingAdjustmentProfile profile, bool isSameItem)
        {
            if(!jc.HasKey(key))
            {
                return;
            }

            var coveredJointsJSON = jc[key].AsObject;
            AssignJointsOnSide(Side.LEFT, coveredJointsJSON, profile, isSameItem);
            AssignJointsOnSide(Side.RIGHT, coveredJointsJSON, profile, isSameItem);
        }

        void AssignJointsOnSide(
            string side,
            JSONClass coveredJoints,
            ClothingAdjustmentProfile profile,
            bool isSameItem
        )
        {
            var coveredJointsIndices = JSONUtils
                .GetJSONArrayEnumerable(coveredJoints, side)
                .Select(node => node.Value)
                .ToList();
            if(coveredJointsIndices.Count == 0)
            {
                return;
            }

            var joints = new List<SoftJoint>();
            if(isSameItem)
            {
                joints.AddRange(softPhysicsHandler.AllSoftJoints[side]
                    .Where(joint => coveredJointsIndices.Contains($"{joint.SoftVerticesSet.targetVertex}"))
                    .Distinct());
                AssignJoints(joints, profile, side);
            }

            // set true even if no joints are assigned because soft physics percentage floats are still loaded from json
            profile.IsCoverageMapped = true;
        }

        void SetTargetProfileFromJSON(JSONClass jc, string key, ClothingAdjustmentProfile profile, string packageId)
        {
            if(!jc.HasKey(key))
            {
                profile.ResetTarget();
                return;
            }

            string targetFullPath = ParsePresetPath(jc[key], packageId);
            if(targetFullPath == "Profile Off")
            {
                SetProfileOffAsTarget(profile);
            }
            else if(!string.IsNullOrEmpty(targetFullPath) && FileUtils.FileExists(targetFullPath))
            {
                PresetHandler.LoadEnabledItemProfileFromPath(targetFullPath, true);
            }
            else
            {
                profile.ResetTarget();
            }
        }

        void RestoreTargetProfileFromJSON(ClothingItem item, ClothingAdjustmentProfile profile, JSONClass jc)
        {
            if(item == null)
            {
                return;
            }

            if(!profile.HasTarget)
            {
                /* Assumes ProfileTransitionJsf doesn't change profile Transition.StorableFloat.val until a target has been selected */
                ProfileTransitionJsf.SetDefaultValNoCallback();
            }

            string prefix = $"{item.Label}:";
            foreach(string key in jc.Keys.ToList())
            {
                JSONUtils.RenameKey(jc, key, $"{prefix}{key}");
            }

            profile.TargetParams[Side.LEFT] = NewClothingAdjustmentParams(prefix, false, false);
            bool isSymmetric = jc[$"{prefix}{SymmetricJsb.name}"].AsBool;
            profile.TargetParams[Side.LEFT].RestoreFromJSON(jc);

            if(isSymmetric)
            {
                profile.TargetParams[Side.RIGHT] = profile.TargetParams[Side.LEFT];
            }
            else
            {
                profile.TargetParams[Side.RIGHT] = NewClothingAdjustmentParams(prefix, true, false);
                profile.TargetParams[Side.RIGHT].RestoreFromJSON(jc);
            }

            if(Script.FirstCalibrationDone)
            {
                profile.SetTargetAdjustmentValues(upDownMassMultiplier, sidewaysMassMultiplier, inOutMassMultiplier, Script.IsBootyMagic);
            }
        }

        string ParsePresetPath(JSONNode jsonNode, string packageId)
        {
            try
            {
                var jsonClass = jsonNode.AsObject;
                if(jsonClass == null)
                {
                    // 1. returns "Profile Off" in the case of target presets
                    // 2. automatically works with pre-v1.0.0 presets and scene files, since in those the path is not wrapped in a JSONClass
                    return jsonNode.Value;
                }

                if(!jsonClass.HasKey("presetName"))
                {
                    return null;
                }

                string value = jsonClass["presetName"].Value;
                if(!string.IsNullOrEmpty(packageId))
                {
                    value = value.Replace("SELF:/", $"{packageId}:/");
                }

                // Debug.Log($"parsed path is {value}");
                return value;
            }
            catch(Exception e)
            {
                Script.Loggr.LogError($"ParsePresetPath: {e}");
                return null;
            }
        }

        internal JSONClass GetSelectedProfileJSON()
        {
            if(SelectedItem == null || SelectedProfile == null)
            {
                return null;
            }

            var jc = new JSONClass();
            var scriptJSON = Script.GetFullBaseJSON();
            if(!scriptJSON.HasKey(SelectedItem.Label))
            {
                Script.Loggr.LogError($"GetSelectedProfileJSON: No JSON for {SelectedItem.Label}");
                return null;
            }

            var profileJSON = scriptJSON[SelectedItem.Label].AsObject;
            bool isSymmetric = profileJSON[SymmetricJsb.name].AsBool;
            foreach(string key in profileJSON.Keys)
            {
                if(isSymmetric && key.EndsWith(" (Right)"))
                {
                    continue;
                }

                jc[key] = profileJSON[key];
            }

            if(SelectedProfile.IsCoverageMapped && SelectedProfile.HasSoftJoints())
            {
                jc["Covered Joints"] = new JSONClass();
                jc["Covered Joints"][Side.LEFT] = JointTargetVertexIndicesArray(SelectedProfile.CoveredSoftJoints[Side.LEFT]);
                jc["Covered Joints"][Side.RIGHT] = JointTargetVertexIndicesArray(SelectedProfile.CoveredSoftJoints[Side.RIGHT]);
            }

            if(SelectedProfile.HasTarget && SelectedProfile.TargetFullPath.Value != SelectedProfile.LoadedFullPath)
            {
                string path = SelectedProfile.TargetFullPath.Value;
                if(path == "Profile Off")
                {
                    jc["Target Profile"] = path;
                }
                else
                {
                    jc["Target Profile"] = new JSONClass
                    {
                        // Hacky way to ensure that path is updated to SELF:/ by package manager when packaging the preset with its target preset in the same var
                        ["presetName"] = SelectedProfile.TargetFullPath.Value,
                    };
                }
            }

            return jc;
        }

        static JSONArray JointTargetVertexIndicesArray(List<SoftJoint> softJoints)
        {
            var indicesArray = new JSONArray();
            foreach(var joint in softJoints)
            {
                indicesArray.Add(joint.SoftVerticesSet.targetVertex.ToString());
            }

            return indicesArray;
        }
    }
}
