using NAT.Common.Storables;

namespace NAT.Common.Handlers.Configs
{
    sealed class PhysicsCalculator
    {
        internal StorableFloat MultiplierJsf { get; set; }
        internal MultiplierConfig PrimaryConfig { get; set; }
        internal MultiplierConfig SecondaryConfig { get; set; }

        internal float Calculate(float mass, float softness, bool usePrimary) =>
            Calculate(usePrimary ? PrimaryConfig : SecondaryConfig, mass, softness);

        static float Calculate(MultiplierConfig config, float valueA, float valueB)
        {
            if(config == null || config.BaseValue == 0)
            {
                return 0;
            }

            float baseValue = config.BaseValue;
            return baseValue +
                baseValue * config.MultiplierA * config.CurveA(valueA) +
                baseValue * config.MultiplierB * config.CurveB(valueB);
        }

        internal float CalculateDynamic(float effect, float valueA, float valueB, bool usePrimary) =>
            CalculateDynamic(usePrimary ? PrimaryConfig : SecondaryConfig, effect, valueA, valueB);

        float CalculateDynamic(MultiplierConfig config, float effect, float valueA, float valueB)
        {
            float multiplier = (MultiplierJsf?.val ?? 1) * effect;
            return multiplier *
            (
                config.BaseValue +
                config.MultiplierA * config.CurveA(valueA) +
                config.MultiplierB * config.CurveB(valueB)
            );
        }
    }
}
