using NAT.Common.Storables;

namespace NAT.Common.Handlers
{
    abstract class ForceHandler : Handler
    {
        protected AngleCalculator AngleCalculator { get; private set; }

        internal StorableFloat BaseJsf { get; set; }
        internal StorableFloat UpJsf { get; set; }
        internal StorableFloat DownJsf { get; set; }
        internal StorableFloat DepthInJsf { get; set; }
        internal StorableFloat DepthOutJsf { get; set; }
        internal StorableFloat SidewaysInJsf { get; set; }
        internal StorableFloat SidewaysOutJsf { get; set; }

        protected float UserUpMultiplier() => Calc.QuadraticRegression(BaseJsf.val * UpJsf.val);
        protected float UserDownMultiplier() => Calc.QuadraticRegression(BaseJsf.val * DownJsf.val);
        protected float UserDepthOutMultiplier() => Calc.QuadraticRegression(BaseJsf.val * DepthOutJsf.val);
        protected float UserDepthInMultiplier() => Calc.QuadraticRegression(BaseJsf.val * DepthInJsf.val);
        protected float UserSidewaysInMultiplier() => Calc.QuadraticRegression(BaseJsf.val * SidewaysInJsf.val);
        protected float UserSidewaysOutMultiplier() => Calc.QuadraticRegression(BaseJsf.val * SidewaysOutJsf.val);

        internal override void Init<T>(ProviderBase provider)
        {
            base.Init<T>(provider);
            AngleCalculator = provider.GetInstance<AngleCalculator>();
        }

        protected StorableFloat NewMultiplierJsf(string name) => new StorableFloat(name, 1.00f, 0.00f, 2.00f)
        {
            setJSONCallbackFunction = OnMultiplierValueChanged,
        };

        protected abstract void OnMultiplierValueChanged(JSONStorableFloat jsf);

        internal void SaveCalibrated()
        {
            BaseJsf?.SaveCalibrated();
            UpJsf?.SaveCalibrated();
            DownJsf?.SaveCalibrated();
            DepthInJsf?.SaveCalibrated();
            DepthOutJsf?.SaveCalibrated();
            SidewaysInJsf?.SaveCalibrated();
            SidewaysOutJsf?.SaveCalibrated();
        }

        // ReSharper disable once UnusedMemberInSuper.Global
        internal abstract void ResetAll();
    }
}
