using System.Linq;
using UnityEngine;

namespace NAT.Common.Models
{
    sealed class Breast : CalibratedJoint
    {
        readonly Rigidbody _nippleRb;

        internal Breast(ConfigurableJoint joint, Rigidbody nippleRb, Person person) : base(joint, person.ChestRb, person)
        {
            _nippleRb = nippleRb;
        }

        internal override void SetFunctionsWhenSoftPhysicsOff(Person person, int[] centerPositionalVerts)
        {
            softPhysicsOn = false;
            if(!person.IsFemale)
            {
                centerVertexIndices = centerPositionalVerts;
            }
        }

        internal override void SetFunctionsWhenSoftPhysicsOn(
            Rigidbody[] softJointOuterRbs,
            Rigidbody[] softJointMainRbs,
            Rigidbody[] softJointCenterRbs
        )
        {
            softPhysicsOn = true;
            this.softJointOuterRbs = softJointOuterRbs;
            this.softJointCenterRbs = softJointCenterRbs;
            combinedSoftRbs = softJointMainRbs.Concat(softJointOuterRbs).ToArray();
        }

        protected override Vector3 RelativePosition()
        {
            Vector3 position;
            if(softPhysicsOn)
            {
                position = Calc.AverageRigidbodiesPosition(combinedSoftRbs);
            }
            else
            {
                position = person.IsFemale ? _nippleRb.position : person.SkinVertsAveragePosition(centerVertexIndices);
            }

            return Calc.RelativePosition(originRb, position);
        }

        protected override float RelativeDepth()
        {
            float result;

            if(person.IsFemale)
            {
                if(softPhysicsOn)
                {
                    var centerToRb = Calc.RelativePosition(
                        simulatedJointPosition,
                        Rigidbody.transform,
                        originRb.transform,
                        Calc.AverageRigidbodiesPosition(softJointCenterRbs)
                    );
                    var outerToRb = Calc.RelativePosition(
                        simulatedJointPosition,
                        Rigidbody.transform,
                        originRb.transform,
                        Calc.AverageRigidbodiesPosition(softJointOuterRbs)
                    );
                    float softJointsDepth = (centerToRb - outerToRb).z;
                    result = (JointRelativeDepth() + softJointsDepth) / 2;
                }
                else
                {
                    result = JointRelativeDepth();
                }
            }
            else
            {
                result = JointRelativeDepth();
            }

            return result;
        }

        internal override void SetRotation(float x, float y, float z)
        {
            var rotation = bone.baseJointRotation;
            rotation.x = -x;
            rotation.y = -y;
            rotation.z = z;
            bone.baseJointRotation = rotation;
        }

        float JointRelativeDepth()
        {
            var position = Calc.RelativePosition(
                simulatedJointPosition,
                Rigidbody.transform,
                originRb.transform,
                Rigidbody.position
            );
            return position.z;
        }
    }
}
