using NAT.Common.Handlers.Configs;
using UnityEngine;

namespace NAT.Common.Models
{
    sealed class ClothingItem : SimItem
    {
        internal DAZClothingItem DazItem { get; }
        internal bool IsBuiltIn { get; }
        internal string Label { get; }
        internal string PresetSubDir { get; }
        internal ClothSimControl SimControl { get; private set; }
        internal ClothingAdjustmentProfile ChestProfile { get; set; }
        internal ClothingAdjustmentProfile HipProfile { get; set; }
        bool _isSim;

        internal ClothingItem(DAZClothingItem dazItem, string label, JSONStorableBool enabledJsb) : base(enabledJsb)
        {
            DazItem = dazItem;
            IsBuiltIn = dazItem.creatorName == "MeshedVR";
            Label = label;
            PresetSubDir = dazItem.creatorName + "/" + dazItem.displayName;
        }

        internal void RefreshControls()
        {
            if(!EnabledJsb.val || retryCount > 10)
            {
                retryCount = 0;
                NeedsRetry = false;
                return;
            }

            if(NeedsRetry)
            {
                retryCount++;
            }

            if(!SimControl)
            {
                SimControl = DazItem.GetComponentInChildren<ClothSimControl>();
            }

            NeedsRetry = !SimControl;
            if(!NeedsRetry)
            {
                retryCount = 0;
                if(SimControl)
                {
                    _isSim = true;
                    SimForceV3 = SimControl.GetVector3JSONParam("force");
                }

                if(shouldSetSimForce)
                {
                    SetSimForce(simForce);
                }
            }
        }

        internal void SetSimForce(Vector3 force)
        {
            if(!_isSim)
            {
                return;
            }

            if(!SimControl || !SimControl.simEnabledJSON.val)
            {
                simForce = force;
                shouldSetSimForce = true;
                return;
            }

            shouldSetSimForce = false;

            // Sanity check
            if(SimForceV3 == null)
            {
                Loggr.Error($"{DazItem.uid}: force storable was null!");
                return;
            }

            if(SimControl.clothSettings)
            {
                SimControl.clothSettings.GravityMultiplier = 0f;
            }

            SimForceV3.val = force;
        }

        internal void UnsetSimForce()
        {
            shouldSetSimForce = false;

            if(!SimControl || !SimControl.simEnabledJSON.val)
            {
                return;
            }

            // Sanity check
            if(SimForceV3 == null)
            {
                Loggr.Error($"{DazItem.uid}: force storable was null!");
                return;
            }

            if(SimControl.clothSettings)
            {
                SimControl.clothSettings.GravityMultiplier = SimControl.GetFloatParamValue("gravityMultiplier");
            }

            SimForceV3.val = Vector3.zero;
        }
    }
}
