using System.Text.RegularExpressions;

namespace NAT.Common.Models
{
    sealed class ConfigurableMorph
    {
        readonly DAZMorph _morph;
        internal JSONStorableBool IsNegativeJsb { get; }
        internal JSONStorableFloat SoftMultiplierJsf { get; } // TODO refactor as a base multiplier
        internal JSONStorableFloat MassMultiplierJsf { get; }
        bool _isNegative;
        float _softMultiplier;
        float _massMultiplier;

        internal string MorphName { get; }
        readonly string _label;
        internal string MorphDisplayName { get; }

        internal ConfigurableMorph(DAZMorph morph, bool isNegative, JSONStorableFloat softMultiplierJsf, JSONStorableFloat massMultiplierJsf)
        {
            _morph = morph;
            MorphName = _morph.morphName;
            _label = IsCenterMorph() || IsLeftRightMorph()
                ? _morph.displayName
                : Regex.Replace(_morph.displayName, " L$", "");
            MorphDisplayName = _label.Substring(MorphName.IndexOf(" ") + 1);

            IsNegativeJsb = new JSONStorableBool("isNegative", isNegative);
            SoftMultiplierJsf = softMultiplierJsf;
            MassMultiplierJsf = massMultiplierJsf;

            IsNegativeJsb.setCallbackFunction = val => _isNegative = val;
            softMultiplierJsf.setCallbackFunction = val => _softMultiplier = val;
            massMultiplierJsf.setCallbackFunction = val => _massMultiplier = val;

            _isNegative = IsNegativeJsb.val;
            _softMultiplier = softMultiplierJsf.val;
            _massMultiplier = massMultiplierJsf.val;
        }

        internal float Calculate(float effect, float mass, float softness)
        {
            float value =
                softness * _softMultiplier * effect +
                mass * _massMultiplier * effect;

            bool inRange = _isNegative ? value < 0 : value > 0;
            return inRange ? Calc.RoundToDecimals(value) : 0;
        }

        internal void SetValue(float value) => _morph.morphValue = value;

        internal void SetValue(float effect, float mass, float softness)
        {
            float value = Calculate(effect, mass, softness);
            _morph.morphValue = value;
        }

        internal void Reset() => _morph.morphValue = 0f;

        bool IsCenterMorph() => !_morph.displayName.EndsWith(" L") && !_morph.displayName.EndsWith(" R");
        bool IsLeftRightMorph() => _morph.displayName.StartsWith("LT") || _morph.displayName.StartsWith("RT");

        /* TODO broken */
        internal string ToCodeString(string getMorphMethodName, string subDir = "")
        {
            string dollar = IsCenterMorph() || IsLeftRightMorph() ? "" : "$";
            string sideStr = IsCenterMorph() || IsLeftRightMorph() ? "" : " {side}";

            return $@"
            new ConfigurableMorph(
                Person.{getMorphMethodName}({dollar}""{subDir}{_label}{sideStr}""),
                {_isNegative.ToString().ToLower()},
                new JSONStorableFloat(""softMultiplier"", {SoftMultiplierJsf.val:0.00}f, -3.00f, 3.00f),
                new JSONStorableFloat(""massMultiplier"", {MassMultiplierJsf.val:0.00}f, -3.00f, 3.00f)
            ),";
        }
    }
}
