using System.Linq;
using UnityEngine;

namespace NAT.Common.Models
{
    sealed class Glute : CalibratedJoint
    {
        internal Glute(ConfigurableJoint joint, Person person) : base(joint, person.HipRb, person)
        {
        }

        internal override void SetFunctionsWhenSoftPhysicsOff(Person person, int[] centerPositionalVerts)
        {
            softPhysicsOn = false;
            centerVertexIndices = centerPositionalVerts;
        }

        internal override void SetFunctionsWhenSoftPhysicsOn(
            Rigidbody[] softJointOuterRbs,
            Rigidbody[] softJointMainRbs,
            Rigidbody[] softJointCenterRbs
        )
        {
            softPhysicsOn = true;
            this.softJointOuterRbs = softJointOuterRbs;
            this.softJointCenterRbs = softJointCenterRbs;
            combinedSoftRbs = softJointMainRbs
                .Concat(softJointCenterRbs)
                .Concat(softJointOuterRbs)
                .ToArray();
        }

        protected override Vector3 RelativePosition()
        {
            Vector3 result;

            if(softPhysicsOn)
            {
                result = Calc.RelativePosition(
                    simulatedJointPosition,
                    originRb.transform,
                    Rigidbody.transform,
                    Calc.AverageRigidbodiesPosition(combinedSoftRbs)
                );
            }
            else
            {
                result = Calc.RelativePosition(
                    simulatedJointPosition,
                    originRb.transform,
                    person.SkinVertsAveragePosition(centerVertexIndices)
                );
            }

            return result;
        }

        protected override float RelativeDepth()
        {
            float result;

            if(softPhysicsOn)
            {
                var centerToRb = Calc.RelativePosition(
                    simulatedJointPosition,
                    originRb.transform,
                    Rigidbody.transform,
                    Calc.AverageRigidbodiesPosition(softJointCenterRbs)
                );
                var outerToRb = Calc.RelativePosition(
                    simulatedJointPosition,
                    originRb.transform,
                    Rigidbody.transform,
                    Calc.AverageRigidbodiesPosition(softJointOuterRbs)
                );
                float softJointsDepth = (centerToRb - outerToRb).z;
                result = (JointRelativeDepth() - softJointsDepth) / 2;
            }
            else
            {
                result = JointRelativeDepth();
            }

            return result;
        }

        internal override void SetRotation(float x, float y, float _)
        {
            var rotation = bone.baseJointRotation;
            rotation.x = x;
            rotation.y = y;
            bone.baseJointRotation = rotation;
        }

        float JointRelativeDepth()
        {
            var position = Calc.RelativePosition(
                simulatedJointPosition,
                originRb.transform,
                Rigidbody.transform,
                originRb.position
            );
            return position.z;
        }
    }
}
