using NAT.Common.Handlers.Configs;
using NAT.Common.Storables;
using UnityEngine;

namespace NAT.Common.Models
{
    abstract class PhysicsParam
    {
        protected readonly string paramName;

        internal JSONStorableFloat BaseValueJsf { get; }
        internal StorableFloat OffsetJsf { get; }

        readonly PhysicsCalculator _calculator = new PhysicsCalculator();
        readonly PhysicsCalculator _quicknessOffsetCalculator = new PhysicsCalculator();
        readonly PhysicsCalculator _slownessOffsetCalculator = new PhysicsCalculator();

        internal MultiplierConfig PrimaryConfig
        {
            set { _calculator.PrimaryConfig = value; }
        }

        internal MultiplierConfig SecondaryConfig
        {
            set { _calculator.SecondaryConfig = value; }
        }

        internal MultiplierConfig QuicknessConfig
        {
            set { _quicknessOffsetCalculator.PrimaryConfig = value; }
        }

        internal MultiplierConfig SlownessConfig
        {
            set { _slownessOffsetCalculator.PrimaryConfig = value; }
        }

        protected PhysicsParam(string side, string name, float min, float max)
        {
            paramName = Utils.ParamName(name, side);
            BaseValueJsf = new StorableFloat($"{paramName} Base Value", 0, min, max);
            OffsetJsf = new StorableFloat($"{paramName} Offset", 0, -max, max);
        }

        internal abstract void DoSync();

        internal void UpdateOffsetMinMax()
        {
            OffsetJsf.min = -(BaseValueJsf.val - BaseValueJsf.min);
            OffsetJsf.max = BaseValueJsf.max - BaseValueJsf.val;
        }

        internal void UpdateBaseValue(float mass, float softness, float quickness, float physicsRateMultiplier, bool usePrimary)
        {
            float value = _calculator.Calculate(mass, softness, usePrimary);

            if(quickness > 0)
            {
                float maxQuicknessOffset = _quicknessOffsetCalculator.Calculate(mass, softness, true);
                value += Mathf.Lerp(0, maxQuicknessOffset, quickness);
            }

            if(quickness < 0)
            {
                float maxSlownessOffset = _slownessOffsetCalculator.Calculate(mass, softness, true);
                value += Mathf.Lerp(0, maxSlownessOffset, -quickness);
            }

            BaseValueJsf.val = physicsRateMultiplier * value;
            DoSync();
        }

        internal void UpdateOffsetValue(float value)
        {
            OffsetJsf.valNoCallback = value;
            DoSync();
        }
    }
}
