namespace NAT.Common.Models
{
    abstract class PhysicsParamGroup
    {
        internal string Name { get; set; }
        internal string InfoText { get; set; }
        internal bool DependsOnPhysicsRate { get; set; }
        internal JSONStorableBool ApplyChangesOnlyToLeftParamJsb { get; }
        internal string ValueFormat { get; set; }

        protected PhysicsParam leftParam, rightParam;

        protected PhysicsParamGroup()
        {
            ApplyChangesOnlyToLeftParamJsb = new JSONStorableBool("applyChangesOnlyToLeftParam", false);
        }

        internal virtual void SetCallbackFunctions()
        {
        }

        internal void UpdateBaseValue(float mass, float softness, float quickness, bool usePrimary = true)
        {
            float physicsRateMultiplier = DependsOnPhysicsRate ? Utils.PhysicsRateMultiplier() : 1;
            leftParam.UpdateBaseValue(mass, softness, quickness, physicsRateMultiplier, usePrimary);
            rightParam.UpdateBaseValue(mass, softness, quickness, physicsRateMultiplier, usePrimary);
        }

        internal void UpdateOffsetMinMax()
        {
            leftParam.UpdateOffsetMinMax();
            rightParam.UpdateOffsetMinMax();
        }

        internal virtual bool IsModified => Name != FAT_COLLIDER_FRICTION.NAME && leftParam.OffsetJsf.val != 0;
    }
}
