using NAT.Common.Handlers.Configs;
using NAT.Common.Storables;
using System;

namespace NAT.Common.Models
{
    sealed class SoftPhysicsParam : PhysicsParam
    {
        internal PhysicsCalculator InverseFrictionCalculator { get; set; }
        float _additiveFrictionAdjustment;

        internal GradientConfig GradientConfig { get; }
        internal LocalDistortionPhysicsConfig LdpConfig { get; }

        internal bool UsesRealMass { get; set; }
        internal Action<float, float> Sync { private get; set; }

        internal SoftPhysicsParam(string domain, string side, string name, float min, float max)
            : base(side, name, min, max)
        {
            GradientConfig = new GradientConfig
            {
                MiddleShiftJsf = new StorableFloat($"{paramName} Gradient Middle Shift", 0.00f, -1.00f, 1.00f),
                OuterMultiplierJsf = NewGroupMultiplierJsf(paramName, "Outer"),
                CenterMultiplierJsf = NewGroupMultiplierJsf(paramName, "Center"),
            };

            if(domain == Constant.TITTY_MAGIC)
            {
                GradientConfig.NippleMultiplierJsf = NewGroupMultiplierJsf(paramName, "Nipple");
            }

            LdpConfig = new LocalDistortionPhysicsConfig
            {
                CompressionEffectJsf = new StorableFloat($"{paramName} Local Compression Effect", 0, -3.00f, 3.00f),
                CompressionSensitivityJsf = new StorableFloat($"{paramName} Local Compression Sensitivity", 0, 0.00f, 1.00f),
                StretchingEffectJsf = new StorableFloat($"{paramName} Local Stretching Effect", 0, -3.00f, 3.00f),
                StretchingSensitivityJsf = new StorableFloat($"{paramName} Local Stretching Sensitivity", 0, 0.00f, 1.00f),
            };
        }

        static StorableFloat NewGroupMultiplierJsf(string paramName, string groupName) => new StorableFloat(
            $"{paramName} Gradient {groupName} Multiplier",
            1.00f,
            0.01f,
            Constant.GRADIENT_MULTI_MAX
        );

        internal override void DoSync() =>
            Sync(BaseValueJsf.val + OffsetJsf.val, _additiveFrictionAdjustment);

        internal void UpdateInverseFrictionValue(float friction, float mass, float softness)
        {
            float value = InverseFrictionCalculator.CalculateDynamic(friction, mass, softness, true);
            _additiveFrictionAdjustment = Calc.RoundToDecimals(value, 10000);
            DoSync();
        }

        internal void ResetInverseFrictionValue()
        {
            _additiveFrictionAdjustment = 0;
            DoSync();
        }
    }
}
