using NAT.Common.Handlers.Configs;

namespace NAT.Common.Models
{
    abstract class SoftPhysicsParamGroup : PhysicsParamGroup
    {
        protected readonly ScriptBase script;

        internal bool HasSoftColliderVisualization { get; set; }
        internal float VisualizationLimit { get; set; }
        internal float ClothingAdjustmentVisualizationLimit { get; set; }
        internal float DistortionLimit { get; set; }
        internal bool VisualizeRelativeChange { get; set; }

        internal float LeftClothingAdjustmentAverage { get; set; }
        internal float RightClothingAdjustmentAverage { get; set; }

        SoftPhysicsParam _left, _right;

        protected SoftPhysicsParamGroup(ScriptBase script)
        {
            this.script = script;
        }

        internal SoftPhysicsParam Left
        {
            get { return _left; }
            set
            {
                _left = value;
                leftParam = value;
            }
        }

        internal SoftPhysicsParam Right
        {
            get { return _right; }
            set
            {
                _right = value;
                rightParam = value;
            }
        }

        protected void UpdateGradient(float value, JSONStorableFloat rightJsf)
        {
            rightJsf.val = ApplyChangesOnlyToLeftParamJsb.val ? rightJsf.defaultVal : value;
            _left.DoSync();
            _right.DoSync();
        }

        protected void SetCompressionMaxAndMultiplier(float value)
        {
            _left.LdpConfig.SetCompressionMaxAndMultiplier(value);
            _right.LdpConfig.SetCompressionMaxAndMultiplier(
                ApplyChangesOnlyToLeftParamJsb.val ? _right.LdpConfig.CompressionEffectJsf.defaultVal : value
            );
        }

        protected void SetCompressionSlope(float value)
        {
            _left.LdpConfig.SetCompressionSlope(value);
            _right.LdpConfig.SetCompressionSlope(
                ApplyChangesOnlyToLeftParamJsb.val ? _right.LdpConfig.CompressionSensitivityJsf.defaultVal : value
            );
        }

        protected void SetStretchingMaxAndMultiplier(float value)
        {
            _left.LdpConfig.SetStretchingMaxAndMultiplier(value);
            _right.LdpConfig.SetStretchingMaxAndMultiplier(
                ApplyChangesOnlyToLeftParamJsb.val ? _right.LdpConfig.StretchingEffectJsf.defaultVal : value
            );
        }

        protected void SetStretchingSlope(float value)
        {
            _left.LdpConfig.SetStretchingSlope(value);
            _right.LdpConfig.SetStretchingSlope(
                ApplyChangesOnlyToLeftParamJsb.val ? _right.LdpConfig.StretchingSensitivityJsf.defaultVal : value
            );
        }

        internal void SetFrictionCalculators(PhysicsCalculator leftCalculator, PhysicsCalculator rightCalculator)
        {
            _left.InverseFrictionCalculator = leftCalculator;
            _right.InverseFrictionCalculator = rightCalculator;
        }

        internal void UpdateInverseFrictionValue(float friction, float mass, float softness)
        {
            _left.UpdateInverseFrictionValue(friction, mass, softness);
            _right.UpdateInverseFrictionValue(friction, mass, softness);
        }

        internal void ResetInverseFrictionValue()
        {
            _left.ResetInverseFrictionValue();
            _right.ResetInverseFrictionValue();
        }

        internal float CalculateDistortionOffset(string side, float distortionBase, float distortion) => side == Side.LEFT
            ? _left.LdpConfig.CalculateDistortionOffset(DistortionLimit, distortionBase, distortion)
            : _right.LdpConfig.CalculateDistortionOffset(DistortionLimit, distortionBase, distortion);

        internal override bool IsModified =>
            base.IsModified ||
            _left.GradientConfig.IsModified ||
            _left.LdpConfig.IsModified;
    }
}
