using NAT.Common.Handlers;
using System;
using System.Collections.Generic;

namespace NAT.Common
{
    abstract class ProviderBase
    {
        readonly Dictionary<Type, object> _instances;

        protected ProviderBase(Dictionary<Type, object> instances)
        {
            _instances = instances;
        }

        internal T GetInstance<T>() where T : class
        {
            var instanceType = typeof(T);
            object instance;
            if(_instances.TryGetValue(instanceType, out instance))
            {
                return instance as T;
            }

            throw new InvalidOperationException($"Instance of type {instanceType} not found");
        }

        internal IEnumerable<object> GetHandlers()
        {
            foreach(var pair in _instances)
            {
                if(pair.Value is Handler)
                {
                    yield return pair.Value;
                }
            }
        }
    }
}
