using SimpleJSON;
using System.Collections.Generic;

namespace NAT.Common.Storables
{
    static class StorableDomainName
    {
        internal const string GENERAL = "GENERAL";
        internal const string JOINT_PHYSICS = "JOINT PHYSICS";
        internal const string SOFT_PHYSICS = "SOFT PHYSICS";
        internal const string FORCE_PHYSICS = "FORCE PHYSICS";
        internal const string HARD_COLLIDERS = "HARD COLLIDERS";
        internal const string MORPHS = "MORPHS";
        internal const string FRICTION = "FRICTION";
        internal const string UI_ACTIONS = "UI ACTIONS";
        internal const string INTERNAL = "INTERNAL - Do not touch";
        internal const string CLOTHING = "CLOTHING";
    }

    sealed class StorableDomain
    {
        internal string Name { get; }
        readonly string _description;
        readonly List<IStorableParam> _storables = new List<IStorableParam>();
        readonly HashSet<string> _temporaryStorableNames = new HashSet<string>();

        internal StorableDomain(string name, string description)
        {
            Name = name;
            _description = description;
        }

        internal void PreRegisterHeader(int i)
        {
            PreRegister(new StorableString($"{new string(' ', i)}", "") // unique spacer storable name
            {
                Temporary = true,
            });
            PreRegister(new StorableString($"      {Name}", _description)
            {
                Temporary = true,
            });
        }

        internal void PreRegister(IStorableParam storableParam)
        {
            _storables.Add(storableParam);
            if(storableParam.Temporary)
            {
                _temporaryStorableNames.Add(storableParam.Name.Split(':')[0]);
            }
        }

        internal void RegisterAllTo(MVRScript script)
        {
            foreach(var storable in _storables)
            {
                storable.RegisterTo(script);
            }
        }

        internal void RemoveTemporaryParamsFrom(JSONClass jc)
        {
            foreach(string name in _temporaryStorableNames)
            {
                jc.Remove(name);
            }
        }
    }
}
