using System;
using System.Diagnostics.CodeAnalysis;

namespace NAT.Common.UI
{
    class Element
    {
        readonly UIDynamic _uiDynamic;
        internal readonly bool showInNested;
        bool _prevActive;

        internal Element(UIDynamic uiDynamic, bool showInNested)
        {
            _uiDynamic = uiDynamic;
            this.showInNested = showInNested;
        }

        internal bool IsVisible() => _uiDynamic.gameObject.activeSelf;
        internal void Show() => _uiDynamic.gameObject.SetActive(true);
        internal void ShowPrev() => _uiDynamic.gameObject.SetActive(_prevActive);

        internal void Hide()
        {
            bool isActive = _uiDynamic.gameObject.activeSelf;
            _uiDynamic.gameObject.SetActive(false);
            ClosePopup();
            _prevActive = isActive;
        }

        internal void SetActiveStyle(bool isActive, bool setInteractable = false, bool highlightIneffective = false) =>
            _uiDynamic.SetActiveStyle(isActive, setInteractable, highlightIneffective);

        internal void SetLabel(string label) => _uiDynamic.SetLabel(label);

        internal void ClosePopup() => (_uiDynamic as UIDynamicPopup)?.SetVisible(false);

        [SuppressMessage("ReSharper", "UnusedMember.Global")]
        internal T GetUIDynamicAs<T>()
        {
            if(_uiDynamic is T)
            {
                return (T) Convert.ChangeType(_uiDynamic, typeof(T));
            }

            return default(T);
        }
    }
}
