using System;
using System.Collections.Generic;

namespace NAT.Common.UI
{
    sealed class Section
    {
        readonly List<Element> _elements = new List<Element>();
        readonly Action<List<Element>> _buildAction;
        internal NavigationButton NavigationButton { get; set; }
        internal bool IsShown { get; private set; }

        internal Section(Action<List<Element>> buildAction)
        {
            _buildAction = buildAction;
        }

        internal void Show()
        {
            NavigationButton.SetActive();
            if(_elements.Count > 0)
            {
                foreach(var element in _elements)
                {
                    element.Show();
                }
            }
            else
            {
                _buildAction(_elements);
            }

            IsShown = true;
        }

        internal void Hide()
        {
            NavigationButton.SetInactive();
            foreach(var element in _elements)
            {
                element.Hide();
            }

            IsShown = false;
        }
    }
}
