using System.Text;
using UnityEngine;
using UnityEngine.UI;

namespace NAT.Common.UI
{
    sealed class SettingsWindow : Window
    {
        internal SettingsWindow(ScriptBase script) : base(script)
        {
        }

        protected override void OnBuild()
        {
            BuildLeftSide();
            BuildRightSide();
            RefreshPauseSceneToggle();
        }

        void BuildLeftSide(bool rightSide = false)
        {
            AddHeaderTextField("User Preferences", rightSide);

            AddInfoTextField(
                $"These settings are saved to file and apply to all {Constant.NATURALIS} instances.",
                rightSide,
                75f
            );

            AddElement(() =>
                script.CreateToggle(script.Bootstrap.CalibrateOnUIClosedJsb, rightSide)
            );

            AddInfoTextField(
                new StringBuilder()
                    .Append("If a calibration is pending, it will trigger when you navigate away from the")
                    .AppendFormat(" {0} or {1} plugin UI", Constant.TITTY_MAGIC, Constant.BOOTY_MAGIC)
                    .ToString(),
                rightSide
            );

            AddElement(() =>
                script.CreateToggle(script.Bootstrap.CalibrateOnPresetLoadedJsb, rightSide)
            );

            AddInfoTextField(
                "Calibrate immediately after loading a preset file. Puts to pending instead if disabled.",
                rightSide,
                75f
            );

            AddElement(() =>
            {
                var toggle = script.CreateToggle(script.Bootstrap.CheckForSizeChangesJsb, rightSide);
                toggle.label = "Monitor: Check For Size Changes";
                return toggle;
            });

            AddInfoTextField(
                "Monitor for size changes to breasts/glutes caused by non-pose morphs and atom scale. A change puts calibration to pending.",
                rightSide
            );

            AddHeaderTextField("Plugin Settings", rightSide);

            AddInfoTextField(
                new StringBuilder()
                    .AppendFormat("These settings save with the current {0} instance.", Constant.NATURALIS)
                    .ToString(),
                rightSide,
                75f
            );

            AddElement(() =>
            {
                var toggle = script.CreateToggle(script.Bootstrap.FreezeMotionSoundJsb, rightSide);
                toggle.AddListener((bool _) => RefreshPauseSceneToggle());
                return toggle;
            });

            AddInfoTextField(
                "Enable the Freeze/Motion Sound toggle (in the below toolbar) during calibration. This prevents all animation from interfering with the calibration, but also disables audio.",
                rightSide,
                125f
            );
        }

        void BuildRightSide(bool rightSide = true)
        {
            AddElement(() =>
            {
                var button = script.CreateButton("Open plugin data directory\nin Windows Explorer", rightSide);
                button.AddListener(FileUtils.OpenDataDirInExplorer);
                button.height = 100f;
                return button;
            });

            AddElement(() =>
            {
                var textField = script.CreateTextField(script.Bootstrap.SavedPrefsInfoJss, rightSide);
                textField.UItext.fontSize = 26;
                textField.backgroundColor = Color.clear;
                var layout = textField.GetComponent<LayoutElement>();
                layout.preferredHeight = 50;
                layout.minHeight = 50;
                return textField;
            });

            AddLabelTextField("LOCAL DISTORTION PHYSICS (LDP)", rightSide);

            AddElement(() =>
                script.CreateToggle(script.Bootstrap.GlutesLdpDefaultEnabledJsb, rightSide)
            );

            AddElement(() =>
                script.CreateToggle(script.Bootstrap.BreastsLdpDefaultEnabledJsb, rightSide)
            );

            AddInfoTextField(
                "Enable LDP by default when plugin is added and when soft physics is enabled. LDP can cause a noticeable framerate drop.",
                rightSide
            );

            AddElement(script.Bootstrap.LdpUpdateRateJssc.name, () =>
                script.CreateScrollablePopup(script.Bootstrap.LdpUpdateRateJssc, rightSide)
            );

            AddInfoTextField(
                "A higher local distortion physics update rate is more responsive but has a higher performance impact.",
                rightSide
            );

            AddSpacer(190, rightSide);

            AddElement(script.Bootstrap.PauseSceneAnimationJsb.name, () =>
                script.CreateToggle(script.Bootstrap.PauseSceneAnimationJsb, rightSide)
            );

            AddInfoTextField(
                "Pause scene animation during calibration. Warning: this has no effect on other animation sources (Animation Pattern, Timeline or other plugin, Cycle Forces, etc.).",
                rightSide,
                125f
            );
        }

        void RefreshPauseSceneToggle()
        {
            var pauseSceneToggle = GetUIDynamic(script.Bootstrap.PauseSceneAnimationJsb.name);
            pauseSceneToggle.SetActiveStyle(!script.Bootstrap.FreezeMotionSoundJsb.val, true);
        }
    }
}
