﻿using System;
using System.Collections;
using UnityEngine;

namespace NAT.Common
{
    class Vector3TransitionParams
    {
        internal readonly Vector3 start;
        internal readonly Vector3 end;
        internal readonly Action<Vector3> callback;

        internal Vector3TransitionParams(Vector3 start, Vector3 end, Action<Vector3> callback)
        {
            this.start = start;
            this.end = end;
            this.callback = callback;
        }
    }

    // ReSharper disable once ClassNeverInstantiated.Global
    class FloatTransitionParams
    {
        internal readonly float start;
        internal readonly float end;
        internal readonly Action<float> callback;

        internal FloatTransitionParams(float start, float end, Action<float> callback)
        {
            this.start = start;
            this.end = end;
            this.callback = callback;
        }
    }

    static class CoroutineUtils
    {
        internal static IEnumerator ExponentialVector3ArrayTransition(float duration, params Vector3TransitionParams[] array)
        {
            float elapsed = 0f;
            while(elapsed < duration)
            {
                elapsed += Time.deltaTime;
                float fraction = elapsed / duration;
                float interpolant = fraction * fraction;
                foreach(var item in array)
                {
                    item.callback(Vector3.Lerp(item.start, item.end, interpolant));
                }

                yield return null;
            }
        }

        internal static IEnumerator ExponentialFloatArrayTransition(float duration, FloatTransitionParams[] array, Action callback = null)
        {
            float elapsed = 0f;
            while(elapsed < duration)
            {
                elapsed += Time.deltaTime;
                float fraction = elapsed / duration;
                float interpolant = fraction * fraction;
                foreach(var item in array)
                {
                    item.callback?.Invoke(Mathf.Lerp(item.start, item.end, interpolant));
                }

                callback?.Invoke();
                yield return null;
            }
        }
    }
}
