using System;
using System.Diagnostics.CodeAnalysis;
using UnityEngine;

namespace NAT.Common
{
    sealed class CustomFixedUpdate
    {
        internal Action Actions { get; set; }

        internal CustomFixedUpdate(float interval)
        {
            _interval = interval;
        }

        [SuppressMessage("ReSharper", "UnusedMember.Global")]
        internal CustomFixedUpdate(JSONStorableFloat frequencyJsf)
        {
            frequencyJsf.setCallbackFunction += SetInterval;
            SetInterval(frequencyJsf.val);
        }

        internal CustomFixedUpdate(JSONStorableStringChooser frequencyJssc)
        {
            frequencyJssc.setCallbackFunction += option => SetInterval(StringOptionToFrequency(option));
            SetInterval(StringOptionToFrequency(frequencyJssc.val));
        }

        float _interval;
        float _timePassed;

        [SuppressMessage("ReSharper", "UnusedMember.Global")]
        internal T Run<T>(Func<T> functions)
        {
            _timePassed += Time.unscaledDeltaTime;
            if(_timePassed >= _interval)
            {
                _timePassed = 0;
                return functions();
            }

            return default(T);
        }

        internal void Run(Action actions)
        {
            _timePassed += Time.unscaledDeltaTime;
            if(_timePassed >= _interval)
            {
                _timePassed = 0;
                actions();
            }
        }

        internal void Run() => Run(Actions);

        void SetInterval(float value) => _interval = 1 / (float) Mathf.RoundToInt(value);

        static float StringOptionToFrequency(string option)
        {
            switch(option)
            {
                case "10 Hz": return 10;
                case "20 Hz": return 20;
                case "30 Hz": return 30;
                case "45 Hz": return 45;
                case "60 Hz": return 60;
                case "Use Physics Rate": return Mathf.RoundToInt(1 / Time.fixedDeltaTime);
                default: throw new ArgumentException($"Invalid option {option}");
            }
        }
    }
}
