using System;
using MVR.FileManagementSecure;
using SimpleJSON;
using System.Collections.Generic;

namespace NAT.Common
{
    static class FileUtils
    {
        readonly static string _dataDir = $"Custom/PluginData/{Constant.EVERLASTER}/{Constant.NATURALIS}";
        internal readonly static string savesDir = $"Saves/PluginData/{Constant.NATURALIS}";

        internal static JSONClass ReadDataJSON(string subPath)
        {
            EnsureDirExists(_dataDir);
            return ReadJSON($"{_dataDir}/{subPath}");
        }

        static JSONClass ReadJSON(string path)
        {
            JSONClass jc = null;

            try
            {
                if(FileManagerSecure.FileExists(path))
                {
                    jc = JSON.Parse(FileManagerSecure.ReadAllText(path)).AsObject;
                }
            }
            catch(Exception e)
            {
                Loggr.Error($"FileUtils.ReadJSON: {e}");
            }

            return jc;
        }

        internal static void WriteDataJSON(string subPath, JSONClass jc, UserActionCallback confirmCallback = null)
        {
            EnsureDirExists(_dataDir);
            string path = $"{_dataDir}/{subPath}";
            FileManagerSecure.WriteAllText(
                path,
                jc.ToString(string.Empty),
                confirmCallback,
                () => Loggr.Message("Save to file cancelled."),
                err => Loggr.Error($"Save to file error: {err}")
            );
        }

        internal static void OpenDataDirInExplorer()
        {
            EnsureDirExists(_dataDir);
            SuperController.singleton.OpenFolderInExplorer(_dataDir);
        }

        internal static List<ShortCut> GetShortCutsForDirectory(string path)
        {
            var cutsForDirectory = FileManagerSecure.GetShortCutsForDirectory(path, true, true, true);
            var rootShortCut = new ShortCut
            {
                displayName = "Root",
                path = FileManagerSecure.GetFullPath("."),
            };
            cutsForDirectory.Insert(0, rootShortCut);
            return cutsForDirectory;
        }

        internal static void EnsureDirExists(string dirName)
        {
            if(!DirectoryExists(dirName))
            {
                FileManagerSecure.CreateDirectory(dirName);
            }
        }

        internal static bool FileExists(string fileName) => FileManagerSecure.FileExists(fileName);
        internal static bool DirectoryExists(string dirName) => FileManagerSecure.DirectoryExists(dirName);

        internal static string ParsePackageIdFromPath(string path)
        {
            if(string.IsNullOrEmpty(path))
            {
                return null;
            }

            int idx = path.IndexOf(":/", StringComparison.Ordinal);
            return idx >= 0 ? path.Substring(0, idx) : null;
        }
    }
}
