//
// ReSharper disable once RedundantUsingDirective
using UnityEngine;

namespace NAT.Common
{
    sealed class Loggr
    {
        readonly string _logPrefix;

        internal Loggr(string domainName)
        {
            _logPrefix = $"{Constant.NATURALIS}.{domainName} v{Bootstrap.NATURALIS_VERSION}";
        }

        internal void LogError(string error, bool report = true) => LogErrorInternal($"{_logPrefix}: {error}", report);

        internal void LogMessage(string message) => SuperController.LogMessage($"{_logPrefix}: {message}");

        internal static void Error(string error, bool report = true) => LogErrorInternal($"{Constant.NATURALIS}: {error}.", report);

        // ReSharper disable once UnusedParameter.Local
        static void LogErrorInternal(string text, bool report)
        {
            if(report)
            {
                text += " Please report the issue!";
            }

            #if ENV_DEVELOPMENT
            {
                SuperController.LogError($"{text}\n{new System.Diagnostics.StackTrace()}");
            }
            #else
            {
                SuperController.LogError(text);
            }
            #endif
        }

        internal static void Message(string message) => SuperController.LogMessage($"{Constant.NATURALIS}: {message}");
    }
}
