// ReSharper disable MemberCanBePrivate.Global UnusedMember.Global
using System;
// ReSharper disable once RedundantUsingDirective
using UnityEngine;

namespace NAT.Common
{
    sealed class SemVer
    {
        readonly Version _version;

        internal SemVer(string versionStr, bool removePrereleaseTag = false)
        {
            try
            {
                /* Without removePrereleaseTag set to true, a versionStr containing a prerelease tag
                 * will result in a null version, and all comparisons will return false.
                 */
                _version = removePrereleaseTag
                    ? new Version(versionStr.Split('-')[0])
                    : new Version(versionStr);
            }
            catch(Exception)
            {
                // Debug.Log("Invalid version string: " + versionStr);
            }
        }

        internal bool IsGte(string other) => IsGte(new SemVer(other));
        internal bool IsMajorGte(string other) => IsMajorGte(new SemVer(other));
        internal bool IsMajorMinorGte(string other) => IsMajorMinorGte(new SemVer(other));
        internal bool IsMajorMinorEqual(string other) => IsMajorMinorEqual(new SemVer(other));

        internal bool IsGte(SemVer other)
        {
            if(_version == null || other._version == null)
            {
                return false;
            }

            return _version >= other._version;
        }

        internal bool IsMajorGte(SemVer other)
        {
            if(_version == null || other._version == null)
            {
                return false;
            }

            return _version.Major >= other._version.Major;
        }

        internal bool IsMajorMinorGte(SemVer other)
        {
            if(_version == null || other._version == null)
            {
                return false;
            }

            var majorMinorA = new Version(_version.Major, _version.Minor);
            var majorMinorB = new Version(other._version.Major, other._version.Minor);

            return majorMinorA >= majorMinorB;
        }

        internal bool IsMajorMinorEqual(SemVer other)
        {
            if(_version == null || other._version == null)
            {
                return false;
            }

            var majorMinorA = new Version(_version.Major, _version.Minor);
            var majorMinorB = new Version(other._version.Major, other._version.Minor);

            return majorMinorA == majorMinorB;
        }
    }
}
