using System.Collections.Generic;
using UnityEngine;

namespace NAT.Common
{
    static class VolumeCalculator
    {
        internal static float SphereVolume(IEnumerable<Vector3> positions, float scale)
        {
            var bounds = CalculateBounds(positions);
            float z = bounds.extents.z * AtomScaleFactor(scale);
            float avgRadius = (bounds.extents.x + bounds.extents.y + z) / 3;
            return 1.33f * Mathf.PI * Mathf.Pow(avgRadius, 3);
        }

        internal static float CupVolume(IEnumerable<Vector3> positions, float scale)
        {
            var bounds = CalculateBounds(positions);
            float z = bounds.size.z * AtomScaleFactor(scale);
            return 4 * Mathf.PI * bounds.size.x / 2 * bounds.size.y / 2 * z / 2 / 3;
        }

        static Bounds CalculateBounds(IEnumerable<Vector3> positions)
        {
            var bounds = new Bounds();
            var min = Vector3.one * float.MaxValue;
            var max = Vector3.one * float.MinValue;
            foreach(var vertex in positions)
            {
                min = Vector3.Min(min, vertex);
                max = Vector3.Max(max, vertex);
            }

            bounds.min = min;
            bounds.max = max;
            return bounds;
        }

        /* This somewhat accurately scales volume to the apparent size when atom scale is adjusted. */
        static float AtomScaleFactor(float scale)
        {
            float atomScaleAdjustment = 1 - Mathf.Abs(Mathf.Log10(Mathf.Pow(scale, 3)));
            float atomScaleFactor = scale >= 1
                ? scale * atomScaleAdjustment
                : scale / atomScaleAdjustment;
            return atomScaleFactor;
        }
    }
}
