﻿using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.Handlers.Configs;
using NAT.Common.Models;
using NAT.Common.Storables;

namespace NAT.TM.Handlers
{
    sealed class NippleErectionHandler : Handler
    {
        BreastSoftPhysicsHandler _softPhysicsHandler;
        BasicMorph _morph;
        internal StorableFloat NippleErectionJsf { get; private set; }
        PhysicsCalculator _springCalculator;
        PhysicsCalculator _damperCalculator;

        internal override void Init<T>(ProviderBase provider)
        {
            base.Init<T>(provider);
            _softPhysicsHandler = provider.GetInstance<BreastSoftPhysicsHandler>();
            NippleErectionJsf = new StorableFloat("Nipple Erection", 0.00f, 0.00f, 1.00f);
            LoadSettings();
        }

        protected override void LoadSettings()
        {
            _morph = new BasicMorph(Person.GetBreastMorph("TM_NippleErection"), 1.0f);
            _springCalculator = new PhysicsCalculator
            {
                PrimaryConfig = new MultiplierConfig
                {
                    BaseValue = 0.50f,
                    MultiplierA = 0.75f,
                    MultiplierB = 0,
                    CurveA = x => 1 - x,
                },
            };
            _damperCalculator = new PhysicsCalculator
            {
                PrimaryConfig = new MultiplierConfig
                {
                    BaseValue = 0.50f,
                    MultiplierA = 0.75f,
                    MultiplierB = 0,
                    CurveA = x => 1 - x,
                },
            };
        }

        internal override void PreRegisterParams()
        {
            Script.PreRegisterMorphsParam(NippleErectionJsf);
        }

        internal void UpdateNipple(float mass, float softness)
        {
            _morph.SetValue(NippleErectionJsf.val);
            if(_softPhysicsHandler.SoftPhysicsOn)
            {
                float springMultiplier = 1 + _springCalculator.CalculateDynamic(NippleErectionJsf.val, mass, softness, true);
                float damperMultiplier = 1 + _damperCalculator.CalculateDynamic(NippleErectionJsf.val, mass, softness, true);
                _softPhysicsHandler.NippleSoftJoints[Side.LEFT].SetAdditionalBaseMultiplier(FAT_SPRING.NAME, springMultiplier);
                _softPhysicsHandler.NippleSoftJoints[Side.LEFT].SetAdditionalBaseMultiplier(FAT_DAMPER.NAME, damperMultiplier);
                _softPhysicsHandler.NippleSoftJoints[Side.RIGHT].SetAdditionalBaseMultiplier(FAT_SPRING.NAME, springMultiplier);
                _softPhysicsHandler.NippleSoftJoints[Side.RIGHT].SetAdditionalBaseMultiplier(FAT_DAMPER.NAME, damperMultiplier);
            }
        }

        internal void Reset()
        {
            _morph.Reset();
            // physics values are not reset here but by SoftPhysicsHandler
        }
    }
}
