using NAT.Common;
using NAT.Common.Models;
using NAT.TM.Handlers;

namespace NAT.TM.Models
{
    sealed class BreastSoftPhysicsParamGroup : SoftPhysicsParamGroup
    {
        readonly BreastSoftPhysicsHandler _softPhysicsHandler;

        internal BreastSoftPhysicsParamGroup(ScriptBase script, BreastSoftPhysicsHandler softPhysicsHandler)
            : base(script)
        {
            _softPhysicsHandler = softPhysicsHandler;
        }

        internal override void SetCallbackFunctions()
        {
            leftParam.OffsetJsf.setCallbackFunction = value =>
            {
                Left.UpdateOffsetValue(value);
                Right.UpdateOffsetValue(ApplyChangesOnlyToLeftParamJsb.val ? 0 : value);
                script.Tabs.PhysicsWindow.OnSoftParamAdjusted(this);
                _softPhysicsHandler.UpdateSoftJointsDistortion(true);
            };

            ApplyChangesOnlyToLeftParamJsb.setCallbackFunction = value =>
            {
                /* Offset */
                Right.UpdateOffsetValue(value ? rightParam.OffsetJsf.defaultVal : leftParam.OffsetJsf.val);

                /* Gradient */
                var lGradient = Left.GradientConfig;
                var rGradient = Right.GradientConfig;
                rGradient.OuterMultiplierJsf.val =
                    value ? rGradient.OuterMultiplierJsf.defaultVal : lGradient.OuterMultiplierJsf.val;
                rGradient.CenterMultiplierJsf.val =
                    value ? rGradient.CenterMultiplierJsf.defaultVal : lGradient.CenterMultiplierJsf.val;
                rGradient.NippleMultiplierJsf.val =
                    value ? rGradient.NippleMultiplierJsf.defaultVal : lGradient.NippleMultiplierJsf.val;
                rGradient.MiddleShiftJsf.val =
                    value ? rGradient.MiddleShiftJsf.defaultVal : lGradient.MiddleShiftJsf.val;
                rightParam.DoSync();

                /* Distortion */
                var lDistortion = Left.LdpConfig;
                var rDistortion = Right.LdpConfig;
                rDistortion.SetCompressionMaxAndMultiplier(
                    value ? rDistortion.CompressionEffectJsf.defaultVal : lDistortion.CompressionEffectJsf.val
                );
                rDistortion.SetCompressionSlope(
                    value ? rDistortion.CompressionSensitivityJsf.defaultVal : lDistortion.CompressionSensitivityJsf.val
                );
                rDistortion.SetStretchingMaxAndMultiplier(
                    value ? rDistortion.StretchingEffectJsf.defaultVal : lDistortion.StretchingEffectJsf.val
                );
                rDistortion.SetStretchingSlope(
                    value ? rDistortion.StretchingSensitivityJsf.defaultVal : lDistortion.StretchingSensitivityJsf.val
                );
                _softPhysicsHandler.UpdateSoftJointsDistortion(Side.RIGHT, true);
            };

            /* Gradient config storable callbacks */
            {
                var lGradient = Left.GradientConfig;
                var rGradient = Right.GradientConfig;
                lGradient.OuterMultiplierJsf.setCallbackFunction = value =>
                    UpdateGradient(value, rGradient.OuterMultiplierJsf);
                lGradient.CenterMultiplierJsf.setCallbackFunction = value =>
                    UpdateGradient(value, rGradient.CenterMultiplierJsf);
                lGradient.NippleMultiplierJsf.setCallbackFunction = value =>
                    UpdateGradient(value, rGradient.NippleMultiplierJsf);
                lGradient.MiddleShiftJsf.setCallbackFunction = value =>
                    UpdateGradient(value, rGradient.MiddleShiftJsf);
            }

            /* Distortion config storable callbacks */
            {
                Left.LdpConfig.CompressionEffectJsf.setCallbackFunction = value =>
                {
                    SetCompressionMaxAndMultiplier(value);
                    _softPhysicsHandler.UpdateSoftJointsDistortion(true);
                    script.Tabs.PhysicsWindow.OnSoftParamAdjusted(this);
                };
                Left.LdpConfig.CompressionSensitivityJsf.setCallbackFunction = value =>
                {
                    SetCompressionSlope(value);
                    _softPhysicsHandler.UpdateSoftJointsDistortion(true);
                    script.Tabs.PhysicsWindow.OnSoftParamAdjusted(this);
                };
                Left.LdpConfig.StretchingEffectJsf.setCallbackFunction = value =>
                {
                    SetStretchingMaxAndMultiplier(value);
                    _softPhysicsHandler.UpdateSoftJointsDistortion(true);
                    script.Tabs.PhysicsWindow.OnSoftParamAdjusted(this);
                };
                Left.LdpConfig.StretchingSensitivityJsf.setCallbackFunction = value =>
                {
                    SetStretchingSlope(value);
                    _softPhysicsHandler.UpdateSoftJointsDistortion(true);
                    script.Tabs.PhysicsWindow.OnSoftParamAdjusted(this);
                };
                SetCompressionMaxAndMultiplier(Left.LdpConfig.CompressionEffectJsf.val);
                SetCompressionSlope(Left.LdpConfig.CompressionSensitivityJsf.val);
                SetStretchingMaxAndMultiplier(Left.LdpConfig.StretchingEffectJsf.val);
                SetStretchingSlope(Left.LdpConfig.StretchingSensitivityJsf.val);
            }
        }

        new void UpdateGradient(float value, JSONStorableFloat rightJsf)
        {
            base.UpdateGradient(value, rightJsf);
            script.Tabs.PhysicsWindow.OnSoftParamAdjusted(this);
        }
    }
}
