using NAT.Common;
using NAT.Common.Handlers;
using NAT.Common.UI;
using NAT.TM.Handlers;
using System.Text;
using UnityEngine;

namespace NAT.TM.UI
{
    sealed class AdditionalClothingSettingsWindow : Window
    {
        readonly ClothingHandler _clothingHandler;
        UIDynamicTextField _profileTransitionStatusTextField;

        internal AdditionalClothingSettingsWindow(ScriptBase script) : base(script)
        {
            _clothingHandler = script.Provider.GetInstance<ChestClothingHandler>();
        }

        protected override void OnBuild()
        {
            if(isBuilt)
            {
                return;
            }

            AddTextField("Target Profile Transition Header", name =>
            {
                string text = new StringBuilder()
                    .AppendSize("\n", 20)
                    .AppendBold("Target Profile Transition")
                    .ToString();
                var uiDynamic = script.CreateCustomTextField(name, text, new Vector2(10, -210), new Vector2(-558, 60));
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.fontSize = 30;
                uiDynamic.UItext.alignment = TextAnchor.LowerCenter;
                return uiDynamic;
            });

            AddButton("Select Target Profile Preset", name =>
            {
                var uiDynamic = script.CreateCustomButton(name, new Vector2(10, -275), new Vector2(-683, 50));
                uiDynamic.AddListener(() => _clothingHandler.LoadProfile(true));
                return uiDynamic;
            });

            AddButton("Set Profile Off As Target", name =>
            {
                var uiDynamic = script.CreateCustomButton(name, new Vector2(405, -275), new Vector2(-730, 50));
                uiDynamic.AddListener(() => _clothingHandler.SetProfileOffAsTarget(_clothingHandler.SelectedProfile));
                return uiDynamic;
            });

            AddButton("Reset", name =>
            {
                var uiDynamic = script.CreateCustomButton(name, new Vector2(753, -275), new Vector2(-950, 50));
                uiDynamic.AddListener(() => _clothingHandler.ResetTargetProfile());
                return uiDynamic;
            });

            AddTextField("Profile Transition Info", name =>
            {
                string text = new StringBuilder()
                    .Append("Transition between the current and target profiles.")
                    .AppendSize("\n", 8)
                    .AppendLine()
                    .Append("The selected target profile and the transition percentage save with the profile.")
                    .ToString();
                var uiDynamic = script.CreateCustomTextField(name, text, new Vector2(545, -530), new Vector2(-555, 160));
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.fontSize = 26;
                uiDynamic.UItext.alignment = TextAnchor.UpperLeft;
                return uiDynamic;
            });

            AddTextField("Profile Transition Status", name =>
            {
                var uiDynamic = script.CreateCustomTextField(name, "", new Vector2(10, -390), new Vector2(-20, 100));
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.fontSize = 26;
                uiDynamic.UItext.alignment = TextAnchor.UpperLeft;
                _profileTransitionStatusTextField = uiDynamic;
                return uiDynamic;
            });

            AddSlider(_clothingHandler.ProfileTransitionJsf, jsf =>
            {
                var uiDynamic = script.CreateCustomSlider(jsf.name, new Vector2(10, -485));
                uiDynamic.valueFormat = "F0";
                ScriptBase.ConfigureCustomSlider(uiDynamic);
                jsf.slider = uiDynamic.slider;
                return uiDynamic;
            });

            AddButton("Set Current State As Current Profile", name =>
            {
                var uiDynamic = script.CreateCustomButton(name, new Vector2(10, -545), new Vector2(-555, 50));
                uiDynamic.AddListener(() => _clothingHandler.SetCurrentTransitionStateAsCurrentProfile());
                return uiDynamic;
            });

            AddTextField("Other", name =>
            {
                string text = new StringBuilder()
                    .AppendSize("\n", 20)
                    .AppendBold("Other")
                    .ToString();
                var uiDynamic = script.CreateCustomTextField(name, text, new Vector2(10, -630), new Vector2(-558, 60));
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.fontSize = 30;
                uiDynamic.UItext.alignment = TextAnchor.LowerCenter;
                return uiDynamic;
            });

            AddSlider(_clothingHandler.MorphingTransitionJsf, jsf =>
            {
                var uiDynamic = script.CreateCustomSlider(jsf.name, new Vector2(10, -755));
                uiDynamic.valueFormat = "F2";
                ScriptBase.ConfigureCustomSlider(uiDynamic);
                jsf.slider = uiDynamic.slider;
                return uiDynamic;
            });

            AddTextField("Morphing Transition Info", name =>
            {
                string text = new StringBuilder()
                    .Append("Timer for how quickly morphing offsets are animated when a clothing item or its profile is toggled on/off.")
                    .AppendFormat(" This saves with the current {0} instance.", Constant.NATURALIS)
                    .ToString();
                var uiDynamic = script.CreateCustomTextField(name, text, new Vector2(545, -755), new Vector2(-558, 120));
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.fontSize = 26;
                uiDynamic.UItext.alignment = TextAnchor.UpperLeft;
                return uiDynamic;
            });

            AddToggle(_clothingHandler.ProfileOffWhenUndressingJsb, jsb =>
            {
                var uiDynamic = script.CreateCustomToggle(jsb.name, new Vector2(10, -845), new Vector2(-555, 50));
                jsb.toggle = uiDynamic.toggle;
                return uiDynamic;
            });

            AddTextField("Undress Info", name =>
            {
                const string text = "Deactivates the profile when Undress is toggled on in the item's physics settings (if sim item). This saves with the profile.";
                var uiDynamic = script.CreateCustomTextField(name, text, new Vector2(545, -915), new Vector2(-558, 120));
                uiDynamic.backgroundColor = Color.clear;
                uiDynamic.UItext.fontSize = 26;
                uiDynamic.UItext.alignment = TextAnchor.UpperLeft;
                return uiDynamic;
            });
        }

        internal override void UpdateDiagnostics()
        {
            _profileTransitionStatusTextField.text = ClothingUIHelpers.TransitionStatus(diagnosticsSb, _clothingHandler);
            ((ClothingWindow) parent).UpdateAdjustmentsInfo();
        }
    }
}
